
=head1 NAME

B<Cost.pm> -- query cost computation.

=head1 DESCRIPTION

Module includes the implementation of a cost function for the query
optimizer of I<Qios>. The standard relational cost functions are used
to estimate the number of bytes processed by the physical query tree
including the cost of materialization of the intermediate query
evaluation results. 

=head1 Statistics

The statistics on the classes is generated dynamically triggered by
the query optimizer and when the class is loaded into the database.
The following data is gathered for a given class: the number of class
objects, the size of objects and the cardinality of attributes. 

=head1 Cost computation

The computation of the estimation of the size of the result for the
operations I<select>, I<project> and I<join> is realized using
standard relational formulas based on the selectivity of attributes
and presupposing the normal distribution of attributes values.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut


package Qios::Cost;


use Qios::Glb;
use Qios::Osm;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( &cost );





BEGIN {
    %attss = ();

    %cmnc = (
        "a1" => { "a2" => 100,
                  "a3" => 200,
                  "a4" => 100 },
        "a2" => { "a3" => 300,
                  "a4" => 100 },
        "a3" => { "a4" => 400 }
    );

    $szcfac = 15;                  # factor for size
    $matfac = 0.8;                 # factor for materialization
    $exefac = 1;                   # factor for execution
}




sub cost {
    my $n = shift;
    CA: {
        if ($n->{nodOp} == $op{select}) {
            &cost($n->{nodOuter});
            return &costSelect($n);
        }
        if ($n->{nodOp} == $op{join}) {
            &cost($n->{nodOuter});
            $n->{nodInner}->{phsMater} = $Ok;
            &cost($n->{nodInner});
            my $cos = &costJoin($n);
            delete $n->{nodInner}->{phsMater};
            return $cos;
        }
        if ($n->{nodOp} == $op{project}) {
            &cost($n->{nodOuter});
            return &costProject($n);
        }
        if ($n->{nodOp} == $op{copy}) {
            &cost($n->{nodOuter});
            return &costCopy($n);
        }
        if (($n->{nodOp} == $op{ext}) ||
            ($n->{nodOp} == $op{ins})) {
            return &costAM($n);
        }
    }
    return undef;
}





sub costSelect() {
    my $n = shift;

    $n->{phsOp} = $op{phsqsl};  

    my $prs = $n->{nodOuter}->{cosRecsz};
    my $prn = $n->{nodOuter}->{cosRecnm};

    $n->{cosRecsz} = $prs;

    my $slc;
    if (defined($n->{nodParam}->{cosSlct})) { 
        $slc = $n->{nodParam}->{cosSlct}; 
    } else { 
        $slc = 1/10;  # ragu
    } 

    $n->{cosRecnm} = $prn * $slc;
    my $bsz = $n->{cosRecsz} * $n->{cosRecnm};

    my $exn = $prn*$prs; 

    my $matc = 0;
    if ($n->{phsMater}) {
        $matc = $bsz * $matfac;
    }

    $n->{optCost} = int($n->{nodOuter}->{optCost} + 
                        $exn * $exefac + $bsz * $szcfac + $matc);
}



sub costJoin() {
    my $n = shift;

    my $prs1 = $n->{nodOuter}->{cosRecsz};
    my $prn1 = $n->{nodOuter}->{cosRecnm};
    my $prs2 = $n->{nodInner}->{cosRecsz};
    my $prn2 = $n->{nodInner}->{cosRecnm};

    $n->{cosRecsz} = $prs1+$prs2;

    my $slc;
    if (defined($n->{nodParam}->{cosSlct})) { 
        $slc = $n->{nodParam}->{cosSlct};
    } else {
        $slc = 1/10;    # ragu
    }

    $n->{cosRecnm} = $prn1*$prn2*$slc;
    my $bsz = $n->{cosRecsz} * $n->{cosRecnm};

    my $exn;
    $n->{phsOp} = &Qios::Eval::decide_phop_join($n);
    if ($n->{phsOp} eq 'phixjn') {
        $exn = $prn1*$prs1  + $n->{cosRecnm}*$prs2;
    } else {
        $exn = $prn1*$prs1 + $prn1*$prn2*$prs2;
    }

    my $matc = 0;
    if ($n->{phsMater}) {
        $matc = $bsz * $matfac;
    }

    $n->{optCost} = int($n->{nodOuter}->{optCost} + 
                        $n->{nodInner}->{optCost} + 
                        $exn * $exefac + $bsz * $szcfac + $matc);
}



sub costProject() {
    my $n = shift;

    my $prs = $n->{nodOuter}->{cosRecsz};
    my $prn = $n->{nodOuter}->{cosRecnm};   #?
    
    $n->{cosRecsz} = &stat_recsz(&obj_fetch($n->{nodType}));
    $n->{cosRecnm} = $prn; 
    my $bsz = $n->{cosRecsz} * $n->{cosRecnm};
    
    my $exn = $prn*$prs;

    my $matc = 0;
    if ($n->{phsMater}) {
        $matc = $bsz * $matfac;
    }

    $n->{optCost} = int($n->{nodOuter}->{optCost} + 
                        $exn * $exefac + $bsz * $szcfac + $matc);
}



sub costCopy() {
    my $n = shift;

    $n->{cosRecsz} = $n->{nodOuter}->{cosRecsz};
    $n->{cosRecnm} = $n->{nodOuter}->{cosRecnm};

    $n->{optCost} = int($n->{nodOuter}->{optCost} + 
                        &ccost($n->{nodOuter}->{nodSite},$n->{nodSite})
                        * $n->{cosRecsz}*$n->{cosRecnm});
}



sub costAM() {
    my $n = shift;

    $n->{cosRecsz} = &obj_fetch($n->{nodOuter})->{$Qios::Osm::recsz};
    $n->{cosRecnm} = &obj_fetch($n->{nodOuter})->{$Qios::Osm::recnm};

    $n->{optCost} = int($n->{cosRecsz} * $n->{cosRecnm} * $szcfac);
}








sub ccost() {
    my ($s1,$s2) = @_;
    
    if (defined $cmnc{$s1}->{$s2}) {
        $cc = $cmnc{$s1}->{$s2};
    } else {
        $cc = 300; 
    }
    
    return $cc;
}





sub statistics() {
    my ($pk,$cid) = @_;
    log_say1 "Computing statistics for $cid.\n";

    my $cl = &obj_fetch($cid);
    &stat_start($cl);

    my $sc = Qios::Osm->scan($cid,$op{phsqex});
    while (defined ($ob = Qios::Osm->next($sc))) {
        &stat_object(&obj_fetch($cid),$ob);
    }
    Qios::Osm->close($sc);
    
    &stat_finish(&obj_fetch($cid));
    return $Ok;
}



sub stat_start() {
    my $cl = shift;

    if (!$cl->{$Qios::Osm::kind}) {
        setError CosNotAClass;
        return undef;
    }

    Qios::Cost->stat_drop($cl->{$idnt});

    $cl->{$Qios::Osm::recsz} = &stat_recsz($cl);
    $cl->{$Qios::Osm::recnm} = 0;
}



sub stat_finish() {
    my $cl = shift;

    my $at;
    my @ks = keys %{$cl->{$Qios::Osm::stat}};
    foreach $at (@ks) {
        $cl->{$Qios::Osm::stat}->{"\#$at"} = 
            1 / (scalar keys %{$cl->{$Qios::Osm::stat}->{$at}}); 
    }
}



sub stat_object() {
    my ($cl,$ob) = @_;

    my $at;
    L1: foreach $at (keys %{$ob}) {
        next L1 if ((&attr_sys($at) || $ob->attr_mvd($at)));
        if (exists $cl->{$Qios::Osm::stat}->{$at}->{$ob->{$at}}) {
            $cl->{$Qios::Osm::stat}->{$at}->{$ob->{$at}}++;
        } else {
            $cl->{$Qios::Osm::stat}->{$at}->{$ob->{$at}} = 1;
        }
    }

    $cl->{$Qios::Osm::recnm}++;
}



sub stat_recsz() {
    my $cl = shift;

    my $s = 0;
    L1: foreach $at (keys %{$cl}) {
        next L1 if (&attr_sys($at));
        if (&attr_mvd($cl,$at)) {
            $s += 4;
        } elsif ($cl->{$at} eq 'number') {
            $s += 6;
        } elsif ($cl->{$at} eq 'string') {
            $s += 30;
        } else {
            $s += 10;
        }
    }
    return $s;
}



sub stat_print {
    my ($pk,$cid) = @_;
    my ($k,$k1,$v,$v1);

    my $cl = &obj_fetch($cid);
    my @ks = sort keys %{$cl->{$Qios::Osm::stat}};
    foreach $k (@ks) {
        $v = $cl->{$Qios::Osm::stat}->{$k};
        if ($k =~ /^\#.*/) {
    	    print "$k = $v\n";
        } else { 
            print "$k stat:\n";
            while (($k1,$v1) = each %{$v}) {
                print "$k1:$v1 ";
            }
            print "\n";
        }
    }
    return $Ok;
}



sub stat_drop {
    my ($pk,$cid) = @_;
    my $cl = &obj_fetch($cid);

    my ($at,$vl);
    if (exists $cl->{$Qios::Osm::stat}) {
        log_say1 "Dropping statistics for class $cl->{$Qios::Osm::idnt}.\n";

        foreach $at (keys %{$cl->{$Qios::Osm::stat}}) {
            if ($at =~ /^\#.*/) {
                undef $cl->{$Qios::Osm::stat}->{$at};
            } else {
                undef %{$cl->{$Qios::Osm::stat}->{$at}};                
            }
	}
        undef %{$cl->{$Qios::Osm::stat}};
        delete $cl->{$Qios::Osm::stat};
    }

    return $Ok;
}





sub slct_join {
    my $n = shift;


    my ($af,$cl,$at);
    my ($f0,$c0) = (0,0); 
    foreach $af (@{$n->{nodParam}->{nodAtts}}) { 
        $c0++;
        ($cl,$at) = split $Qios::Osm::attrGlue, $af;

        my $clp = &obj_fetch($cl);
        if (defined $clp->{$Qios::Osm::stat}->{"\#$af"}) { 
            $f0 = $f0 + $clp->{$Qios::Osm::stat}->{"\#$af"};
        } else {
            $f0 = $f0 + 0.1;
        }
    }
    
    if ($c0 != 0) { 
        $n->{nodParam}->{cosSlct} = $f0/$c0;
    } else {
        $n->{nodParam}->{cosSlct} = 0.1;
    }
    return $Ok;
}



sub slct_select {
    my $n = shift;

    my $af = $n->{nodParam}->{nodAtts}->[0];
    my ($cl,$at) = split $Qios::Osm::attrGlue, $af;


    my $clp = &obj_fetch($cl);
    my $f1 = undef;
    if (defined $clp->{$Qios::Osm::stat}->{"\#$af"}) { 
        $f1 = $clp->{$Qios::Osm::stat}->{"\#$af"};
    } else {
        $f1 = 0.1;
    }

    $n->{nodParam}->{cosSlct} = $f1;
    return $Ok;
}


1;



















