
=head1 NAME

Eval.pm - query executor.

=head1 DESCRIPTION

This module implements the query executor of Qios. The query executor
comprises the computation of query execution plan and the query
evaluation engine.

=head1 Query execution plan

The physical query execution plan is computed from the optimized
query trees obtained using module I<Optor.pm>. The selection of query
execution plan is implemented by the procedure which computes
sub-optimal physical execution plan by selecting physical operations
for all logical operations (query nodes) forming the physical query
tree in a bottom-up manner. 

The result of the logical query optimization are bushy query trees.
The results of the inner sub-trees have to be materialized in order to
avoid their repeating evaluation. Each of the logical operations
(query nodes) can be implemented using one or more physical operations
which are listed in the following section. The procedures for the
selection of physical operations are integrated in the routines that
open scan operator for query nodes (logical operators). The use of
physical operators based on B+-trees are planned to be implemented.

=head1 Physical operations

After the selection of the physical operations, the same skeleton of
the query tree is employed as the iterator tree for the evaluation of
physical query. The physical operations are implemented as scans. The
currently implemented physical operations are the following.

I<Access methods>
    
    phsqex  - sequentially extension
    phixex  - hash-index on extension
    phsqin  - sequentially all instances
    phixin  - hash-index on all instances

I<Select>

    phsqse  - select loop
    phixse  - hash access for equality

I<Project>

    phsqpr  - project loop

I<Join>
   
    phsqjn  - nested-loop join
    phixjn  - hash-join for equality

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Eval;


use Qios::Glb;
use Qios::Rsm;
use Qios::Osm;




sub eval_query {
    my ($pk,$n,$mt) = @_;
    log_say1 "Evaluating a query.\n";

    $n->scan($mt);
    my $cn = 0;
    my $ob;
    while (defined ($ob = $n->next())) {
         Qios::Glb->print($ob);
         $cn++;
    }
    $n->close();
    print "\n$cn objects printed\n";
}





sub scan_AM {
    my ($n,$at,$av) = @_;
    my $sc;

    $n->{phsOp} = &decide_phop_AM($n,$at);

    $n->{phsScan} = Qios::Osm->scan($n->{nodOuter},
                              $n->{phsOp},$at,$av);
    $n->{phsCobj} = Qios::Osm->obj_new([$n->{nodType}],$isgr)->{$Qios::Osm::idnt};
    $n->{phsMater} = $Ok;

    return $Ok;    
}



sub close_AM {
    my $n = shift;

    if (defined $n->{phsScan}) {
        Qios::Osm->close($n->{phsScan});
    }
    if (defined $n->{phsCobj}) { 
        &obj_fetch($n->{phsCobj})->obj_delete();   # clean? (!)
    }

    delete $n->{phsScan}; 
    delete $n->{phsCobj};
    delete $n->{phsOp};

    return $Ok;
}



sub next_AM {
    my $n = shift;

    my $n1; 
    if (!defined ($n1 = Qios::Osm->next($n->{phsScan}))) { 
        if (&error()) {
            setError EvalErrNextObj; 
        }
        return undef; 
    } 
 
    my $n2 = &obj_fetch($n->{phsCobj}); 
    $n2->obj_clean(); 

    $n1->obj_copy($n2); 
    my $l = [];
    &obj_fetch($n->{nodType})->obj_atts($l);
    $n2->obj_project($l); 

    return $n2;
}



sub scan_select {
    my ($n,$mt,$at,$av) = @_;


    $n->{nodParam}->{nodBind} = $False;

    if ($n->{phsMater} == $Ok) {

        if (defined $at) {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType},
                                      $op{phixex},$at,$av);
        } else {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType},
                                      $op{phsqex});
        }
    } elsif ($mt) {

        $n->{phsMater} = $Null;    
        log_say1 "Materializing $n->{nodExpr} in class $n->{nodType}.\n";

        &scan_select($n);
        my $n1;
        while (defined ($n1 = &next_select($n))) {
        }
        &close_select($n);

        $n->{phsMater} = $Ok;
        &scan_select($n,$False,$at,$av);

    } else {

        $n->{nodOuter}->scan();
        if ($n->{phsMater} == $False) {
            $n->{phsCobj} = 
            Qios::Osm->obj_new([$n->{nodType}],$isgr)->{$Qios::Osm::idnt};
        }
    }

    return $Ok;
}



sub close_select {
    my $n = shift;

    if ($n->{phsMater} == $Ok) {
        Qios::Osm->close($n->{phsScan});
    } else {
        $n->{nodOuter}->close();

        if ($n->{phsMater} == $False) {
            if (defined $n->{phsCobj}) {
                &obj_fetch($n->{phsCobj})->obj_delete();
            }
        }
    }
 
    delete $n->{phsScan}; 
    delete $n->{phsCobj};
    delete $n->{phsOp};

    return $Ok;    
}



sub next_select {
    my $n = shift;

    my $n1;
    if ($n->{phsMater} == $Ok) {
        my $nob = Qios::Osm->next($n->{phsScan});
        if (&error()) {
            setError EvalErrNextObj; 
            return undef;
        }
        if (defined $nob) {
            $n->{phsCobj} = $nob->{$Qios::Osm::idnt};
        }    
        return $nob;

    } else {
        L1: while () {
            if (!defined ($n1 = $n->{nodOuter}->next())) {
                if (&error()) {
                    setError EvalErrNextObj;
                } 
                return undef; 
            }

            if (&eval_fpara($n->{nodParam}) == $Ok) {
                last L1;
            } 
        }
    }

    my $n2;
    if ($n->{phsMater} == $Null) {
        $n2 = Qios::Osm->obj_new([$n->{nodType}],$isgr);
    } else {
        $n2 = &obj_fetch($n->{phsCobj});
        $n2->obj_clean();
    }

    $n1->obj_copy($n2);
    my $l = [];
    &obj_fetch($n->{nodType})->obj_atts($l);
    $n2->obj_project($l); 
    $n->{phsCobj} = $n2->{$Qios::Osm::idnt};

    return $n2;
}



sub scan_project {
    my ($n,$mt,$at,$av) = @_;

    if ($n->{phsMater} == $Ok) {

        if (defined $at) {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType}, 
                                      $op{phixex},$at,$av);
        } else {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType}, 
                                      $op{phsqex});
        }
    } elsif ($mt) {

        $n->{phsMater} = $Null;    
        log_say1 "Materializing $n->{nodExpr} in class $n->{nodType}.\n";

        &scan_project($n);
        my $n1;
        while (defined ($n1 = &next_project($n))) { 
        }
        &close_project($n);

        $n->{phsMater} = $Ok;
        &scan_project($n,$False,$at,$av);
    } else {

        $n->{nodOuter}->scan();
        if ($n->{phsMater} == $False) {
            $n->{phsCobj} = 
            Qios::Osm->obj_new([$n->{nodType}],$isgr)->{$Qios::Osm::idnt};
	}
    }

    return $Ok;    
}



sub close_project {
    my $n = shift;

    if ($n->{phsMater} == $Ok) {

        Qios::Osm->close($n->{phsScan});
    } else  {

        $n->{nodOuter}->close();

        if ($n->{phsMater} == $False) {
            if (defined $n->{phsCobj}) {
                &obj_fetch($n->{phsCobj})->obj_delete();
            }
        }
    }

    delete $n->{phsScan}; 
    delete $n->{phsCobj};
    delete $n->{phsOp};

    return $Ok;    
}



sub next_project {
    my $n = shift;

    my $n1;
    if ($n->{phsMater} == $Ok) {

        my $nob = Qios::Osm->next($n->{phsScan});
        if (&error()) {
            setError EvalErrNextObj; 
            return undef;
        }
        if (defined $nob) {
            $n->{phsCobj} = $nob->{$Qios::Osm::idnt};
        }
        return $nob;

    } else {
        if (!defined ($n1 = $n->{nodOuter}->next())) { 
            if (&error()) {
                setError EvalErrNextObj;
            }
            return undef;
        }
    }

    my $n2;
    if ($n->{phsMater} == $Null) {
        $n2 = Qios::Osm->obj_new([$n->{nodType}],$isgr);
    } else {
        $n2 = &obj_fetch($n->{phsCobj});
        $n2->obj_clean();
    }

    $n1->obj_copy($n2);
    my $l = $n->{nodParam}->{nodCval};
    $n2->obj_project($l); 
    $n->{phsCobj} = $n2->{$Qios::Osm::idnt};

    return $n2;
}



sub scan_join {
    my ($n,$mt,$at,$av) = @_;


    $n->{nodParam}->{nodBind} = $False;

    if ($n->{phsMater} == $Ok) {

        if (defined($at)) {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType}, 
                                      $op{phixex},$at,$av);
        } else {
            $n->{phsScan} = Qios::Osm->scan($n->{nodType},
                                      $op{phsqex});
        }
    } elsif ($mt) {

        $n->{phsMater} = $Null;    
        log_say1 "Materializing $n->{nodExpr} in class $n->{nodType}.\n";

        &scan_join($n);   # we donot have relat yet
        my $n1;
        while (defined ($n1 = &next_join($n))) { 
        }
        &close_join($n);

        $n->{phsMater} = $Ok;
        &scan_join($n,$False,$at,$av);

    } else {

        $n->{nodOuter}->scan(); 
        my $n1;
        if (!defined ($n1 = $n->{nodOuter}->next())) {
            if (&error()) {
                setError EvalErrNextObj; 
                return undef;
            }
            return $Ok;
        }
        $n->{phsLastOut} = $n1->{$Qios::Osm::idnt};
        
        if ($n->{phsOp} == $op{phixjn}) {
            $n->{phsIxAtrVal} = &Qios::Osm::attr_get(
                                    &obj_fetch($n->{nodOuter}->{phsCobj}),
                                    $n->{phsIxAtrOut});
            $n->{nodInner}->scan($Ok,$n->{phsIxAtrIn},$n->{phsIxAtrVal});
        } else {
            $n->{nodInner}->scan($Ok); #070901
        }

    }

    return $Ok;    
}



sub close_join {
    my ($n,$a) = @_;

    if ($n->{phsMater} == $Ok) {

        Qios::Osm->close($n->{phsScan});
    } else {

        $n->{nodOuter}->close();
        $n->{nodInner}->close();

        if ($n->{phsMater} == $False) {
            if (defined $n->{phsCobj}) {
                &obj_fetch($n->{phsCobj})->obj_delete();
            }
        }
    }

    delete $n->{phsScan};
    delete $n->{phsLastOut};
    delete $n->{phsCobj};
    delete $n->{phsIxAtrOut};
    delete $n->{phsIxAtrIn};
    delete $n->{phsIxAtrVal};
    delete $n->{phsOp};

    return $Ok;    
}



sub next_join {
    my $n = shift;
    my $n1;
    my $eok = $False;

    if ($n->{phsMater} == $Ok) {
  
        my $nob = Qios::Osm->next($n->{phsScan});
        if (&error()) {
            setError EvalErrNextObj; 
            return undef;
        }
        if (defined $nob) {
            $n->{phsCobj} = $nob->{$Qios::Osm::idnt};
        } 
        return $nob;

    } else {

        my ($n11,$n12);
        if (!defined $n->{phsLastOut}) {
            return undef;
        }

        $n11 = &obj_fetch($n->{phsLastOut}); 
        L1: while () {

            if (!defined ($n12 = $n->{nodInner}->next())) {
                if (&error()) {
                    setError EvalErrNextObj; 
                    return undef;

                } elsif (!defined ($n11 = $n->{nodOuter}->next())) {
                    if (&error()) {
                        setError EvalErrNextObj; 
                    }
                    return undef;

                } else {
                    $n->{phsLastOut} = $n11->{$Qios::Osm::idnt}; 

                    $n->{nodInner}->close();     
                    if ($n->{phsOp} == $op{phixjn}) {
                        $n->{phsIxAtrVal} = 
                            &Qios::Osm::attr_get(
                                  &obj_fetch($n->{nodOuter}->{phsCobj}),
                                  $n->{phsIxAtrOut});
                        $n->{nodInner}->scan($Ok,
                                             $n->{phsIxAtrIn},
                                             $n->{phsIxAtrVal});
                    } else {
                        $n->{nodInner}->scan($False);
                    }
                    next L1;
                }
            }

            if (&eval_fpara($n->{nodParam})) {
                if (!defined 
                     ($n1 = &obj_fetch(&Qios::Osm::obj_join($n11,
                                                      $n12,
                                                      $n->{nodType})))) {
                    return undef;
                }
                $eok = $Ok;
                last L1;
            } else {
                next L1; 
            }
        }
    }

    if (!$eok) {
        return undef;
    }

    if ($n->{phsMater} == $False) {
        if (defined $n->{phsCobj}) {
            &obj_fetch($n->{phsCobj})->obj_delete();
        }
    }

    my $l = [];
    &obj_fetch($n->{nodType})->obj_atts($l);
    $n1->obj_project($l); 
    $n->{phsCobj} = $n1->{$Qios::Osm::idnt};

    return $n1;
}



sub scan_copy {
    my $n = shift;
}



sub close_copy {
    my $n = shift;
}



sub next_copy {
    my $n = shift;
}




sub decide_phop_join() {
    my $n = shift;
    my $op = $n->{nodParam}->{nodOuter};
 

    if ($op->{nodOp} == $op{ieq} &&
        $op->{nodOuter}->{nodOp} == $op{attv} &&
        $op->{nodInner}->{nodOp} == $op{attv}) {

        $n->{phsIxAtrOut} = $op->{nodOuter}->{nodName};
        $n->{phsIxAtrIn} = $op->{nodInner}->{nodName};
        return $op{phixjn};
    } else {

        return $op{phsqjn};
    }
}



sub decide_phop_AM() {
    my ($n,$a) = @_;

    if ($n->{nodOp} == $op{ext}) {
        if (defined($a)) {
            return $op{phixex};
        } else {
            return $op{phsqex};
        }
    } elsif ($n->{nodOp} == $op{ins}) {
        if (defined($a)) {
            return $op{phixin};
        } else {
            return $op{phsqin};
        }
    }
}





sub eval_fpara {
    my $n = shift;

    if (!$n->{nodBind}) {
        my $pa;
        foreach $pa (@{$n->{nodPvars}}) {
            $pa->{nodOuter} = $n->{nodVRmap}->{$pa->{nodName}};
        }

        $n->clean_fpara();
        if (!($n->typ_fpara())) {
            setError EvalTypeErr;
            return undef;
        }

        $n->{nodBind} = $Ok;
    }

    return &eval_pexpr($n->{nodOuter});
}



sub eval_pexpr {
    my $n = shift;

    C1: {
        if (&member($n->{nodOp},
                    [$op{var},$op{attv},$op{const}])) {
            return &eval_term($n);
        }

        if (&member($n->{nodOp},\@binops)) {
            return &eval_bin($n);
        }

        if (&member($n->{nodOp},\@uniops)) {
            return &eval_uni($n);
        }
    }

    return undef;
}



sub eval_bin {
    my $n = shift;

    my $e1 = &eval_pexpr($n->{nodOuter});
    my $e2 = &eval_pexpr($n->{nodInner});

    if ((!defined $e1) || (!defined $e2)) {
         return $False;
    }

    my $op = $n->{nodOp};
    C1: {
        if ($op == $op{land}) {
            return $e1 && $e2;
        } 

        if ($op == $op{lor}) {
            return $e1 || $e2;
        }    

        if ($op == $op{plus}) {
            return $e1 + $e2;
        }

        if ($op == $op{minus}) {
            return $e1 - $e2;
        }

        if ($op == $op{times}) {
            return $e1 * $e2;
        }

        if ($op == $op{delete}) {
            return $e1 / $e2;
        }

        if ($op == $op{ieq}) {
            return $e1 eq $e2;
        }

        if ($op == $op{deq}) {
            return $e1 == $e2;               # to come (!)
        }

        if ($op == $op{mtch}) {
            return ($e1 =~ /$e2/);  
        }

        if ($op == $op{lt}) {
            return $e1 < $e2;    
        }

        if ($op == $op{le}) {
            return $e1 <= $e2;    
        }

        if ($op == $op{gt}) {
            return $e1 > $e2;    
        }

        if ($op == $op{ge}) {
            return $e1 >= $e2;    
        }

        if ($op == $op{in}) {
            return &member($e1,$e2);    
        }

        if ($op == $op{subset}) {
            return &subset($e1,$e2);
        }
    }

    setError EvalNoSuchOp;
    return undef;
}



sub eval_uni {
    my $n = shift;

    my $e1 = &eval_pexpr($n->{nodOuter});

    if ($n->{nodOp} == $op{uneg}) {
        return ! $e1;
    }

    setError EvalNoSuchOp;
    return undef;
}



sub eval_term {
    my $n = shift;

    if ($n->{nodOp} == $op{const}) {
        return $n->{nodCval};
    }

    if ($n->{nodOp} == $op{var}) {
        $n->{nodCval} = $n->{nodOuter}->{phsCobj};
        return $n->{nodCval};
    }

    if ($n->{nodOp} == $op{attv}) {


        my ($e1,$vl);
        if (!defined ($e1 = &eval_pexpr($n->{nodOuter}))) {
            return undef;
        }
        
        if (!&Qios::Osm::obj_exists($e1)) {
            setInternal EvalIllObj;
            return undef;
        }
        if (!defined ($n->{nodCval} = 
                      &obj_fetch($e1)->attr_get($n->{nodName}))) {
            return undef;
        }
        return $n->{nodCval};  
    }

    return undef;
}



1;



