
=head1 NAME

I<Glb.pm> -- common data structures and procedures.

=head1 DESCRIPTION

Module I<Glb.pm> includes the definitions of the common data
structures for the representation of Qios operations, error messages
and datafile configuration records. Further, module I<Glb.pm>
implements the mechanisms for error handling and system logging.

=head2 Operations

The operations of Qios are represented in the associative array B<%op>
by which the integer numbers are assigned to the keywords
representing operations. The keywords are used to symbolically
identify the operations in the code. For the purpose of fast
implementation, Qios manipulates internally the integer codes of
operations.

=head2 Errors

The error messages are represented using the associative array
B<%error> with the keywords identifying errors and the corresponding
error messages as strings.

The state of a module describes the position in which the module is in
a given timepoint. The state can be one of the following: I<Ok>,
I<Info>, I<Warn>, I<Error> and I<Fatal>. The state of the particular
module is recorded using the variable I<$state>. The state of the
system is recorded in the module I<Glb.pm>. (Note: only system wide
state is currently used.)

The errors are gathered on the stack I<@error> to be able to make a
precise diagnose of the error possibly from more than one modules. The
main function for checking the errors in the system examines only the
size of stack I<@error>. 

=head2 Logging

Two log streams are realized. The standard log stream is used to
record system events during normal operation. The debug log stream is
used for debugging. 

Log level of the log record denotes the importance (or severity) of
the log record. The log level thresholds (parameters of system) are
used to filter the standard and debug log streams: only log records
which levels are smaller than or equal to the threshold log level are
channeled to the stream. The log level of the log record is controlled
by using the appropriate procedure to report the log record.

The log level thresholds are specified in the datafile configuration:
I<loglvl> variable records the level of standard logging and the
variable I<dbglvl> filters debug log stream. Furthermore, the log
streams can be turned off and on using the configuration parameters: 
I<logon> and I<debug>.

The procedures for reporting the log records can report various types
of events: all types of errors, info messages, descriptions,
announcements, printing various debug info, etc. The procedures are:
I<log_sayX>, I<log_heyX>, I<log_warn>, I<log_error>, I<log_fatal>,
I<log_enterX>, I<log_exitX> and I<log_infoX>. The variable 'X' in the
procedure names denotes the reporting log level (0..2) of the log
record.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut


package Qios::Glb;


require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( %state %error %op $Ok $False $Null $nnl $idnt $db %db
              @binops @uniops &member &subset &trim &op_key &nl &print 
              $ftc &error &setError &setWarn &setInternal &getError 
              &outputErrors &cleanErrors &sysAttr &log_say0 &log_say1 
              &log_say2 &log_hey0 &log_hey1 &log_hey2 &log_enter0 
              &log_enter1 &log_enter2 &log_exit0 &log_exit1 &log_exit2 
              &log_info0 &log_info1 &log_info2 );



BEGIN { 

    $Null   = -1;
    $Ok     = 1;
    $False  = 0;

    $fnl = $False;    # first-new-line magic
    $ftc = $False;    # first type check
    
    $guid = "Heyon";

    $defaultSite = 'a1';


    $config = './config';
    $dbname = undef;    # most relevant
    %db = ();
    $db = undef;
}



BEGIN {

    $n = 1;    # node identifiers

    %op = (                 # all operations

    ieq     => $n++,        # ident equality
    deq     => $n++,        # deep equality 
    leq     => $n++,        # local equality
    ext     => $n++,        # extension
    ins     => $n++,        # instances 
    cls     => $n++,        # class_of 
    lto     => $n++,        # more specific 
    leo     => $n++,        # more specific or equal
    gto     => $n++,        # more general 
    geo     => $n++,        # more general or equal
    lub     => $n++,        # least upper bound
    glb     => $n++,        # great lower bound
    set     => $n++,        # set constructor
    in      => $n++,        # set membership
    subset  => $n++,        # subset 
    mtch    => $n++,        # pattern matching

    lt      => $n++,        # less
    le      => $n++,        # less or equal
    gt      => $n++,        # greater
    ge      => $n++,        # greater or queal
    plus    => $n++,        # +
    minus   => $n++,        # -
    times   => $n++,        # *
    delete  => $n++,        # /
    addr    => $n++,        # \
    lor     => $n++,        # logical or
    land    => $n++,        # logical and
    uneg    => $n++,        # unary negation

    select  => $n++,        # selection
    project => $n++,        # projection
    intrsc  => $n++,        # set intersection
    union   => $n++,        # set union
    join    => $n++,        # join
    copy    => $n++,        # copy
    group   => $n++,        # group-by
    unnest  => $n++,        # unnest
    flat    => $n++,        # flatten
    coll    => $n++,        # tuple collapse

    phsqex  => $n++,        # extension sequential
    phixex  => $n++,        # extension with index
    phsqin  => $n++,        # instances sequential 
    phixin  => $n++,        # instances with index
    phsqsl  => $n++,        # select sequential
    phixsl  => $n++,        # select with index
    phsqpr  => $n++,        # project with loop
    phsqjn  => $n++,        # nested-loops join
    phixjn  => $n++,        # indexed nested-loop join

    name    => $n++,        # name
    const   => $n++,        # constant
    attv    => $n++,        # attribute 
    obj     => $n++,        # object name
    var     => $n++,        # variable
    fpara   => $n++,        # filter param expression
    cpara   => $n++,        # copy param
    tpara   => $n++,        # project param 
    span    => $n++,        # span for rules

    lpar    => $n++,        # lpar
    rpar    => $n++,        # rpar
    lbrk    => $n++,        # lbrk
    rbrk    => $n++,        # rbrk
    vdef    => $n++,        # var definition
    cntx    => $n++,        # cntx binding
    rdef    => $n++,        # rule definition
    lcon    => $n++,        # list constructor
    dot     => $n++         # dot operation

    );

    @binops = ( $op{ieq}, $op{deq}, $op{leq}, $op{lor}, $op{land}, 
                $op{lto}, $op{leo}, $op{gto}, $op{geo}, $op{lt}, $op{le}, 
                $op{gt}, $op{ge}, $op{plus}, $op{minus}, $op{times}, 
                $op{delete}, $op{in}, $op{subset}, $op{mtch} );

    @uniops = ( $op{uneg}, $op{lub}, $op{glb}, $op{cls} );

    $trmchr = ["\ ","\:","\#","\(","\)","\,","\.","\[","\]",
               "\=","\>","\<","\@","\*","\+","\-","\&","\/",
               "\!","\t","\n","\r","\f"];

}



BEGIN {

    %state = ( 
        Ok    => 0,        # ok
        Info  => 1,        # info messg
        Warn  => 1,        # warning
        Error => 2,        # error
        Fatal => 3         # internal error
    ); 

    %error = (


        GlbErrOpenLog      => "Error on opening log file",
        GlbErrOpenConf     => "Error on opening config file",
        GlbNoDbDir         => "Qios base directory not defined",


        ParseConsErr       => "Error in constant",
        ParseVarErr        => "Error in variable",
        ParseVarNotExist   => "Variable has not appeared before",
        ParseObjNotExist   => "Object does not exist",

        ParseDotExpct      => "Dot expected",
        ParseLconExpct     => "Comma expected",
        ParseLParExpct     => "Left parenthesis expected",
        ParseRParExpct     => "Right parenthesis expected",
        ParseLBrkExpct     => "Left bracket expected",
        ParseRBrkExpct     => "Right bracket expected",
        ParseNameExpct     => "Name expected",
        ParseTabExprExpct  => "Table expression expected",
        ParseRVdefExpct    => "Parameter or span definition expected",
        ParseQueOpExpct    => "Query operation was expected",
        ParseUniOpExpct    => "Unary operation expected",
        ParseBinOpExpct    => "Binary operation expected",

        ParseSelectErr     => "Error in select()",
        ParseProjectErr    => "Error in project()",
        ParseJoinErr       => "Error in join()",
        ParseCopyErr       => "Error in copy()",
        ParseSpanErr       => "Error in span() definition",
        ParseTabExprErr    => "Error in table expression",
        ParseParExprErr    => "Error in parameter expression",
        ParseFParaErr      => "Error in filter expression",
        ParseTParaErr      => "Error in project expression",
        ParseCParaErr      => "Error in copy() parameter",
        ParseBOpExprErr    => "Error in bin operation expression",


        QueNoSuchOp        => "No such operation",
        QueNoSuchObj       => "Can not access object",
        QueNoSuchTyp       => "No such type",
        QueNoSiteInf       => "Info about site missing",
        QueNoCVal          => "CVal of parameter is not provided",
        QueNoTyp           => "Type is not provided in node",
        QueNoAMClass       => "Class is not provided for AM",
        QueCantCreTyp      => "Type can not be created",
        
        QueSlctCompErr     => "Error by computing selectivity",
        QueSelectErr       => "Type error in select()",
        QueProjectErr      => "Type error in project()",
        QueJoinErr         => "Type error in join()",
        QueCopyErr         => "Type error in copy()",
        QueFParaErr        => "Type error in filter parameter",
        QuePExprErr        => "Type error in parameter expression",
        QueTermErr         => "Error in term",
        QueConstErr        => "Error with constant",

        QueAttrNamErr      => "Can not get full attribute name",
        QueCompTypErr      => "Error by finding component type",
        QueMvdCompTypErr   => "Error by finding type of mvd component",
        QueAttsProjErr     => "Error by projecting attributes of type",
        QueCompAttsProjErr => "Error by projecting component type",

        QueBoolOprExpct    => "Both operands of type boolean expected",
        QueNumOprExpct     => "Both operands of type number expected",
        QueTypNotEq        => "Types of operands are not equal",
        QueTypNotCmp       => "Types in set operation not compatible",
        QueFNameErr        => "No such attribute",


        OsmDbNotClosed     => "Database is not closed",
        OsmDbNotOpened     => "Database is not opened",
        OsmObjNotDef       => "Object not defined",
        OsmAtrNotDef       => "Attribute not defined",
        OsmOidExists       => "Oid already defined",
        OsmOidNotExists    => "Oid does not exist",
        OsmOidNotInUse     => "Oid is not in use",
        OsmObjNotCreated   => "Object was not created",
        OsmObjExists       => "Object already exists",
        OsmObjNotExists    => "Object does not exist",
        OsmNoSuchPhsOp     => "Undefined physical operation",
        OsmAliasRdErr      => "Error by converting alias",
        OsmAtrCnvErr       => "Error by coverting attribute",


        RsmDbNameExist     => "Database name already defined",
        RsmRecNotDef       => "Record id does not exist",
        RsmNoSuchPhsOp     => "Undefined physical operation",
        RsmDbSpaceFull     => "No more space in database file",
        RsmRecIdExist      => "Record ident already in use",
        RsmAliasRdErr      => "Error by converting alias",


        OptTyNotDef        => "Type of query node undefined",
        OptObjNotDef       => "Object is not defined",
        OptDupExpr         => "Query expression shuld not appear",


        CosNotAClass       => "Parameter object should be a class",


        EvalTypeErr        => "Error by typechecking parameter expression",
        EvalIllObj         => "Illegal object",
        EvalNoSuchOp       => "No such operation",
        EvalErrNextObj     => "Error in Osm::next().",


        LoadTypNotExist    => "Type is not defined",
        LoadMvTypCnflct    => "Conflict in types of mv elements",
        LoadXMLSynErr      => "Error in XML syntax",
        LoadWrngRef        => "Illegal ref of element type",
        LoadStckEmpty      => "Stack of elements is empty",
        LoadWrngElmNm      => "Illegal element name",
        LoadMssmPar        => "Mismatched parenthesis",
        LoadWrngStr        => "Unexpected string",
        LoadUndfElmAt      => "Element of attribute undefined",
        LoadElmUndf        => "Undefined element",
        LoadTypErr         => "Type error by loading"

    );


    @error = ();
    @msg   = ();
    $state = $state{Ok};
}    






sub log_debug {
    my ($act,$par,$lvl,$idt) = @_;
    my @cinfo = ();

    $idt = 0 if (!defined $idt);
    $par = "" if (!defined $par);

    if (defined $lvl) {
        @cinfo = caller($lvl);
    } else { 
        @cinfo = caller(1);   
    } 


    my @sub = split "::", $cinfo[3];
    my $sub = pop @sub;
    my $pck = pop @sub;
    my $spc = "";
    while ($idt--) { $spc .= " " };
    if (-w $db->{logfile}) {
        open(LOG,">>".$db->{logfile});
    } else {
        open(LOG,">".$db->{logfile}) ||
            die "Can\'t create log file ".$db->{logfile}."\n";
    }
    printf LOG "%-35s %-25s %-10s\n", 
               "$spc$act: $sub()", $par, $pck;
    close LOG;
}



sub log_tell {
    my ($act,$par,$lvl,$idt) = @_;
    my @cinfo = ();

    $idt = 0 if (!defined $idt);
    $par = "" if (!defined $par);

    if (defined $lvl) {
        @cinfo = caller($lvl);
    } else { 
        @cinfo = caller(1);   
    } 


    my @sub = split "::", $cinfo[3];
    my $sub = pop @sub;
    my $pck = pop @sub;
    my $spc = "";
    while ($idt--) { $spc .= " " };
    if (-w $db->{logfile}) {
        open(LOG,">>".$db->{logfile});
    } else {
        open(LOG,">".$db->{logfile}) ||
            die "Can\'t create log file ".$db->{logfile}."\n";
    }
    print LOG "$spc$act: $par - $pck\:\:$sub\n";
    close LOG;
}



sub log_say {
    my ($msg,$idt) = @_;

    $idt = 0 if (!defined $idt);
    $msg = "" if (!defined $msg);
    my @cinfo = caller(1);


    my @sub = split "::", $cinfo[3];
    my $sub = pop @sub;
    my $pck = pop @sub;
    my $spc = "";
    while ($idt--) { $spc .= " " };
    if (-w $db->{logfile}) {
        open(LOG,">>".$db->{logfile});
    } else {
        open(LOG,">".$db->{logfile}) ||
            die "Can\'t create log file ".$db->{logfile}."\n";
    }
    printf LOG "$spc$msg";
    close LOG;
}



sub log_hey {
    my $msg = shift;

    $msg = "" if (!defined $msg);
    my @cinfo = caller(1);


    my @sub = split "::", $cinfo[3];
    my $sub = pop @sub;
    my $pck = pop @sub;
    if (-w $db->{logfile}) {
        open(LOG,">>".$db->{logfile});
    } else {
        open(LOG,">".$db->{logfile}) ||
            die "Can\'t create log file ".$db->{logfile}."\n";
    }
    printf LOG "$guid: $msg";
    close LOG;
}



sub log_ok {
    my $l = shift;
    if ($db->{'logon'}) { 
        return ($l <= $db->{loglvl});
    } else { 
        return $False;
    }
}



sub dbg_ok {
    my $l = shift;
    if ($db->{debug}) { 
        return ($l <= $db->{dbglvl});
    } else { 
        return $False;
    }
}



sub log_fatal {
    return if !&log_ok(0);
    my $par = join ":", @_;
    log_tell "fatal", $par, 3, 0;    
}

sub log_error {
    return if !&log_ok(1);
    my $par = join ":", @_;
    log_tell "error", $par, 3, 0;    
}

sub log_warn {
    return if !&log_ok(2);
    my $par = join ":", @_;
    log_tell "warn", $par, 3, 0;    
}



sub log_say0 {
    return if !&log_ok(0);
    my $msg = shift;
    log_say $msg;
}

sub log_say1 {
    return if !&log_ok(1);
    my $msg = shift;
    log_say $msg;
}

sub log_say2 {
    return if !&log_ok(2);
    my $msg = shift;
    log_say $msg;
}



sub log_hey0 {
    return if !&log_ok(0);
    my $msg = shift;
    log_hey $msg;
}

sub log_hey1 {
    return if !&log_ok(1);
    my $msg = shift;
    log_hey $msg;    
}

sub log_hey2 {
    return if !&log_ok(2);
    my $msg = shift;
    log_hey $msg;
}




sub log_enter0 {
    return if !&dbg_ok(0);
    my $par = join ":", @_;
    log_debug "call", $par, 2, 0;    
}

sub log_enter1 {
    return if !&dbg_ok(1);
    my $par = join ":", @_;
    log_debug "call", $par, 2, 4;    
}

sub log_enter2 {
    return if !&dbg_ok(2);
    my $par = join ":", @_;
    log_debug "call", $par, 2, 8;
}

sub log_exit0 {
    return if !&dbg_ok(0);
    my $par = join ":", @_;
    log_debug "exit", $par, 2, 0;    
}

sub log_exit1 {
    return if !&dbg_ok(1);
    my $par = join ":", @_;
    log_debug "exit", $par, 2, 4;    
}

sub log_exit2 {
    return if !&dbg_ok(2);
    my $par = join ":", @_;
    log_debug "exit", $par, 2, 8;    
}

sub log_info0 {
    return if !&dbg_ok(0);
    my $par = join ":", @_;
    log_debug "info", $par, 2, 0;    
}

sub log_info1 {
    return if !&dbg_ok(1);
    my $par = join ":", @_;
    log_debug "info", $par, 2, 4;    
}

sub log_info2 {
    return if !&dbg_ok(2);
    my $par = join ":", @_;
    log_debug "info", $par, 2, 8;    
}





sub setWarn {
    my ($wrn,$msg) = @_;

    if (!exists $error{$wrn}) {
        log_info0 "No such warning: $wrn\n";
    }

    push @error, ($wrn);
    push @msg, ($msg);
    $state = $state{Warn};
    log_warn "$error{$wrn} $msg" if ($db->{logon});
}



sub setError {
    my ($err,$msg) = @_;

    if (!exists $error{$err}) {
        print "No such error: $err\n";
    }

    push @error, ($err);
    push @msg, ($msg);
    $state = $state{Error};
    log_error "$error{$err} $msg" if ($db->{logon});
}



sub setInternal {
    my ($err,$msg) = @_;

    if (!exists $error{$err}) {
        log_say0 "No such error: $err\n";
    }

    push @error, ($err);
    push @msg, ($msg);
    $state = $state{Fatal};
    log_fatal "$error{$err} $msg" if ($db->{logon});
}



sub error {
    return $#error >= 0;
}



sub getError {
    if ($#error < 0) {
        return undef;
    }    

    my $er = shift @error;
    $state = $state{Ok} if ($#error<0);
    return $er;
}



sub outputErrors {

    if ($#error < 0) {
        return undef;
    }    

    my ($er,$msg);
    while (defined($er = shift @error)) {
        $msg = shift @msg;
        print "$error{$er} $msg\n";
    }
    $state = $state{Ok};
}



sub cleanErrors {

    if ($#error < 0) {
        return undef;
    }    
    @error = ();
    $state = $state{Ok};
}





sub member {
    my ($a,$l) = @_;
    
    my $e;
    foreach $e (@{$l}) {
        if ($e eq $a) {
            return $Ok;
        }
    }
    return $False;
}



sub subset {
    my ($l1,$l2) = @_;
    
    my $e;
    foreach $e (@{$l1}) {
        if (!&member($e,$l2)) {
            return $False;
        }
    }
    return $Ok;
}



sub trim {
    my $str = shift;
    $str =~ s/^\s+//g;
    $str =~ s/\s+$//g;
    return $str;
}



sub nl() { print "\n"; }



sub stop {
    return $Ok;    
}



sub file2str {
    my $fn = shift;
    my $s = undef;

    if (!open QUE, $fn) {
        return undef;
    }
    while (<QUE>) {
        next if (/^--.*/);
        $s .= $_;
    }
    close QUE;
    return $s;
}





sub op_key {
    my $opid = shift;
    my $k;
    for $k (keys %Qios::Parse::token) {
	if ($Qios::Parse::token{$k} == $opid) {
            return $k;
        }
    }
    return $False;
}



sub op_name {
    my $opid = shift;
    my $k;
    for $k (keys %op) {
	if ($op{$k} == $opid) {
            return $k;
        }
    }
    return $False;
}




sub print {
    my ($pk,$n) = @_;

    if ($db->{'outtype'} == 0) { 
        Qios::Glb->print_raw($n);
    } elsif ($db->{'outtype'} == 1) {
        Qios::Glb->print_xml($n);
    }
}



sub print_raw {
    my ($pk,$n) = @_;

    print "\n";
    $n->{'##addr'} = $n;
    if (exists($n->{nodOp})) {
        $n->{nodOpKey} = &op_key($n->{nodOp});
    }
    if (exists($n->{phsOp})) {
        $n->{phsOpKey} = &op_name($n->{phsOp});
    }

    if (exists($n->{$Qios::Osm::idnt}) && !$debug) {
        &print_val_raw('#ident',$n->{$Qios::Osm::idnt});
    }
    if (exists($n->{$Qios::Osm::idnt1}) && !$debug) {
        &print_val_raw('#alias',$n->{$Qios::Osm::idnt1});
    }

    my $k;
    foreach $k (sort keys %{$n}) {
        if ($debug) {
            &print_val_raw($k,$n->{$k});
        } elsif ($k !~ /^[\#\<\>]/)  {
            &print_val_raw($k,$n->{$k});
        }           
    }

    delete($n->{'#addr'});
    if (exists($n->{nodOp})) {
        delete($n->{nodOpKey});
    }
}



sub print_val_raw() {
    my ($n,$v) = @_;

    printf "%-10s : ",$n;

    my $k;
    if (ref($v) eq "HASH") {
        print "\{ ";
        foreach $k (sort keys(%$v)) { 
            print $k.":".$v->{$k}." ";
        } 
        print "\}";

    } elsif (ref($v) eq "ARRAY") {
        print " \[ ";
        foreach $k (@$v) {
            print "$k "; 
	}
        print "\]"; 
 
    } else {
        print "$v";
    }
    print "\n";
}




sub print_xml {
    my ($pk,$n) = @_;

    print "\n";
    $n->{'##addr'} = $n;
    if (exists($n->{nodOp})) {
        $n->{nodOpKey} = &op_key($n->{nodOp});
    }
    if (exists($n->{phsOp})) {
        $n->{phsOpKey} = &op_name($n->{phsOp});
    }

    if (exists($n->{$Qios::Osm::idnt1}) && !$debug) {
        print "\<".$n->{$Qios::Osm::idnt1}."\>\n";
    } elsif (exists($n->{$Qios::Osm::idnt})) {
        print "\<".$n->{$Qios::Osm::idnt}."\>\n";
    }

    if (exists($n->{$Qios::Osm::idnt}) && !$debug) {
        &print_val_xml('ident',$n->{$Qios::Osm::idnt});
    }

    my $k;
    foreach $k (sort keys %{$n}) {
        if ($debug) {
            &print_val_xml($n->obj_atrno($k),$n->{$k});
        } elsif ($k !~ /^[\#\<\>]/)  {
            &print_val_xml($n->obj_atrno($k),$n->{$k});
        }           
    }

    if (exists($n->{$Qios::Osm::idnt1}) && !$debug) {
        print "\<\/".$n->{$Qios::Osm::idnt1}."\>\n";
    } elsif (exists($n->{$Qios::Osm::idnt})) {
        print "\<\/".$n->{$Qios::Osm::idnt}."\>\n";
    }


    delete($n->{'#addr'});
    if (exists($n->{nodOp})) {
        delete($n->{nodOpKey});
    }
}



sub print_val_xml() {
    my ($n,$v) = @_;

    my ($k,$st);
    if (ref($v) eq "HASH") {
        print "\{ ";
        foreach $k (sort keys(%$v)) { 
            print $k.":".$v->{$k}." ";
        } 
        print "\}";

    } elsif (ref($v) eq "ARRAY") {
        print "\<$n\>";
        $st = undef;
        foreach $k (@$v) {
            print " " if ($st);
            print "$k";
            $st = $Ok;
	}
        print "\<\*\/$n\>\n"; 
 
    } else {
      print "\<$n\>$v\<\/$n\>\n";
    }
    return $Ok;
}





sub root_defs {
    my $pk = shift; 

    my $db0 = &Qios::Osm::obj_bless();

    if (defined $ENV{'QIOS_BASE'}) { 
        $db0->{dbdir} = $ENV{'QIOS_BASE'};
    } elsif (defined $ENV{'HOME'}) {
        $db0->{dbdir} = $ENV{'HOME'}."/.qios_base";
    } else {
        $db0->{dbdir} = '.';
    }
    if (!-d $db0->{dbdir}) {
        mkdir $db0->{dbdir};
        print "Qios base directory ".
               $db0->{dbdir}." created.\n";
    }

    $Qios::Glb::config = $db0->{dbdir}."/config";

    $db0->{bufsize} = 100000;                  # vspace buffer size
    $db0->{config}  = $Qios::Glb::config;      # db conf file (for record)
    $db0->{dbname}  = 'base';                  # database name 
    $db0->{dbfile}  = $db0->{dbdir}.           # default db file
                     "/base.db"; 
    $db0->{dbsize}  = 100000;                  # default size of db
    $db0->{debug}   = 0;                       # debug mode off        
    $db0->{dbglvl}  = 1;                       # debug log level 
    $db0->{logon}   = 1;                       # log turned on 
    $db0->{collsvd} = 1;                       # collapse svds
    $db0->{collmvd} = 1;                       # collapse mvds

    $db0->{logfile} = $db0->{dbdir}.           # default log file
                      "/log";             
    $db0->{loglvl}  = 1;                       # message log level  
    $db0->{matres}  = 0;                       # materialise results off
    $db0->{outtype} = 1;                       # output type

    if (-r $Qios::Glb::config) {
        my $db1 = $db;
        $db = $db0;
        &root_conf_load($db0);
        $db = $db1;
    }

    $db0->{bufcnt} = undef;                    # vspace buff count
    $db0->{bufloc} = undef;   
    $db0->{buflst} = undef;
    $db0->{oidfree} = undef;
    $db0->{oidnxt} = undef;
    $db0->{quefn} = '';                        # default query file

    return $db0;
}



sub root_log {
    my $pk = shift; 

    log_say1 "Root object:
    config: $db->{config}
    dbdir: $db->{dbdir}
    dbname: $db->{dbname}
    dbfile: $db->{dbfile}
    dbsize: $db->{dbsize}
    quefn: $db->{quefn}
    bufsize: $db->{bufsize}
    bufcnt: $db->{bufcnt}
    bufloc: $db->{bufloc}
    buflst: $db->{buflst}
    oidfree: $db->{oidfree}
    oidnxt: $db->{oidnxt}
    logon: $db->{logon}
    loglvl: $db->{loglvl}
    logfile: $db->{logfile}
    debug: $db->{debug}
    dbglvl: $db->{dbglvl}
    matres: $db->{matres}
    collsvd: $db->{collsvd}
    collmvd: $db->{collmvd}
    outtype: $db->{outtype}\n";

    return $Ok;
}



sub root_print {
    my $pk = shift; 
    log_hey1 "Root object printed.\n";

    print "dbname: ".$db->{dbname}."\n";
    print "dbfile: ".$db->{dbfile}."\n";
    print "dbsize: ".$db->{dbsize}."\n";
    print "dbdir: ".$db->{dbdir}."\n";
    print "quefn: ".$db->{quefn}."\n-------\n";
    print "bufsize: ".$db->{bufsize}."\n";
    print "bufcnt: ".$db->{bufcnt}."\n";
    print "bufloc: ".$db->{bufloc}."\n";
    print "buflst: ".$db->{buflst}."\n";
    print "oidfree: ".$db->{oidfree}."\n";
    print "oidnxt: ".$db->{oidnxt}."\n-------\n";
    print "logon: ".$db->{logon}."\n";
    print "loglvl: ".$db->{loglvl}."\n";
    print "logfile: ".$db->{logfile}."\n";
    print "debug: ".$db->{debug}."\n";
    print "dbglvl: ".$db->{dbglvl}."\n";
    print "matres: ".$db->{matres}."\n";
    print "collsvd: ".$db->{collsvd}."\n";
    print "collmvd: ".$db->{collmvd}."\n";
    print "outtype: ".$db->{outtype}."\n";

    return $Ok;
}



sub root_conf_save {
    my $db0 = shift;

    if (!open(CF,">".$Qios::Glb::config)) {
        setError GlbErrOpenConf;
        return undef;
    }
    print CF "dbdir     = ".$db0->{dbdir}."\n";
    print CF "dbfile    = ".$db0->{dbfile}."\n";
    print CF "dbname    = ".$db0->{dbname}."\n";
    print CF "dbsize    = ".$db0->{dbsize}."\n";
    print CF "bufsize   = ".$db0->{bufsize}."\n";
    print CF "quefn     = ".$db0->{quefn}."\n";
    print CF "logfile   = ".$db0->{logfile}."\n";
    print CF "loglvl    = ".$db0->{loglvl}."\n";
    print CF "dbglvl    = ".$db0->{dbglvl}."\n";
    print CF "debug     = ".$db0->{debug}."\n";
    print CF "matres    = ".$db0->{matres}."\n";
    print CF "collsvd   = ".$db0->{collsvd}."\n";
    print CF "collmvd   = ".$db0->{collmvd}."\n";
    print CF "outtype   = ".$db0->{outtype}."\n";
   
    close CF;
    return $Ok;
}



sub root_conf_load {
    my $db0 = shift;
    my ($txt,$val);

    if (!open(CF,"<".$Qios::Glb::config)) {
        setError GlbErrOpenConf;
        return undef;
    }
    while (<CF>) {
       ($txt,$val) = split "=";
       chop $val;
       $db0->{dbname}  = &trim($val) if ($txt =~ /^dbname/);
       $db0->{dbdir}   = &trim($val) if ($txt =~ /^dbdir/);
       $db0->{dbfile}  = &trim($val) if ($txt =~ /^dbfile/);
       $db0->{dbsize}  = &trim($val) if ($txt =~ /^dbsize/);
       $db0->{bufsize} = &trim($val) if ($txt =~ /^bufsize/);
       $db0->{quefn}   = &trim($val) if ($txt =~ /^quefn/);
       $db0->{logfile} = &trim($val) if ($txt =~ /^logfile/);
       $db0->{loglvl}  = &trim($val) if ($txt =~ /^loglvl/);
       $db0->{dbglvl}  = &trim($val) if ($txt =~ /^dbglvl/);
       $db0->{debug}   = &trim($val) if ($txt =~ /^debug/);
       $db0->{matres}  = &trim($val) if ($txt =~ /^matres/);
       $db0->{collsvd} = &trim($val) if ($txt =~ /^collsvd/);
       $db0->{collmvd} = &trim($val) if ($txt =~ /^collmvd/);
       $db0->{outtype} = &trim($val) if ($txt =~ /^outtype/);
   }
}


1;





