
=head1 NAME

I<Load.pm> -- load data into Qios database. 

=head1 DESCRIPTION

Module I<Load.pm> includes the functions for loading XML data,
conversion of the DTD schemata into internal database schemata, and
re-engineering of the database schemata from XML data. 

=head1 Schema discovery

The method for the re-engineering of the schema from the data is
implemented. It tends to group the information to construct the
object-relational schema (objects which besides scalar-valued
attributes include also multi-valued attributes ranging over scalars
or objects) rooted around the top-level data elements. The types of
elements are computed by determining the least general type that
covers all instances.

Two parameters by which the creation of the object-relational schemata
can be controlled in schema re-discovery and in DTD load process are
as follows. The boolean parameter I<collsvd> controls the collapse of
the function elements into the subsuming relation. By turning the
option I<collsvd> on, all the single-valued elements (including a
single data entry) are included in the upper-level relational
schema. Similarly, the boolean parameter I<collmvd> controls the
collapse of relations including a single multi-valued attribute
modeling thus sequences of simple elements including a single data
entry. The parameters are accessible thru the database environment
using the command I<set>.

=head1 XML 

I<Qios> includes the method for the translation of the DTD schemata
into the internal database format. Similarly as the procedure for
re-engineering the schema from the XML data, the procedure for the
translation of DTD schemata into internal format constructs
object-relational schemata. User can control the conversion of the
functional and multi-valued path expressions---they can be
either collapsed into single or multi-valued attributes, or retained.
Same as by the procedure for the re-engineering of the object-relational
schemata from the XML data, the conversion of DTD can be controlled 
using system parameters I<collsvd> and I<collmvd>.

The basic rules for the translation of DTD files into
object-relational schemata are as follows. XML attributes are
converted into functional (relational) attributes describing the
relation of the host element. The elements described with more than
one components (elements) and attributes are converted into
relations. The creation of the relations for the attributes including
a single data entry (XML attribute or component element) is controlled
by the parameters I<collsvd> and I<collmvd> described in the previous 
section.

=head1 Data loader

The method for loading XML data into the internal database follows
strictly the structure of the data defined by the database schemata
(ie. the set of class objects) to unravel on-the-fly the structure of
the XML records. The method tends not to use the parameters described
above for the decisions during the loading process. 

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Load;


use Qios::Glb;
use Qios::Osm;
use LWP::Simple;
use XML::Parser;



BEGIN { 

    @estck = ();

    %ldcls = ();

    $resch = new XML::Parser(Handlers => { Start => \&reschStart,
					   End   => \&reschEnd,
					   Char  => \&reschChar
					  });
    $loader = new XML::Parser(Handlers => { Start => \&loadStart,
		 			    End   => \&loadEnd,
					    Char  => \&loadChar
					  });
    $lodtd = new XML::Parser(Handlers => { Element => \&dtd_element,
		 			   Attlist => \&dtd_attlist,
					   Char    => \&dtd_char
					  });
}



sub xml_load_url {
    my ($pk,$url) = @_;
    undef %ldcls;
    log_say1 "Loading XML data from $url.\n";

    my $xml = &get($url);
    $loader->parse($xml);

}



sub xml_load_file {
    my ($pk,$fn) = @_;
    undef %ldcls;
    log_say1 "Loading XML data from file $fn.\n";
   
    open(XF,$fn);
    $loader->parse(*XF);
    close XF;

}



sub xml_reeng_url {
    my ($pk,$url) = @_;
    undef %ldcls;
    log_say1 "Extracting schema from URL $url.\n";
   
    my $xml = &get($url);
    $resch->parse($xml);
}



sub xml_reeng_file {
    my ($pk,$fn) = @_;
    undef %ldcls;
    log_say1 "Extracting schema from file $fn.\n";
   
    open(XF,$fn);
    $resch->parse(*XF);
    close XF;
}



sub dtd_load_url {
    my ($pk,$url) = @_;

    %elem = (); 
    log_say1 "Loading DTD from URL $url.\n";
   
    my $dtd = &get($url);
    $lodtd->parse($dtd);
   
    &schema_clean();
    &schema_convert();

    return $Ok;
}



sub dtd_load_file {
    my ($pk,$fn) = @_;

    %elem = (); 
    log_say1 "Loading DTD from file $fn.\n";
   
    open(XF,$fn);
    $lodtd->parse(*XF);
    close XF;

    &schema_clean();
    &schema_convert();

    return $Ok;
}





sub new {
    my $pk = shift;

    $el = {};
    bless $el;
}



sub delete {
    my $el = shift;

    my $c;
    undef %{$el->{ldComp}};
 
    undef %{$el};
    return $Ok;
}




BEGIN {
    $data_cont = $False;     # denoting the cont of char-data
    $ld_natr = "comp";       # nested attribute string
    $ld_natr_c = 1;          # nested attribute counter
    $ld_val = "val";         # denoting single attr name
    $ld_val_c = 1;           # counter of ld_val
    $ldTyp = "\&\&kind";     # kind of type spec var
    $ldCat = "\&\&cat";      # marker for single attr elems
    $ldAty = "\&\&aty";      # marker for single attr elems
    $ldMat = "\&\&mat";      # materialize relation
    %elem = ();              # elements of curr schema
    @mdl = ();               # cleaned element model 
}



sub dtd_element {
    my ($xp,$en,$mo) = @_;

    $mo =~ s/([\|\,\(\)\*\?\+])/ $1 /g;        
    $mo =~ s/\s+/ /g;   # single blank

    @mdl = split / /, &trim($mo);

    $parc = 0; 
    &model_parse($en,\@mdl);
    return undef if &error();

    if ($parc != 0) {
        setError LoadMssmPar;
        return undef;
    }  
    return $Ok;
} 



sub model_parse {
    my ($en,$md) = @_;
    my ($el0,$el1);
    
    $el0 = shift @{$md};
    if ($el0 eq '(') {
        $parc++;
        $el0 = shift @{$md};
    }
    
    my $em = Qios::Load->new();
    $em->{$ldTyp} = "AND";
    my $ld_natr_c = 1;                # reset cntr
    L1: while () {

        if ($el0 eq '(') {
            unshift @{$md}, ($el0);

            $el0 = "$ld_natr".$ld_natr_c++;
            $em->{$el0} = &model_parse($el0,$md);
        } elsif ($el0 =~ /\w*/) {

            $em->{$el0} = $el0;
        } else {
            setError LoadWrngElmNm;
            return undef;
        }
 
        $el1 = shift @{$md};
        $em->{"\#\#$el0"} = "f";
        if (($el1 eq '*') || 
            ($el1 eq '+')) {
            $em->{"\#\#$el0"} = "m";
            $el1 = shift @{$md};
        } elsif ($el1 eq '?') {
            $el1 = shift @{$md};
        }

        if (!defined($el1)) {
            last L1;
        } elsif ($el1 eq ',') { 
            $el0 = shift @{$md};
            next L1;
        } elsif ($el1 eq '|') { 
            $el0 = shift @{$md};
            $em->{$ldTyp} = "OR";
            next L1;
        } elsif ($el1 eq ')') {
            $parc--;
            last L1;
        } else {
            setError LoadWrngStr;
            return undef;
        }
    } 

    if (defined $en) {
        $elem{$en} = $em;
    }
    return $em;
}



sub elem_print {
    my ($en,$em) = @_;

    if (ref($em) ne 'Qios::Load') {
        print "$en\:\:$em\n";
        return $Ok;
    }

    print "$en\:\:(";
    foreach $k (keys %{$em}) {
        if (ref($em->{$k}) eq 'Qios::Load') {        # test this!
            print " ";
            &elem_print($k,$em->{$k});
        } else {
            print " $k\:\:".$em->{$k};
        }
    }
    print " )\n";
    return $Ok;
}



sub schema_print {
    print "----------\n";
    print "Schema.\n";

    foreach $k (sort keys %elem) {
        &elem_print($k,$elem{$k});
    }
    return $Ok;
}



sub elem_clean {
    my ($en,$em) = @_;

    sub get_pair () {
        my $ats = shift;
        foreach $k (keys %{$ats}) {
            if ($k =~ /^\#\#.*/) {
                $a2 = $k;
            } elsif ($k ne $ldTyp) {
                $a1 = $k;
            } 
        }
        return ($a1,$a2);
    }

    foreach $k0 (keys %{$em}) {
        if ($k0 eq 'EMPTY') {
            delete $em->{$k0};
            delete $em->{"\#\#$k0"};
        }
        if (ref($em->{$k0}) eq 'Qios::Load') {
            &elem_clean($k0,$em->{$k0});
            delete $em->{"\#\#$k0"};
        }
    }

    if ((scalar keys %{$em}) <= 3) {

        ($k1,$k2) = &get_pair($em);
        my $ty = $em->{$k2};

        if (((($ty eq "a") || ($ty eq "f")) && $db->{collsvd}) ||
            (($ty eq "m") && $db->{collmvd})) {
               
            if (($em->{$k1} eq '#PCDATA') ||
                ($em->{$k1} eq 'CDATA') || 
                ($em->{$k1} eq 'IDREF') ||
                ($em->{$k1} eq 'IDREFS')) {
                $elem{$en}->{$ldCat} = 'string';
            } else {
                $elem{$en}->{$ldCat} = $em->{$k1};
            }
            $elem{$en}->{$ldAty} = $ty;
        }
    }
    return $Ok;
}



sub schema_clean {

    while (($k,$v) = each %elem) {
        &elem_clean($k,$v);
    }
    return $Ok;
}


sub elem_creo {
    my $elnm = shift;
        
    if (!defined($elem{$elnm})) {
        setError LoadElmUndf;
        return undef;
    }   
    if (&Qios::Osm::obj_exists($elnm)) {
        return &Qios::Osm::obj_fetch($elnm); 
    } else {    
        return Qios::Osm->obj_new([object],$Qios::Osm::iscl,$elnm); 
    }
}


sub elem_fetch {
    my ($elnm,$elty) = @_;

    if ($elnm eq 'string') {
        return ($elnm,"f");
    };

    if (!defined($elem{$elnm})) {
        setError LoadElmUndf;
        return undef;
    }   

    my ($elnm0,$elty0) = ($elnm,$elty);
    my $elct1 = $elem{$elnm}->{$ldCat};
    my $elty1 = $elem{$elnm}->{$ldAty};
    if ($elct1 && (($elty1 eq "a") || ($elty1 eq "f")) && $db->{collsvd}) {
        ($elnm0,$elty0) = &elem_fetch($elct1,$elty);

    } elsif ($elct1 && ($elty1 eq "m") && $db->{collmvd}) {
        if (($elty eq "f") || ($elty eq "a")) { 
            ($elnm0,$elty0) = &elem_fetch($elct1,"m");
            $elty0 = "m";
        } else {
            ($elnm0,$elty0) = ($elnm,"m");
        }
    }

    if ($elnm0 eq 'string') {
        return ($elnm0,$elty0);
    };

    my $elob0;
    if (&Qios::Osm::obj_exists($elnm0)) {
        $elob0 = &Qios::Osm::obj_fetch($elnm0)->{$Qios::Osm::idnt}; 
        return ($elob0,$elty0)
    } else {    
        $elob0 = Qios::Osm->obj_new([object],$Qios::Osm::iscl,$elnm0)
                          ->{$Qios::Osm::idnt};  
        return ($elob0,$elty0);
    }
}



sub elem_convert {
    my ($en,$em) = @_;
    my ($is_sc);
    $ld_val_c = 1;


    if ($em->{$ldCat}) {
        return $Ok;
    }

    $ty = &elem_creo($en);

    L2: while (($k,$v) = each %{$em}) {
        if ($k =~ /^\&\&.*/) { 
            next L2;

        } elsif ($k !~ /^\#\#.*/) {

            if ($k eq '#PCDATA') {
                $ty->attr_set("$ld_val".$ld_val_c++,'string');
                next L2;
            }

            if ($em->{"\#\#$k"} eq "a") {
                $ty->attr_set($k,'string');
                next L2;
            }
           
            if ($em->{"\#\#$k"} eq "f") {

                my ($elob,$elty) = &elem_fetch($em->{$k},"f");
                if (($elty eq "f") || ($elty eq "a")) {
                    $ty->attr_set($k,$elob);
                } else {
                    $ty->attr_setmv($k,0,$elob);
                }
                next L2;
            }

            if ($em->{"\#\#$k"} eq "m") {

                my ($elob,$elty) = &elem_fetch($em->{$k},"m");
                if (($elty eq "a") || ($elty eq "f")) {
                    $ty->attr_setmv($k,0,$elob);
                } else {
                    $ty->attr_setmv($k,0,$elob);

                    $elem{$em->{$k}}->{$ldMat} = $Ok;
                }
                next L2;
            }
	}
    }
    return $Ok;
}



sub schema_convert {

    while (($k,$v) = each %elem) {
        &elem_convert($k,$v);
    }
}



sub dtd_attlist {
    my ($xp,$en,$an,$ty,$df,$fx) = @_;

    if (!exists($elem{$en})) {
        setError LoadUndfElmAt;
        return undef;
    }
    $elem{$en}->{$an} = $ty;
    $elem{$en}->{"\#\#$an"} = "a";

    return $Ok;
} 



sub dtd_char {
    my ($xp,$data) = @_;
    return $Ok;
}





sub reschStart {
    my ($xp,$en,@as) = @_;
    
    $data_cont = $False;

    my $el = Qios::Load->new();
    $el->{ldElem} = $en;
    $el->{ldComp} = undef;

    my ($at,$vl);
    while ($#as != -1) {
        $at = shift @as;
        $vl = shift @as;
        push @{$el->{ldComp}->{$at}}, ($vl);      # at name (!)
    }
   
    push @estck, ($el);
    return $Ok;
} 



sub reschEnd {
    my $xp = shift;
    my $en = shift;

    $data_cont = $False;

    my $el = pop @estck;
    if ($en ne $el->{ldElem}) {
        setError LoadXMLSynErr;
        return undef;
    }

    my ($ty,$ob); 
    my @compKeys = keys %{$el->{ldComp}};
    my $oneComp = ($#compKeys <= 0);
    my $key = $compKeys[0];
    my $singleVal = $Ok;

    if (exists $el->{ldComp}->{$key}) {
        $singleVal = ($#{$el->{ldComp}->{$key}} <= 0);
    }
    my $lastEl = $estck[$#estck];


    if ($oneComp && ($singleVal || $db->{collmvd})) {

        if ($singleVal) {
            if (!exists $el->{ldComp}->{$key}) {
                push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                     (undef);
	    } else {
                push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                     ($el->{ldComp}->{$key}->[0]);
	    }
        } else {
            $lastEl->{ldComp}->{$el->{ldElem}} = $el->{ldComp}->{$key};
        }
        return $Ok;
    }

    if (!$oneComp || ($oneComp && !$singleVal && !$db->{collmvd})) { 

        if (&Qios::Osm::obj_exists($el->{ldElem})) {
            $ty = &Qios::Osm::obj_fetch($el->{ldElem});
        } else {
            $ty = Qios::Osm->obj_new([object],$Qios::Osm::iscl,$el->{ldElem});
        }

        $ldcls{$el->{ldElem}} = $Ok;

        my ($k,$cm,$mvty);
        foreach $k (keys %{$el->{ldComp}}) {
            if ($#{$el->{ldComp}->{$k}} <= 0) {
                $cm = $el->{ldComp}->{$k}->[0];
                $mvty = $False;
            } else {
                $cm = $el->{ldComp}->{$k};
                $mvty = $Ok;
            }
    
            my $dty = &Qios::Que::typ_data($cm);

            if ($k eq 'main_city_list') {
                $dtys = $dty;
                if (ref($cm) eq 'ARRAY') {
                    $cms = '['.join("-", @{$cm}).']';
                }
                if (ref($dty) eq 'ARRAY') {
                    $dtys = '['.join("-", @{$dty}).']';
                }
                print "***mcl: $k--$mvty--$dtys\n";
            }

	    if (!($ty->attr_exists($k))) {
                $ty->attr_set($k,$dty);

            } else {
                if ($ty->attr_mvd($k) && $mvty) {
           
                    if (($ty->attr_getmv($k,0)) ne $dty->[0]) { 
                        setError LoadMvTypCnflct;
                        return undef;
		    }

                } elsif ($ty->attr_mvd($k) && !$mvty) {

                    if (($ty->attr_getmv($k,0)) ne $dty) { 
                        setError LoadMvTypCnflct;
                        return undef;
		    }
        
                } else {

                    if (defined $dty) {
                        if (!&Qios::Rsm::oid_is_sysgen(
			     $ty->attr_get($k))) {
                             $ty->attr_set($k,$dty);
                        }
                    }
                }
            }
        }

        if ($#estck > -1) {
            push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                 ($ty->{$Qios::Osm::idnt});
        }
    }

    $el->delete();
    return $Ok;
} 



sub reschChar {
    my ($xp,$data) = @_;


    return if (&trim($data) eq '');

    if ($#estck == -1) {
        setInternal LoadStckEmpty;
        return undef;
    }

    if ($data_cont) {
        $estck[$#estck]->{ldComp}->{cdata}->[0] .= &trim($data);
    } else {
        push @{$estck[$#estck]->{ldComp}->{cdata}}, 
             (&trim($data));
    }

    $data_cont = $Ok;
    return $Ok;
}




sub loadStart {
    my ($xp,$en,@as) = @_;
    
    $data_cont = $False;

    my $el = Qios::Load->new();
    $el->{ldElem} = $en;
    $el->{ldComp} = undef;
    $el->{ldVcnt} = 1;

    my ($at,$vl);
    while ($#as != -1) {
        $at = shift @as;
        $vl = shift @as;
        push @{$el->{ldComp}->{$at}}, ($vl);      # at name (!)
    }
   
    push @estck, ($el);
    return $Ok;
} 



sub loadEnd {
    my $xp = shift;
    my $en = shift;

    $data_cont = $False;

    my $el = pop @estck;
    if ($en ne $el->{ldElem}) {
        setError LoadXMLSynErr;
        return undef;
    }

    my ($ty,$ob); 
    my @compKeys = keys %{$el->{ldComp}};
    my $oneComp = ($#compKeys <= 0);
    my $key = shift @compKeys;
    my $singleVal = $Ok;
    if (exists $el->{ldComp}->{$key}) {
        $singleVal = ($#{$el->{ldComp}->{$key}} <= 0);
    }
    my $lastEl = $estck[$#estck];
    

    

    my ($ty,$lastTy);
    if ($oneComp && ($#estck >= 0)) {

        if (!&Qios::Osm::obj_exists($el->{ldElem})) {

            if (&Qios::Osm::obj_exists($lastEl->{ldElem})) {
                $lastTy = &Qios::Osm::obj_fetch($lastEl->{ldElem});
                if (&error()) {
                    setError LoadTypNotExist;
                    return undef;
                }
            
                if ($lastTy->attr_svd($el->{ldElem})) {       # && $singleVal)
                    push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                         @{$el->{ldComp}->{$key}};
                    return $Ok;
                }

                if ($lastTy->attr_mvd($el->{ldElem})) {      # && !$singleVal) 
                    push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                         @{$el->{ldComp}->{$key}};
                    return $Ok;
                }
            }

            push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                 @{$el->{ldComp}->{$key}};
            return $Ok;
        }
    }


    $ty = &Qios::Osm::obj_fetch($el->{ldElem});
    $ob = Qios::Osm->obj_new([$el->{ldElem}],$Qios::Osm::isgr);
    if (&error()) {
        setError LoadTypErr;
        return undef;
    }

    my ($k,$cm);
    foreach $k (keys %{$el->{ldComp}}) {


        if ($ty->attr_svd($k)) {            # ($#{$el->{ldComp}->{$k}} == 0)) 
            $cm = $el->{ldComp}->{$k}->[0];

            $ob->attr_set($k,$cm);
        }

        if ($ty->attr_mvd($k)) {
            $cm = $el->{ldComp}->{$k};

            $ob->attr_set($k,$cm);
        }

    }

    if ($#estck >= 0) {
        push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
             ($ob->{$Qios::Osm::idnt});
    }

    $el->delete();
    return $Ok;
} 



sub loadChar {
    my ($xp,$data) = @_;


    return if (&trim($data) eq '');

    if ($#estck == -1) {
        setInternal LoadStckEmpty;
        return undef;
    }

    my $valsym;
    if ($data_cont) {
        $estck[$#estck]->{ldVcnt}--;
        $valsym = "val".$estck[$#estck]->{ldVcnt};
        $estck[$#estck]->{ldComp}->{$valsym}->[0] .= &trim($data);
    } else {
        $valsym = "val".$estck[$#estck]->{ldVcnt};
        push @{$estck[$#estck]->{ldComp}->{$valsym}}, 
             (&trim($data));
    }
    $estck[$#estck]->{ldVcnt}++;
    
    $data_cont = $Ok;
    return $Ok;
}





sub hndlStart {
    my ($xp,$en,@as) = @_;
    
    my $el = Qios::Load->new();
    $el->{ldElem} = $en;
    $el->{ldComp} = undef;

    my ($at,$vl);
    while ($#as != -1) {
        $at = shift @as;
        $vl = shift @as;
        push @{$el->{ldComp}->{$at}}, ($vl);      # at name (!)
    }
   
    push @estck, ($el);
    return $Ok;
} 



sub hndlEnd {
    my $xp = shift;
    my $en = shift;

    my $el = pop @estck;
    if ($en ne $el->{ldElem}) {
        setInternal LoadStckEmpty;
        return undef;
    }
    my $isObject = ((scalar keys %{$el->{ldComp}}) > 1);

    my ($ty,$ob); 
    if ($isObject) {

        if (!Qios::Osm::obj_exists($el->{ldElem})) {
            $ty = Qios::Osm->obj_new([object],$Qios::Osm::iscl,$el->{ldElem});
        } else {
            $ty = &Qios::Osm::obj_fetch($el->{ldElem});
        }

        $ldcls{$el->{ldElem}} = $Ok;

        $ob = Qios::Osm->obj_new([$el->{ldElem}],$Qios::Osm::isgr);
    } else {
        $ob = undef;
    }

    my ($k,$cm);
    foreach $k (keys %{$el->{ldComp}}) {

        if ($#{$el->{ldComp}->{$k}} == 0) {
            $cm = $el->{ldComp}->{$k}->[0];
        } else {
            $cm = $el->{ldComp}->{$k};
        }

        if ($isObject) {

	    if (!$ty->attr_exists($k)) {
                $ty->attr_set($k,&Qios::Que::typ_data($cm));
            }
            $ob->attr_set($k,$cm);
        } else {

            $ob = $cm;          
        }
    }

    if ($#estck > -1) {
        my $lastEl = $estck[$#estck];
        if ($isObject) {

            push @{$lastEl->{ldComp}->{$el->{ldElem}}}, 
                 ($ob->{$Qios::Osm::idnt});
        } elsif (ref($ob) eq "LIST") {

            $lastEl->{ldComp}->{$el->{ldElem}} = $ob;
        } else {

            push @{$lastEl->{ldComp}->{$el->{ldElem}}}, ($ob);
        }            
    }

    $el->delete();
    return $Ok;
} 



sub hndlChar {
    my ($xp,$data) = @_;

    return if (&trim($data) eq '');

    if ($#estck == -1) {
        setInternal LoadStckEmpty;
        return undef;
    }

    push @{$estck[$#estck]->{ldComp}->{cdata}}, 
         (&trim($data));
    return $Ok;
}


1;




