


=head1 NAME

I<Optor.pm> -- query optimization.

=head1 DESCRIPTION

This package includes the implementation of data structures and
optimization algorithms for the logical optimization of the query
expressions.

The core of the module is the data structure I<mesh> for the
representation of sets of queries. Common sub-expressions of queries
are shared ie. each query is represented in I<mesh> only once. Queries
are organized into equivalence classes.

The module is the platform for the study of query optimization
algorithms appropriate for the data environment of Qios. The module
I<Rule.pm> is used to perform logical transformations of the query
trees and the module I<Cost.pm> is applied for the computation of the
cost of the newly generated query trees. The module currently includes
the implementation of sub-optimal optimization algorithm based on
dynamic programming.

=head1 Mesh

I<mesh> is a data structure for storing query trees. The basic
structure of I<mesh> is directed acyclic graph (dag). The query trees
have to share the common parts hence there is only one representation
of a query expression in I<mesh>. Further, the query trees are
organized into equivalence classes. I<mesh> has three entry points.

First, queries in I<mesh> can be accessed using the unique 
identifiers of queries which are created when query is entered in
I<mesh>. The mapping is realized between query trees (roots of) and
the entries in I<mesh> holding the queries. 

Second, queries in I<mesh> can be accessed thru the equivalence
classes which again have unique identification generated on the
creation of the equivalence class. The inverse relationship from
the query trees (roots of) to the equivalence classes is also defined.

Third, queries can be accessed using the normalized query expressions
(character strings) as the hash index targeting roots of query trees.

In spite of various access methods, queries are in I<mesh> defined
free of the cover data structures as they are created using module
I<Que.pm>. Each query node includes solely a reference to the
equivalence class to which it belongs.

=head1 Optimization algorithms 

The algorithm implemented in I<Optor.pm> is based on dynamic
programming. The optimization is performed bottom-up: the leaves of
the query tree are optimized first progressing then upwards toward
the root of the query tree. For each query tree all possible logical
transformations of the root of query tree are generated and optimized,
recursively. At this point of the algorithm, the search space can be
reduced by selecting only the promising queries based on cost
estimation. 

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Optor;


use Qios::Glb;
use Qios::Osm;
use Qios::Que;
use Qios::Rule;
use Qios::Cost;



BEGIN {

    @mesh = ();        
                       
    @ques = ();        
                       
    %qexp = {};        

    @ocst = ();        
  
    $method = "EXHTD";
}






sub new() {
    my $cl = shift;
    $e = {};
    bless $e;

    $e->{eqcCost} = undef;
    $e->{eqcBest} = undef;
    $e->{eqcCq} = 0;

    push @mesh, $e;
    $e->{eqcMeid} = $#mesh;    

    return $e;
}



sub delete() {
    my $e = shift;

    $mesh[$e->{eqcMeid}] = undef;

    undef @{$e->{eqcElem}};
    undef %{$e};
}



sub add() {
    my ($e,$n) = @_;
  
    push @ques, ($n);
    my $quid = $#ques;

    $n->{nodQuid} = $quid;
    $n->{nodEquc} = $e;
    my $c = &cost($n);

    push @{$e->{eqcElem}}, ($quid);
    if (!defined($e->{eqcCost})) {
        $e->{eqcCost} = $c;
        $e->{eqcBest} = $quid;
    } elsif ($c < $e->{eqcCost}) {
        $e->{eqcCost} = $c;
        $e->{eqcBest} = $quid;
    }

    $qexp{$n->{nodExpr}} = $n;
    return $Ok;
}



sub clean {
    my $e = shift;

    undef @{$e->{eqcElem}};
    $e->{eqcElem} = undef;
    $e->{eqcCost} = undef;
    $e->{eqcBest} = undef;
    $e->{eqcCq} = 0;
}






sub mesh_clean() {
    my $pk = shift;
    log_say1 "Cleaning Mesh.\n";


    my $n;
    L1: foreach $n (@ques) {

        if (&Qios::Osm::obj_exists($n->{nodType})) {
            if (! &obj_fetch($n->{nodType})->cl_empty()) {

                if (defined $n->{nodDefty}) {
                    $n->delete(); 
                    next L1;
                }
                &obj_fetch($n->{nodType})->cl_delete_ext();
            }
            &obj_fetch($n->{nodType})->obj_delete();

        } else {
            setError OptObjNotDef, "$n->{nodType}:$n->{nodExpr}";
        }
        $n->delete(); 
    }

    my $e;
    foreach $e (@mesh) {
        undef @{$e->{eqcElem}};
        $e->delete();
    }

    undef @ocst;
    undef @mesh;
    undef %qexp;
    undef @ques;
   
    log_exit1;
    return $Ok;
}



sub mesh_print {
    my $pk = shift;
    log_hey1 "Printing all equivalence classes and queries from Mesh.\n";

    my ($e,$q,$sz,$cr);
    foreach $e (@mesh) {
        Qios::Glb->print_raw($e);

        print "nodExpr    : ".$ques[$e->{eqcBest}]->{nodExpr}."\n";
        $sz = $#{$e->{eqcElem}} + 1;
        print "eqcNumQue  : ".$sz."\n";

        foreach $q (@{$e->{eqcElem}}) { 
            Qios::Glb->print_raw($ques[$q]);
        }

        print "\n<next>"; 
        $cr = <STDIN>;
    }
}



sub mesh_print_eqc {
    my $pk = shift;
    log_hey1 "Printing equivalence classes.\n";

    my ($e,$q,$sz,$cr);
    foreach $e (@mesh) {
        Qios::Glb->print_raw($e);

        print "eqcOExpr   : ".$ques[$e->{eqcBest}]->{nodExpr}."\n";
        $sz = $#{$e->{eqcElem}} + 1;
        print "eqcNumQue  : ".$sz."\n";
    }
    return $Ok;
}



sub mesh_print_qex {
    my $pk = shift;
    log_hey1 "Printing queries thru hash of query expressions.\n";

    my ($k,$v);
    my @qexs = ();
    while (($k,$v) = each %qexp) {
        push @qexs, ($v->{nodEquc}->{eqcMeid}."\t".
                     $v->{nodEquc}->{eqcBest}."\t".
                     $v->{nodQuid}."\t".
                     $k."\n");
    }
  
    foreach $v (sort {$a <=> $b} @qexs) {
        print $v;
    }
    print "\n";
    return $Ok;
}



sub mesh_print_que {
    my $pk = shift;
    log_hey1 "Printing all queries from Mesh.\n";

    my ($i,$v);
    for ($i=0; $i<=$#ques; $i++) {
        print $ques[$i]->{nodQuid}."\t".
              $ques[$i]->{nodEquc}->{eqcBest}."\t".
              $ques[$i]->{nodEquc}->{eqcMeid}."\t".
              $ques[$i]->{nodExpr}."\n";
    }
    print "\n";
    return $Ok;
}






sub optimize {
    my ($pk,$n) = @_;
    log_say1 "Optimizing.\n";

    &mesh_clean();

    my $n1;
    C1: {
        if ($method eq "EXHTD") {   
            return $ques[&optimize_exhtd($n)];
        }
    }
    return undef;
}



sub link_node {
    my ($n,$s1,$s2) = @_;

    $n->{nodOuter} = $s1;
    
    my $p = $n->{nodParam};
    if ($n->{nodOp} == $op{select} ||
        $n->{nodOp} == $op{join}) {
        $p->{nodVRmap}->{$p->{nodPnam}->[0]} = $s1;
    }

    if (defined $s2) {
        $n->{nodInner} = $s2;
        if (($n->{nodOp} == $op{select}) ||
            ($n->{nodOp} == $op{join})) {
            $p->{nodVRmap}->{$p->{nodPnam}->[1]} = $s2;
        }
    }

    my $v;
    foreach $v (@{$p->{nodPvars}}) {
        $v->{nodOuter} = $p->{nodVRmap}->{$v->{nodName}};
    }

    $n->clean();
    my $ty = $n->typ_query();
    return $ty;
}



sub optimize_exhtd {
    my $n = shift;
    
    if (exists $qexp{$n->{nodExpr}}) {
        setInternal OptDupExpr;
        return undef;
    }
 
    my $e = Qios::Optor->new();
    $e->add($n);

    if ($n->{nodOp} == $op{ins} || 
        $n->{nodOp} == $op{ext}) {
        $e->{eqcOptim} = $Ok;
        return $n->{nodQuid};
    }

    my ($nq,$cq,$qid);
    L1: while () {

        if (defined ($qid = $e->{eqcElem}->[$e->{eqcCq}++])) {
            $cq = $ques[$qid];
        } else {
            last L1;
        }

        if ($qexp{$cq->{nodExpr}} != $cq) { 
            next L1;
        }

        my @qcmp = ();  # curr comps
        my @qres = ();  # opt comps
        my $ncmp = 0;   # num of comps
        $qcmp[0] = $cq->{nodOuter};
        if (defined $cq->{nodInner}) {
            $qcmp[1] = $cq->{nodInner};
            $ncmp = 1;
        }

        my ($qcng,$c) = ($False,0);
        for ($c=0; $c<=$ncmp; $c++) {
            if (!exists($qexp{$qcmp[$c]->{nodExpr}})) {
                $qres[$c] = $ques[&optimize_exhtd($qcmp[$c])];
            } else {
                if ($qcmp[$c]->{nodEquc}->{eqcOptim}) {
                    $qres[$c] = $ques[$qcmp[$c]->{nodEquc}->{eqcBest}];
                } else {
                    $qexp{$qcmp[$c]->{nodExpr}}->{nodEquc}->add($qcmp[$c]);
                    $qres[$c] = $qcmp[$c];
                }
            }
            $qcng = $qcng || ($qcmp[$c] != $qres[$c]);
	}

        if ($qcng) { 
            $nq = $cq->duplicate();
            $nq->{nodParam} = $cq->{nodParam}->duplicate();
            &link_node( $nq,$qres[0],$qres[1] );

            $cq->{nodParam}->{nodBind} = $False;

            if (!exists $qexp{$nq->{nodExpr}}) {
                $e->add($nq);
	    } else {
                $nq->{nodParam}->delete();
                &obj_fetch($nq->{nodType})->obj_delete();
                $nq->delete();
            }   
        }

        L2: foreach $r (@Qios::Rule::rule) {

            if ($r->match($cq)) {

                if (exists $qexp{$r->{rulOutpat}->{nodExpr}}) {
                    $r->clean(1);
                    next L2;
                }

                $nq = $r->apply($cq);
                $e->add($nq);
                $r->clean();     # apply cleans rule

            } else {
                $r->clean(1);
            }
        }
    }

    $e->{eqcOptim} = $Ok;
    return $e->{eqcBest};
}


1;




