
=head1 NAME

Osm.pm -- persistent object storage manager.

=head1 DESCRIPTION

Persistent objects are Perl objects that are tied to the database via
the persistent storage manager. Each object has an identifier which
is unique within the database file. Object identifiers are system
generated. External identifiers which are unique within the datafile
can be assigned to objects when they are created.

I<Osm.pm> implements object cache that store recently used
objects. Cache is realized using LRU (least recently used) strategy to
choose the objects to be removed from cache. The size of object cache
can be set as the system parameter via the configuration record of a
datafile in run-time.

Objects can be accessed either directly using object identifiers, or
thru scan operator (iterator). When accessed using object identifiers,
objects are manipulated thru the in-memory pointers to Perl objects in
cache. Routines for the manipulation of objects mark the object
``changed'' when components are altered, so they don't need to be
treated thereafter. However, the user is responsible that object is
accessed each time before it is used.

=head1 OBJECT REPRESENTATION 

Every object has an identifier denoted as object identifier or oid. 
Objects can have functional or multi-valued attributes. The term
``attribute'' corresponds to what is also called object component,
data member, data field, etc. The range of the functional attribute
can be an atomic value: number, string or object (oid). Multi-valued
attribute is an array of atomic values. The data model of Qios is
formally presented in [].

From a programmer's perspective, an object is a reference to the
associative array (blessed;) and an attribute is a hash entry which
can be either a Perl scalar (typed) or a reference to an array of
scalars.

Storage manager differentiates between two kinds of objects: class
objects and individual objects (instances). Each class object is 
associated to the set of its sub-classes and its instances. The storage
manager implementation is based on the hierarchical and network data
models: class, its instances and sub-classes are linked in a ring
realized in I<Osm.pm> by a double linked list.

=head1 Types

Types are viewed as data structures. Qios distinguishes between two
sorts of types: atomic type and tuple-structured type. Atomic types
are: I<number>, I<string> and I<object>. The type I<object> is a
generic term that stands for object identifiers referring to
user-defined types (actually I<object> is oid of the root class
object; see presentation of the Qios data model in [] for details).

Tuple-structured types have components or attributes which are either
functional or multi-valued (implemented as refs to arrays). The range
of functional and multi-valued attributes can only be atomic types.

As suggested by the above description, types (ie. data structures)
together with the corresponding object identifiers form I<class
objects>. The structure of the Qios database can then be viewed
uniformly as partially ordered set of objects: class objects and
instances (individual objects) interrelated by the inheritance (isa)
relationship.

=head1 OBJECT MANIPULATION

The basic functions on objects are used to create, delete, and
manipulate objects. The functions are: I<obj_new>, I<obj_create>,
I<obj_fetch>, I<obj_delete>, I<obj_dispose>, I<obj_set>, and
I<obj_get>.

Object state can be read and altered using the procedures for the
manipulation of object attributes. The procedures for reading and
altering functional attributes are: I<attr_set> and I<attr_get>.  The
procedures for manipulation of multi-valued attributes are work same
as Perl procedures for list manipulation; they are: I<attr_setmv>,
I<attr_getmv>, I<attr_push>, I<attr_unshift>, I<attr_pop>,
I<attr_shift> and I<attr_del>.

The procedures used for the manipulation of object values which are
used during query processing are: I<obj_copy>, I<obj_project>,
I<obj_diff>, I<obj_dupl> and I<obj_join>. The first three operations
preserve the identity of objects while the last two generate new
objects.

The iterative access to the instances and sub-classes of class object
is provided by procedures: I<scan>, I<next> and I<close>. The names 
of the attributes of a given object can be accessed using procedures: 
I<scan_attr>, I<next_attr> and I<close_attr>.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Osm;

 

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( &obj_fetch $iscl $isgr $idnt $idnt1 &attr_sys &attr_mvd );

use Qios::Glb;
use Qios::Rsm;
use Qios::Load;



BEGIN {

    $iscl  = 1;                 # is-a-class variable
    $isgr  = 0;                 # is-a-ground-object variable
    $attrGlue = '#';            # attribute name glue

}




BEGIN {

    $root = undef;            # root of obj hierarch
    $iter = undef;            # iterator root obj
    $cntx = undef;            # current context

    $kind   = '#kind';        # kind of object (class=1,ground=0)
    $idnt   = '#idnt';        # basic identifier of ob
    $idnt1  = '#idnt1';       # second identifier of ob
    $name   = '#name';        # ob name (not in use!)
    $attc   = '#attc';        # pointer to parent of attribute object
    $supc   = '#supc';        # classes of object (list of idents)  
    $site   = '#site';        # site of object
    $recnm  = '#recnm';       # statistics; num of recs
    $recsz  = '#recsz';       # statistics; size of rec
    $card   = '#carda';       # statistics; cardinality of attr

    $stat   = '##stat';       # class statistics
    $indx   = '##indx';       # ?
    $chng   = '##chng';       # object changed in vspace
    $bufnxt = '##bufnxt';     # next in vspace
    $bufpre = '##bufpre';     # prev in vspace


}





sub obj_new {
    my ($pk,$csp,$cg,$id) = @_;
    log_enter0 $csp->[0].":$id";

    my $ob = &buff_newframe();
    
    $ob->{$kind} = $cg;
    my $oid = &Qios::Rsm::oid_new();
    $ob->{$idnt} = $oid;
    $db->{vspace}->{$oid} = $ob;

    if (defined $id) {
        &obj_alias($oid,$id);
    } 

    if ($cg) {
        $ob->{">$oid"} = $oid;
        $ob->{"<$oid"} = $oid;
    }
    
    &Qios::Rsm::subo_assoc($ob,$csp);

    $ob->{$chng} = $Ok;

    return $ob;
}



sub obj_create {
    my ($pk,$csp,$cg,$vp,$id) = @_;
    log_enter1 "$csp->[0]:$id";

    my $ob = &buff_newframe();

    $ob->{$kind} = $cg;
    my $oid = &Qios::Rsm::oid_new();
    $ob->{$idnt} = $oid;
    $db->{vspace}->{$oid} = $ob;

    if (defined $id) {
        &obj_alias($oid,$id);
    } 

    my $k; 
    foreach $k (keys %{$vp}) {
        $ob->{$k} = $vp->{$k};
    }    

    if ($cg) {
        $ob->{">$oid"} = $oid;
        $ob->{"<$oid"} = $oid;
    }
    
    &Qios::Rsm::subo_assoc($ob,$csp);
    $ob->{$chng} = $Ok;

    return $ob;
}



sub obj_fetch {
    my $id = shift;
    my $ob = undef;

    if (exists $db->{vspace}->{$id}) {
        log_enter1 "+$id";
        $ob = $db->{vspace}->{$id};

        $ob->buff_touch();

    } elsif (&Qios::Rsm::oid_exists($id)) {
        log_enter1 "-$id";
        $ob = &obj_read($id);

    } else {
        setError OsmOidNotExists, "$id";
        return undef;
    }

    return $ob;
}



sub obj_delete {
    my $ob = shift;
    log_enter0 $ob->{$idnt};

    my $id = $ob->{$idnt};
    if (!obj_exists($id)) {
        setInternal OsmObjNotExists;
        return undef;
    }

    if ($ob->{$kind}) { 
        Qios::Rsm->inx_drop_all($ob);
        Qios::Cost->stat_drop($id);
    }

    &Qios::Rsm::subo_disoc($ob);

    $ob->buff_delete();
    &Qios::Rsm::oid_free($id);                             # first id
    delete $db->{vspace}->{$ob->{$idnt}};
    if (exists $ob->{$idnt1}) {                     # check secnd idnt
        delete $db->{vspace}->{$ob->{$idnt1}};
        delete $db->{pspace}->{$ob->{$idnt1}};
    }

    my $atr;
    foreach $atr (keys %{$ob}) {
        if (ref($ob->{$atr}) eq "ARRAY") {
            undef @{$ob->{$atr}};
        }
    }

    undef %{$ob}; 

    return $Ok;
}



sub obj_dispose {
    my $ob = shift;
    log_enter0 $ob->{$idnt};

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setError OsmObjNotExists, "$id";
        return undef;
    }


    if ($ob->{$kind}) { 
        Qios::Rsm->inx_drop_all($ob);
        Qios::Cost->stat_drop($id);
    }

    $ob->buff_delete();                          
    delete $db->{vspace}->{$id};                     # first id
    if (exists $ob->{$idnt1}) {                      # check secnd id
        delete $db->{vspace}->{$ob->{$idnt1}};
    }

    my $atr;
    foreach $atr (keys %{$ob}) {
        if (ref($ob->{$atr}) eq "ARRAY") {
            undef @{$ob->{$atr}};
        }
    }

    undef %{$ob}; 

    return $Ok;
}



sub obj_set {
    my ($ob,$vp) = @_;

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setError OsmObjNotExists;
        return undef;
    }

    my ($at,$v);
    while (($at,$v) = each %$vp) {
        $at = $ob->attr_inam($at);
        $ob->{$at} = $v;
    }

    $ob->{$chng} = $Ok;

    return $Ok;
}



sub obj_get { 
    my $ob = shift;

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setError OsmObjNotExists;
        return undef;
    }

    %val = ();
    my ($at,$v);
    while (($at,$v) = each %$ob) {
        if (!&attr_sys($at)) {
            $val{$at} = $v;
        }
    }

    return \%val;
}



sub obj_write { 
    my $ob = shift;
    log_enter1 $ob->{$idnt};

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setInternal OsmObjNotExists;
        return undef;
    }

    if (!$ob->{$chng}) {
        return $Ok;
    } 

    my $rc = $ob->obj_pack();
    &Qios::Rsm::rec_write($rc,$ob->{$idnt});

    $ob->{$chng} = $False;


    return $Ok;
}



sub obj_read { 
    my $id = shift; 
    log_enter1 $id;

    if (!&Qios::Rsm::oid_exists($id)) {
        setInternal OsmObjNotExists;
        return undef;
    }

    my $ob = &buff_newframe();

    my $rc = &Qios::Rsm::rec_read($id);
    $ob->obj_unpack($rc);

    $db->{vspace}->{$ob->{$idnt}} = $ob;
    if (exists $ob->{$idnt1}) {
        $db->{vspace}->{$ob->{$idnt1}} = $ob;
    }

    return $ob;
}



sub obj_clean { 
    my $ob = shift;

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setError OsmObjNotExists;
        return undef;
    }



    my ($at,$v);
    while (($at,$v) = each %$ob) {
        if (!&attr_sys($at)) {
            if (ref($ob->{$at}) eq "ARRAY") {
                undef @{$ob->{$at}};
            }
            delete $ob->{$at};
        }
    }


    $ob->{$chng} = $Ok;

    return $Ok;
}



sub obj_clean0 { 
    my $ob = shift;
    log_enter1 $ob->{$idnt};

    my $id = $ob->{$idnt};
    if (!exists $db->{vspace}->{$id}) {
        setError OsmObjNotExists;
        return undef;
    }

    if ($ob->{$kind}) { 
        Qios::Rsm->inx_drop_all($ob);
        Qios::Cost->stat_drop($id);
    }

    delete $db->{vspace}->{$ob->{$idnt}};
    $ob->{$idnt} = $Qios::Rsm::dbnull;
    if (exists $ob->{$idnt1}) { 
        delete $db->{vspace}->{$ob->{$idnt1}};
        delete $ob->{$idnt1};
    }

    my ($at,$v);
    while (($at,$v) = each %$ob) {
        if (!&attr_sys0($at)) {
            if (ref($ob->{$at}) eq "ARRAY") {
                undef @{$ob->{$at}};
            }
            delete $ob->{$at};
        }
    }

    $ob->{$chng} = $Ok;

    return $Ok;
}



sub obj_alias {
    my ($oid,$id) = @_;

    &Qios::Rsm::oid_alias($oid,$id);
    $db->{vspace}->{$id} = &obj_fetch($oid);  
    my $ob = &obj_fetch($oid);
    $ob->{$idnt1} = $id;
    $ob->{$chng} = $Ok;

    return $Ok;
}




sub obj_pack {
    my $ob = shift;


    my @ob1 = ();
    my ($k,$val,$refob);
    W1: foreach $k (keys %{$ob}) {
 
        next W1 if ($k =~ /^\#\#.*/);

        $refob = ref($ob->{$k});
        if ($refob eq '') {
            $val = $ob->{$k};
        } elsif ($refob eq 'ARRAY') {
            $val = join "\000", (undef,@{$ob->{$k}});
        } else {  
            next W1;
        }

        push @ob1, join "\001", ($k,$val);
    }

    $Qios::Rsm::dbrec = join "\002", @ob1;

    return \$Qios::Rsm::dbrec;
}



sub obj_unpack {
    my ($ob,$rc) = @_;
    my ($k,$v,$co);
    my @al; 

    my @ob2 = split "\002", $$rc;
    
    foreach $co (@ob2) {
        ($k,$v) = split "\001", $co;

        @al = split "\000", $v;
        if ($#al <= 0) {
            $ob->{$k} = $al[0];
        } else {
            shift @al;
            $ob->{$k} = [ @al ];
        }
    }

    $ob->{$chng} = $False;

    return $Ok;
}





sub obj_exists {
    my $id = shift; 

    if (exists $db->{vspace}->{$id}) {
        return $Ok;
    } else {
        return &Qios::Rsm::oid_exists($id);
    }
}



sub obj_isclass {
    my $id = shift; 

    if (!&obj_exists($id)) {
         setError OsmObjNotExists;
         return undef;
    }

    my $ob = &obj_fetch($id);
    return ($ob->{$kind} == $iscl);
}



sub obj_isground {
    my $id = shift; 

    if (!&obj_exists($id)) {
        setError OsmObjNotExists;
        return undef;
    }
 
    my $ob = &obj_fetch($id);
    return ($ob->{$kind} == $isgr);
}





sub attr_sys0 {
    my $at = shift;
    return ($at =~ /^\#\#.*/);
}



sub attr_sys {
    my $at = shift;
    return ($at =~ /^[\#\<\>\~]/);
}



sub attr_exists {
    my ($ob,$at) = @_;

    $at = $ob->attr_inam($at);
    return exists $ob->{$at};
}



sub attr_defined {
    my ($ob,$at) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }
    return defined($ob->obj_atrfn($at));
}



sub attr_mvd {
    my ($ob,$at) = @_;

    $at = $ob->attr_inam($at);
    return (ref($ob->{$at}) eq "ARRAY");
}



sub attr_svd {
    my ($ob,$at) = @_;

    $at = $ob->attr_inam($at);
    return (!ref($ob->{$at}));
}



sub attr_inam {
    my ($ob,$at) = @_;

    if ($at =~ /^[\&\#\<\>\~]/) {
        return $at;

    } else {
        if ($ob->{$kind}) {
            return join $attrGlue, ($ob->{$idnt},$at);
        } else {
            return join $attrGlue, 
                        (&obj_fetch($ob->{$supc}->[0])->{$idnt},$at);
        }
    }
}



sub attr_set {
    my ($ob,$at,$v) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);
    $ob->{$at} = $v;
    $ob->{$chng} = $Ok;
    return $Ok;
}



sub attr_get {
    my ($ob,$at) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);


    return $ob->{$at};
}



sub attr_setmv {
    my($ob,$at,$i,$v) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);
    $ob->{$at}->[$i] = $v;
    $ob->{$chng} = $Ok;
    return $Ok;
}



sub attr_getmv {
    my ($ob,$at,$i) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);
    return $ob->{$at}->[$i];
}



sub attr_push {
    my ($ob,$at,$v) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);
    push @{$ob->{$at}}, ($v);  
    $ob->{$chng} = $Ok;
    return $Ok;
}



sub attr_unshift {
    my($ob,$at,$v) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);
    unshift @{$ob->{$at}}, ($v);
    $ob->{$chng} = $Ok;
    return $Ok;
}



sub attr_pop {
    my ($ob,$at) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);


    $ob->{$chng} = $Ok;
    return pop @{$ob->{$at}};
}



sub attr_shift {
    my($ob,$at) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);


    $ob->{$chng} = $Ok;
    return shift(@{$ob->{$at}});
}



sub attr_del {
    my ($ob,$at) = @_;

    if (ref($ob) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    }

    $at = $ob->attr_inam($at);


    if (ref($ob->{$at}) eq "REF") {
        undef @{$ob->{$at}};      
    }

    delete $ob->{$at};
    $ob->{$chng} = $Ok;
    return $Ok;
}





sub obj_atrnm {
    my ($n,$at) = @_;

    my $cl;
    if ($n->{$kind}) {
        $cl = $n;
    } else {
        $cl = $n->{$supc}->[0];
    }

    my @ac = split $attrGlue, $at;
    foreach $k (keys %{$cl}) {
        my @ak = split $attrGlue, $k;

        if (($#ac == 0) && ($#ak == 0)) { 
            if ($ac[0] eq $ak[0]) {
                return $at;
            }
        } 
        if (($#ac == 0) && ($#ak == 1)) { 
            if ($ac[0] eq $ak[1]) {
                return join($attrGlue,($ak[0],$ac[0]));
	    }
        } 
        if (($#ac == 1) && ($#ak == 0)) { 
            if ($ac[1] eq $ak[0]) {
                return $ac[1];
	    }
        } 
        if (($#ac == 1) && ($#ak == 1)) { 
            if (($ac[0] eq $ak[0]) && ($ac[1] eq $ak[1])) {
                return $at; 
	    }
        } 
    }
    return undef;
}




sub more_atts {
    my ($n1,$nm) = @_;
    my $cn = 0;
    my @kc;
    my $k1;
    foreach $k1 (keys %{$n1}) { 
        @kc = split $attrGlue, $k1;
        if ($kc[1] eq $nm) {
            $cn++
        }
    }
    return $cn > 1;
}    



sub obj_atrno {
    my ($n,$at) = @_;
    my ($cl,$k);
    my $cnt = 0;
    my @ks = ();

    my @atc = split $attrGlue, $at;
    for $k (keys %{$n}) { 

        @ks = split $attrGlue, $k;
        if ($atc[1] eq $ks[1]) {
            if ($atc[0] eq $ks[0]) {

                if (&more_atts($n,$atc[1])) {
                    undef $cl;
                    if (!defined($cl = &obj_fetch($atc[0]))) {
                        setError OsmAtrCnvErr;
                        return undef;
                    }
                    return $cl->{$idnt1}."_$atc[1]";
                } else {
                    return $atc[1];
                }

            } else {
                $cnt++;
            }
        }
    } 
    setError OsmAtrCnvErr;
    return undef;
}



sub obj_atrfn {
    my ($n,$at) = @_;

    my $cl;
    if ($n->{$kind}) {
        $cl = $n;
    } else {
        $cl = &obj_fetch($n->{$supc}->[0]);
    }

    if ($at eq $idnt) {
        return "$cl$at";
    }

    my @ac = split $attrGlue, $at;
    if ($#ac == 1) {
        $ac[0] = &obj_fetch($ac[0])->{$idnt};
    }

    my @ak;
    L1: foreach $k (keys %{$cl}) { 
        next L1 if (&attr_sys($k));
        @ak = split $attrGlue, $k;

        if (($#ac == 0) && ($#ak == 0)) { 
            if ($ac[0] eq $ak[0]) {
                return join $attrGlue, ($cl->{$idnt},$ac[0]);
            }
        } 
        if (($#ac == 0) && ($#ak == 1)) { 
            if ($ac[0] eq $ak[1]) {
                return $k;
  	    }
        } 
        if (($#ac == 1) && ($#ak == 0)) { 
            if ($ac[1] eq $ak[0]) {
                return join $attrGlue, @ac;
	    }
        } 
        if (($#ac == 1) && ($#ak == 1)) { 
            if (($ac[0] eq $ak[0]) && ($ac[1] eq $ak[1])) {
                return $k;
	    }
        } 
    }
    return undef;
}



sub obj_atts {
    my ($o,$l) = @_;

    my $k; 
    foreach $k (keys %{$o}) {
        if (!&attr_sys($k)) { 
            push @{$l}, ($k);
        }
    }
    return $Ok;
}



sub obj_bless {
    return bless {};
}





sub obj_copy {
    my ($o1,$o2) = @_;

    if ((ref($o1) ne "Qios::Osm") || (ref($o2) ne "Qios::Osm")) {
        setError OsmObjNotExists;
        return undef;
    } 

    L1: foreach $k (keys %{$o1}) { 
        if ($k !~ /^[\#\<\>]/) { 
   	    if (ref($o1->{$k}) eq "ARRAY") { 
                $o2->{$k} = [ @{$o1->{$k}} ]; 
            } else { 
                $o2->{$k} = $o1->{$k}; 
            }
        }
    }

    $o2->{$chng} = $Ok;
    return $Ok;
}



sub obj_project {
    my ($o1,$l) = @_;

    if (ref($o1) ne "Qios::Osm") {
        setError OsmObjNotExists;
        return undef;
    } 

    my $k;
    foreach $k (keys %{$o1}) { 
        if (!&attr_sys($k)) { 
            if (!&member($k,$l)) {
                delete $o1->{$k};
            }
        }
    }

    $o1->{$chng} = $Ok;
    return $Ok;
}



sub obj_diff {
    my ($o1,$o2) = @_;

    if ((ref($o1) ne "Qios::Osm") || (ref($o2) ne "Qios::Osm")) {
        setError OsmObjNotExists;
        return undef;
    } 

    my $k;
    foreach $k (keys %{$o1}) { 
        if (!&attr_sys($k)) { 
            if (exists $o2->{$k}) {
                delete $o1->{$k};
            }
        }
    }

    $o1->{$chng} = $Ok;
    return $Ok;
}





sub obj_dupl { 
    my ($o,$ty) = @_; 

    if (ref($o) ne "Qios::Osm") {
        setError ObjNotDef;
        return undef;
    }    

    if (!defined($ty)) {
        $ty = 'temp';
    }

    my $n = Qios::Osm->obj_new( [$ty],$o->{$kind} ); 

    foreach $k (keys %{$o}) { 
        if (!&attr_sys($k)) { 
  	    if (ref($o->{$k}) eq "ARRAY") { 
                $n->{$k} = [ @{$o->{$k}} ]; 
            } else { 
                $n->{$k} = $o->{$k}; 
            }
        }
    }
    return $n->{$idnt};
}



sub obj_join {
    my ($o1,$o2,$ty) = @_;

    if ((ref($o1) ne "Qios::Osm") || (ref($o2) ne "Qios::Osm")) {
        setError OsmObjNotExists;
        return undef;
    }    

    if (!defined($ty)) {
        $ty = 'temp';
    }

    my $n = Qios::Osm->obj_new( [$ty], 
                          $o1->{$kind} || $o2->{$kind} ); 

    my $k;
    foreach $k (keys %{$o1}) { 
        if (!&attr_sys($k)) { 
  	    if (ref($o1->{$k}) eq "ARRAY") { 
                $n->{$k} = [ @{$o1->{$k}} ]; 
            } else { 
                $n->{$k} = $o1->{$k}; 
            }
        }
    }

    foreach $k (keys %{$o2}) { 
        if (!&attr_sys($k)) { 
  	    if (ref($o2->{$k}) eq "ARRAY") { 
                $n->{$k} = [ @{$o2->{$k}} ]; 
            } else { 
                $n->{$k} = $o2->{$k}; 
            }
        }
    }
    return $n->{$idnt};
}







sub buff_push { 
    my $ob = shift;
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    $ob->{$bufpre} = $db->{buflst};
    $ob->{$bufnxt} = undef;   
 
    if (defined $db->{buflst}) { 
        $db->{buflst}->{$bufnxt} = $ob; 
        $db->{buflst} = $ob;
    } else {
        $db->{bufloc} = $ob;
        $db->{buflst} = $ob;
    }

    $db->{bufcnt}++;
}



sub buff_pop { 

    if (!defined $db->{buflst}) {
        setInternal OsmBuffEmpty;
        return undef;
    }

    my $ob = $db->{buflst};
    if (defined $ob->{$bufpre}) {
        $ob->{$bufpre}->{$bufnxt} = undef;
        $db->{buflst} = $ob->{$bufpre};
    } else {
        $db->{buflst} = undef;
        $db->{bufloc} = undef;
    }
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    $db->{bufcnt}--;
    return $ob;  
}



sub buff_unshift { 
    my $ob = shift;
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    $ob->{$bufnxt} = $db->{bufloc};
    $ob->{$bufpre} = undef;

    if (defined $db->{bufloc}) {
        $db->{bufloc}->{$bufpre} = $ob;
        $db->{bufloc} = $ob;
    } else { 
        $db->{bufloc} = $ob;
        $db->{buflst} = $ob;
    }

    $db->{bufcnt}++;
    return $Ok;
}



sub buff_shift { 

    if (!defined $db->{bufloc}) {
        setError OsmBuffEmpty;
        return undef;
    }

    my $ob = $db->{bufloc};
    if (defined $ob->{$bufnxt}) {
        $ob->{$bufnxt}->{$bufpre} = undef;
        $db->{bufloc} = $ob->{$bufnxt};
    } else {
        $db->{buflst} = undef;
        $db->{bufloc} = undef;
    }
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    $db->{bufcnt}--;
    return $ob;  
}



sub buff_newframe { 
    my $ob;

    if ($db->{bufcnt} >= $db->{bufsize}) {
        log_enter0 "-bufcnt:$Qios::Glb::db->{bufcnt}";

        $ob = $db->{buflst};
        if ($ob->{$chng}) {
            $ob->obj_write();
        }

        $ob->obj_clean0();
        $ob->buff_touch();

    } else {
        log_enter0 "+bufcnt:$Qios::Glb::db->{bufcnt}";
        $ob = {};
        bless $ob;

        $ob->buff_unshift();
    }
    return $ob;
}

   

sub buff_touch { 
    my $ob = shift;
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    if (defined $ob->{$bufpre}) {
        $ob->{$bufpre}->{$bufnxt} = $ob->{$bufnxt};
    } else {
        $db->{bufloc} = $ob->{$bufnxt};
        if (defined $ob->{$bufnxt}) { 
            $ob->{$bufnxt}->{$bufpre} = undef;
        }
    }

    if (defined $ob->{$bufnxt}) {
        $ob->{$bufnxt}->{$bufpre} = $ob->{$bufpre};
    } else {
        $db->{buflst} = $ob->{$bufpre};
        if (defined $ob->{$bufpre}) { 
            $ob->{$bufpre}->{$bufnxt} = undef;
        } 
    }


    $ob->{$bufnxt} = $db->{bufloc};
    $ob->{$bufpre} = undef;

    if (defined $db->{bufloc}) {
        $db->{bufloc}->{$bufpre} = $ob;
        $db->{bufloc} = $ob;
    } else { 
        $db->{bufloc} = $ob;
        $db->{buflst} = $ob;
    }

    return $Ok;
}



sub buff_delete { 
    my $ob = shift;
    log_enter1 "$ob->{$idnt}:$ob->{$idnt1}";

    if (defined $ob->{$bufpre}) {
        $ob->{$bufpre}->{$bufnxt} = $ob->{$bufnxt};
    } else {
        $db->{bufloc} = $ob->{$bufnxt};
        if (defined $ob->{$bufnxt}) { 
            $ob->{$bufnxt}->{$bufpre} = undef;
        }
    }

    if (defined $ob->{$bufnxt}) {
        $ob->{$bufnxt}->{$bufpre} = $ob->{$bufpre};
    } else {
        $db->{buflst} = $ob->{$bufpre};
        if (defined $ob->{$bufpre}) { 
            $ob->{$bufpre}->{$bufnxt} = undef;
        } 
    }

    $db->{bufcnt}--;
    $ob->{$bufpre} = undef;
    $ob->{$bufnxt} = undef;
    return $Ok;
}



sub buff_print { 
    my $pck = shift;
    log_say1 "Printing vspace buffer.\n";

    my $ob = $db->{bufloc};
    while (defined $ob) {
        Qios::Glb->print($ob);
        $ob = $ob->{$bufnxt};
    }
    return $Ok;

}



sub buff_print_ids { 
    my ($pck,$par) = @_;
    my @ids = ();

    my $ob = $db->{bufloc};
    $ob = $db->{buflst} if ($par eq '-');

    while (defined $ob) {

        if (exists $ob->{$idnt1}) {
            push @ids, ("$ob->{$idnt}($ob->{$idnt1}) ");
        } else {
            push @ids, ("$ob->{$idnt} ");
        }

        $ob = $ob->{$bufnxt} if ($par ne '-');
        $ob = $ob->{$bufpre} if ($par eq '-');
    }

    print @ids;
    print "\n";
    return $Ok;

}



sub buff_print_raw { 
    my ($pck,$par) = @_;
    log_say1 "Printing raw vspace.\n";

    my ($k,$v,$oid,$als);
    while (($k,$v) = each %{$db->{vspace}}) {
        if (exists $db->{vspace}->{$k}->{$idnt}) {
            $oid = $db->{vspace}->{$k}->{$idnt};
        }
        if (exists $db->{vspace}->{$k}->{$idnt1}) {
            $als = $db->{vspace}->{$k}->{$idnt1};
        }
        print "$k\t\t $oid/$als/$v\n";
    }
    return $Ok;
}





sub scan {
    my ($pk,$cid,$ty,$at,$av) = @_;
    my $opk = &Qios::Glb::op_name($ty);

    my $cl = &obj_fetch($cid);
    my $sc = Qios::Osm->obj_new([iterator],$isgr);
    log_info1 "$sc->{$idnt} created";
    my $scid = $sc->{$idnt};
    $sc->{phsOp} = $ty;
    $sc->{scnClass} = $cid;
    $sc->{scnCobj} = $cid;
    
    if (&member($ty,[$op{phixex},$op{phixin}])) {
        $sc->{scnIxAtr} = $at;
        $sc->{scnIxVal} = $av;
        $sc->{scnIxCval} = 0;

        if (!&Qios::Rsm::inx_defined($cl,$at)) {
            Qios::Rsm->inx_create($cl,$at,$ty); 
        }
    }

    $sc = &obj_fetch($scid);
    return $sc->{$idnt};
}



sub next {
    my ($pk,$sc) = @_;
 
    my $self;
    if (!defined ($self = &obj_fetch($sc))) {
        return undef;
    }

    my $cl = &obj_fetch($self->{scnClass}); 
    my $co = &obj_fetch($self->{scnCobj});
    if (&error()) {
        return undef;
    }

    if ($self->{phsOp} == $op{phsqin}) {
        L1: while (1) {

	    $co = &obj_fetch($co->{">".$cl->{$idnt}});
            if (&error()) {
                return undef;
            }

            if ($co == $cl) {
                if ($#{$self->{scnStack}} == -1) {
                    return undef;
                } else {
                    $co = &obj_fetch(pop @{$self->{scnStack}});
                    $cl = &obj_fetch(pop @{$self->{scnStack}});
                    if (&error()) {
                        return undef;
                    }
                    $self->{scnClass} = $cl->{$idnt};
                    next L1;
                }
            }

            if ($co->{$kind}) {
                push @{$self->{scnStack}}, ($cl->{$idnt},$co->{$idnt});
                $cl = $co;
                $self->{scnClass} = $cl->{$idnt};
            } else {
                $self->{scnCobj} = $co->{$idnt};
                return $co;
            }
        }
    } elsif ($self->{phsOp} == $op{phsqex}) {

        while (1) {
	    $co = &obj_fetch($co->{">".$cl->{$idnt}});
            if (&error()) {
                return undef;
            }
            if ($co == $cl) {
                return undef;
            } 
            if (!$co->{$kind}) {
                $self->{scnCobj} = $co->{$idnt};
                return $co;
            }
        }
    } elsif (($self->{phsOp} == $op{phixex}) || 
             ($self->{phsOp} == $op{phixin})) {

        my $at = $self->{scnIxAtr};
        my $av = $self->{scnIxVal};
        my $cx = $self->{scnIxCval}++;
        my $cl = &obj_fetch($self->{scnClass});

        if (!&Qios::Rsm::inx_defined($cl,$at)) {
            Qios::Rsm->inx_create($cl,$at,$ty); 
        }

        my $lst = $cl->{$indx}->{$at}->{$av};
        if (&error()) {
            setInternal OsmObjNotDef;
            return undef;
        }
        if (defined $lst->[$cx]) {
            return &obj_fetch($lst->[$cx]);     # oids stored!
        } else {
            return undef;
        }
    } else {
        setInternal OsmNoSuchPhsOp;
        return undef;
    }
}



sub close {
    my ($pk,$sc) = @_;

    &obj_fetch($sc)->obj_delete();
}





sub scan_attr {
    my ($ob,$ty) = @_;

    my $sc = Qios::Osm->obj_new([iterator],$isgr);
    $sc->{phsOp} = $ty;
    $sc->{scnCobj} = $ob;

    if ($ty) {
        $sc->{scnAtts} = [keys (%{$ob})];
    } else {
        $sc->{scnAtts} = [];
        my $a;
        foreach $a (keys %{$ob}) {
            if (!&attr_sys($a)) {
                push @{$sc->{scnAtts}}, ($a);
            }
        }
    }
   
    $sc->{scnCval} = 0;
    return $sc;
}



sub next_attr {
    my $sc = shift;
    return $sc->{scnAtts}->[$sc->{scnCval}++];
}



sub close_attr {
    my $sc = shift;

    undef @{$sc->{scnAtts}};
    undef %{$sc}; 
}





sub cl_empty {
    my $ob = shift;
 
    my $cn = $ob->{$idnt};
    if ($ob->{">$cn"} eq $ob->{"<$cn"}) {
        return $Ok;
    } else {
        return $False;
    }
}



sub cl_print_sub {
    my $ob = shift;
    my $obn = undef;

    Qios::Glb->print($ob);

    my $id = $ob->{$idnt};
    my $idn = $ob->{"\>$id"};
    while ($idn ne $id) {

        $obn = &obj_fetch($idn);
        if ($obn->{$kind}) {
	    $obn->cl_print_sub();
        } else {
            Qios::Glb->print($ob);
        }

        $idn = $obn->{"\>$id"};
    }
}



sub cl_print_subids {
    my ($ob,$par) = @_;
    my ($idn,$obn);

    print "\[$ob->{$idnt}($ob->{$idnt1}) ";

    my $id = $ob->{$idnt};
    $idn = $ob->{"\>$id"};
    $idn = $ob->{"\<$id"} if ($par eq '-');

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            print "No object: $idn ";
            last W1;
        }
       
        $obn = &obj_fetch($idn);
        
        if ($obn->{$kind}) {
	    $obn->cl_print_subids($par);
        } else {
            print "$obn->{$idnt} ";
        }

        $idn = $obn->{"\>$id"};
        $idn = $obn->{"\<$id"} if ($par eq '-');
    }
    print "\]";
    return $Ok;
}



sub cl_print_sbc {
    my $ob = shift;
    my $obn = undef;

    Qios::Glb->print($ob);

    my $id = $ob->{$idnt};
    my $idn = $ob->{"\<$id"};

    L1: while ($idn ne $id) {
        $obn = &obj_fetch($idn);

        if ($obn->{$kind}) {
	    $obn->cl_print_sbc();
        } else {
            last L1;
        }
        $obn = &obj_fetch($idn);   # once more to be sure it's in buff
        $idn = $obn->{"\<$id"};
    }
}



sub cl_print_sbcids {
    my ($ob,$par) = @_;
    my ($idn,$obn);

    $par = '+' if ($par ne '-');

    print "$ob->{$idnt}($ob->{$idnt1}) ";

    my $id = $ob->{$idnt};
    $idn = $ob->{"\>$id"};
    $idn = $ob->{"\<$id"} if ($par eq '-');

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            print "No object: $idn ";
            last W1;
        }
       
        $obn = &obj_fetch($idn);
        
        if ($obn->{$kind}) {
	    $obn->cl_print_sbcids($par);
        } elsif ($par eq '-') {
            last W1;
        }

        $obn = &obj_fetch($idn);    # ... it could vanish from vspace
        $idn = $obn->{"\>$id"};
        $idn = $obn->{"\<$id"} if ($par eq '-');
    }
    return $Ok;
}



sub cl_print_ext {
    my $cl = shift;

    my $sc = Qios::Osm->scan($cl->{$idnt},$op{phsqex});
    my $ob = undef;

    while (defined ($ob = Qios::Osm->next($sc))) {
        Qios::Glb->print($ob);
    }
    Qios::Osm->close($sc);
}



sub cl_print_extids {
    my ($ob,$par) = @_;
    my ($idn,$obn);

    my $id = $ob->{$idnt};
    $idn = $ob->{"\>$id"};
    $idn = $ob->{"\<$id"} if ($par eq '-');

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            print "No object: $idn ";
            last W1;
        }
       
        $obn = &obj_fetch($idn);
        
        if ($obn->{$kind}) {
            print "\n$obn->{$idnt}($obn->{$idnt1}) ";
	    $obn->cl_print_extids($par);
        } else {
            print "$obn->{$idnt} ";
        }

        $idn = $obn->{"\>$id"};
        $idn = $obn->{"\<$id"} if ($par eq '-');
    }
    return $Ok;
}



sub cl_print_ins {
    my $cl = shift;

    my $sc = Qios::Osm->scan($cl->{$idnt},$op{phsqin});
    my $ob = undef;

    while (defined ($ob = Qios::Osm->next($sc))) {
        Qios::Glb->print($ob);
    }
    Qios::Osm->close($sc);
}



sub cl_print_insids {
    my ($ob,$par) = @_;
    my ($idn,$obn);

    my $id = $ob->{$idnt};
    $idn = $ob->{"\>$id"};
    $idn = $ob->{"\<$id"} if ($par eq '-');

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            print "No object: $idn ";
            last W1;
        }
       
        $obn = &obj_fetch($idn);
        
        if ($obn->{$kind}) {
            print "\n$obn->{$idnt}($obn->{$idnt1}) ";
	    $obn->cl_print_insids($par);
        } else {
            print "$obn->{$idnt} ";
        }

        $idn = $obn->{"\>$id"};
        $idn = $obn->{"\<$id"} if ($par eq '-');
    }
    return $Ok;
}



sub cl_delete_ext {
    my $ob = shift;

    my $id = $ob->{$idnt};
    my $idn = $ob->{"\>$id"};
    my $obn = undef;

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            print "No object: $idn ";
            last W1;
        }
       
        $obn = &obj_fetch($idn);
        $idn = $obn->{"\>$id"};
        $obn->obj_delete();
    }

    return $Ok;
}



sub cl_delete_sub {
    my $ob = shift;

    my $id = $ob->{$idnt};
    my $idn = $ob->{"\>$id"};
    my $obn = undef;

    W1: while ($idn ne $id) {

        if (!obj_exists($idn)) {
            log_say1 "Error: no such id $idn "; 
            last W1;
        }

        $obn = &obj_fetch($idn);
        $idn = $obn->{"\>$id"};

        if ($obn->{$kind}) {
	    $obn->cl_delete_sub();
        } 
        $obn->obj_delete();
    }

    return $Ok;
}






sub cntx_new {
    my ($pk,$cx) = @_;

    if (!defined $db) {
        setError OsmDbNotOpened;
        return undef;
    }
    if (&obj_exists($cx)) {
        setError OsmCntxExists;
        return undef;
    }

    my $cxo = Qios::Osm->obj_new([context],$iscl,$cx);
    $cxo->{vspace}->{$cx} = $cxo;  # self ref

    return $cxo;
}



sub cntx_delete {
    my ($pk,$cx) = @_;

    if (!defined $db) {
        setError OsmDbNotOpened;
        return undef;
    }
    if (!&obj_exists($cx)) {
        setError OsmCntxNotExists;
        return undef;
    }

    return &obj_fetch($cx)->obj_delete();
}



sub cntx_open { 
    my ($pk,$cx) = @_;

    if (!defined $db) {
        setError OsmDbNotOpened;
        return undef;
    }
    if (&obj_exists($cx)) {
        setError OsmCntxExists;
        return undef;
    }

    my $cxo = &obj_read($cx);
    $cxo->{vspace}->{$cx} = $cxo; 
    return $cxo;
}



sub cntx_close { 
    my $cxo = shift;
    
    if (!defined $db) {
        setError OsmDbNotOpened;
        return undef;
    }

    undef %{$cxo->{vspace}};
    delete $cxo->{vspace};

    $cxo->release();
}



sub cntx_flush { 
    my $cxo = shift;

    $vspc = $cxo->{vspace};
    delete $cxo->{vspace};
    
    $cxo->obj_write();

    $cxo->{vspace} = $vspc;
}





sub test1 {
    my $pk = shift;

    $c1 = Qios::Osm->obj_new([object],$iscl,person);
    $c2 = Qios::Osm->obj_new([object],$iscl,thingy);
    $o1 = Qios::Osm->obj_new([person],$isgr,man);
    $o2 = Qios::Osm->obj_new([thingy],$isgr,perl);
    $o3 = Qios::Osm->obj_new([thingy],$isgr,C);

    $o3->obj_delete();
    $o2->obj_delete();
    $o1->obj_delete();
    $c2->obj_delete();
    $c1->obj_delete();

    $c1 = Qios::Osm->obj_new([object],$iscl,person);
    $c2 = Qios::Osm->obj_new([object],$iscl,thingy);
    $o1 = Qios::Osm->obj_new([person],$isgr,man);
    $o2 = Qios::Osm->obj_new([thingy],$isgr,perl);
    $o3 = Qios::Osm->obj_new([thingy],$isgr,C);
}



sub test2 {
    my $pk = shift;

    $c1 = Qios::Osm->obj_new([object],$iscl,person);   
    $c1->attr_set( name,string );
    $c1->attr_set( age,number );
    $c1->attr_set( address,string );
    $c1->attr_set( city,string );
    $c1->attr_push( friends,person );
    $c1->attr_set( work,string );

    $c2 = Qios::Osm->obj_new( [object],$iscl,thingy );   
    $c2->attr_set( name,string );
    $c2->attr_set( weight,number );
    $c2->attr_set( material,string );
    $c2->attr_set( color,string );
    $c2->attr_push( owner,person );
    $c2->attr_push( parts,thingy );

    $o2 = Qios::Osm->obj_new( [thingy],$isgr,perl );
    $o3 = Qios::Osm->obj_new( [thingy],$isgr,C );

    $rec1 = { name => "Jim",
              address => "Alte Strasse",
              city => "Merzhausen",
              work => "maler" };
    $o1 = Qios::Osm->obj_create( [person],$isgr,$rec1,'jim' );

    $o2 = Osm->Qios::obj_new( [person],$isgr,'john' );
    $o2->{nodName} = "John";
    $o2->{address} = "Berthold Strasse";
    $o2->{city} = "Freiburg";
    $o2->{work} = "builder";
    $o2->{friends} = [$o1,$o3];

    $o3 = Qios::Osm->obj_new( [person],$isgr,martin );
    $o3->attr_set( name,"Martin" );
    $o3->attr_set( address,"Lerchen Strasse" );
    $o3->attr_set( city,"Freiburg" );
    $o3->attr_set( work,"keeper" );
    $o3->attr_push( friends,$o2 );

    $c3 = Qios::Osm->obj_new( [person],$iscl,man );
    $o4 = Qios::Osm->obj_new( [man],$isgr,man1 );
    $o5 = Qios::Osm->obj_new( [man],$isgr,man2 );

    $c4 = Qios::Osm->obj_new( [person],$iscl,ona );
    $o6 = Qios::Osm->obj_new( [ona],$isgr,ona1 );
    $o7 = Qios::Osm->obj_new( [ona],$isgr,ona2 );
}


sub test3 {
    my $pk = shift;

    $c1 = Qios::Osm->obj_new( [object],$iscl,r1 );
    $c1->attr_set( name,string );
    $c1->attr_set( age,number );
    $c1->attr_set( price,number );
    $c1->attr_push( list,r2 );

    $c2 = Qios::Osm->obj_new( [object],$iscl,r2 );
    $c2->attr_set( name,string );
    $c2->attr_set( price,number );
    $c2->attr_set( size,number );
    
    $i1 = Qios::Osm->obj_new([r1],$isgr);
    $i1->{nodName} = "No";
    $i1->{age} = 40;
    $i1->{list} = [ 1,2,3 ];
}


sub test4 {
    my $pk = shift;

    my $c1 = Qios::Osm->obj_new([object],$iscl,S);    
    $c1->attr_set( $site,a1 );
    $c1->attr_set( sid,number );
    $c1->attr_set( name,string );
    $c1->attr_set( status,number );
    $c1->attr_set( addr,string );

    my $c2 = Qios::Osm->obj_new( [object],$iscl,P );    
    $c2->attr_set( $site,a1 );
    $c2->attr_set( pid,number );
    $c2->attr_set( name,string );
    $c2->attr_set( color,string );
    $c2->attr_set( price,number );
    $c2->attr_set( weight,number );
    $c2->attr_set( addr,string );

    my $c3 = Qios::Osm->obj_new([object],$iscl,J);    
    $c3->attr_set( $site,a1 );
    $c3->attr_set( jid,number );
    $c3->attr_set( name,string );
    $c3->attr_set( addr,string );

    my $c4 = Qios::Osm->obj_new([object],$iscl,SPJ);  
    $c4->attr_set( $site,a1 );
    $c4->attr_set( sid,number );
    $c4->attr_set( pid,number );
    $c4->attr_set( jid,number );
    $c4->attr_set( num,number );
}


sub test5 {
    my $pk = shift;
 
    Qios::Load->xmlRead('P.xml');
    Qios::Rsm->inx_create(&obj_fetch(P),name,3);
    Qios::Rsm->inx_create(&obj_fetch(P),pid,3);    
    Qios::Rsm->inx_drop(&obj_fetch(P),name);    
}




BEGIN { 


}



1;


