

=head1 NAME 

B<Parse.pm> -- query parsing and translation into query trees.

=head1 DESCRIPTION

This module includes the implementation of a parser for the algebra
expressions. The algebra expressions are checked for syntactical
errors and then translated into query trees. 

Lexical analysis is implemented using very simple lexer which converts
a query into a sequence of tokens including operation codes and
constants. Parsing and translation are realized by a top-down
examination of the query expressions based on methods for LL(0)
grammars.

Module is used for translating queries as well as the rules for
query optimization which are loaded into the system at the startup
time. Briefly, the rules are pairs of query expressions which
terminate by using special operators called spans; see man pages of
I<Rule.pm> for more details. 

=head1 Query tree

The query expressions are represented during the query compilation and
execution as query trees. Two types of nodes are used for the
representation of queries: the query nodes representing set operations
(operations I<select>, I<join>, etc.), and the query nodes for the
representation of parameter expressions (eg. join parameter
expression).

The basic skeleton of the query tree is constructed during the parsing
process. The variables and the names of the data sources and spans are
stored in symbol table I<%symtab>. The basic syntax of the rule
terminators is checked and the rules are inserted in the pool of
active rules. The data sources are checked and read into the object
buffer.

=head1 Symbol table

A symbol name appearing in the query expression can be either the name
of a data source, a query variable or the name of a span. The entry of
the symbol table I<%symtab> includes the name of symbol and the
reference to the query node representing data source, variable or
span. The symbol table is used for handling query variables during 
parsing and type-checking, and for preparing rules for matching.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Parse;


use Qios::Glb;
use Qios::Osm;
use Qios::Que;






BEGIN {

    $questr = '';      
    @quelst = ();      
    @quetok = ();      

    $pp = 0;
}



BEGIN {

    %token = (

    '=='        => $op{ieq},
    '='         => $op{deq},
    '=/'        => $op{leq},
    'ext'       => $op{ext},
    'ins'       => $op{ins},
    'cls'       => $op{cls},
    '@<'        => $op{lto},
    '@<='       => $op{leo},
    '@>'        => $op{gto},
    '@>='       => $op{geo},
    'lub'       => $op{lub},
    'glb'       => $op{glb},
    'set'       => $op{set},
    'in'        => $op{in},
    'subset'    => $op{subset},
    '=~'        => $op{mtch},

    '<'         => $op{lt}, 
    '<='        => $op{le}, 
    '>'         => $op{gt}, 
    '>='        => $op{ge},
    '+'         => $op{plus},
    '-'         => $op{minus},
    '*'         => $op{times},
    '/'         => $op{delete},
    '&'         => $op{addr},
    'or'        => $op{lor},
    'and'       => $op{land},
    '!'         => $op{uneg},

    'select'    => $op{select}, 
    'project'   => $op{project},
    'intrsc'    => $op{intrsc}, 
    'union'     => $op{union},  
    'join'      => $op{join}, 
    'copy'      => $op{copy},
    'group'     => $op{group},
    'unnest'    => $op{unnest},
    'flat'      => $op{flat},
    'coll'      => $op{coll},

    'name'      => $op{name},
    'const'     => $op{const},
    'attv'      => $op{attv},
    'var'       => $op{var},
    'obj'       => $op{obj},
    'fpara'     => $op{fpara},
    'cpara'     => $op{cpara},
    'tpara'     => $op{tpara},


    ':'         => $op{vdef},
    '::'        => $op{cntx},
    '#'         => $op{rdef},
    ','         => $op{lcon},
    '.'         => $op{dot},
    '('         => $op{lpar},
    ')'         => $op{rpar},
    '{'         => $op{lbrc},
    '}'         => $op{rbrc},
    '['         => $op{lbrk},
    ']'         => $op{rbrk} 
    );

    %symtab = ();

    %para = ();

    @rstck = ();

    @pvars = ();
}




sub clean() {
    my $pk = shift; 
    log_say1 "Parsing a query.\n";

    @quelst = ();
    @quetok = ();
    $questr = ();
    $pp = 0;

    @rstck = ();
    %symtab = ();
    @pvars = ();
    %para = ();                   # note; arrays are referenced from rules
}






sub tokenize {
    $questr = &trim(shift);

    $questr =~ s/([\:\#\(\)\,\.\[\]\=\>\<\@\*\+\-\&\/\!])/ $1 /g;        

    $questr =~ s/\s+/ /g; 

    $questr =~ s/([\:\=\<\>\@\/\(\{]) ([\:\=\<\>\@\/\)\}\~])/$1$2/g;

    @quelst = split / /, $questr;

    @quetok = ();
    for $str (@quelst) {
        if ($str =~ /(^[+-]?\d+)|(^[\"\']\w+[\"\']$)|(^\".*\"$)|(^\&\d+)/) {
            push @quetok, ($token{const});      # constant 
        } elsif (exists $token{$str}) { 
 	    push @quetok, ($token{$str});       # add token if exists 
        } else {
   	    push @quetok, ($token{name});       # name
        }
    }

    $pp = 0;
}



sub match_opkey {
    my $i = shift; 
 
    my ($n,$l,$k,$k1);
    my $e = undef;
    my $n = 0;
    L3: foreach $k (keys %token) {
        $l = length($k);
        if ((($k1 = substr($questr,$i,$l)) eq $k)) { # &&
            if (($k =~ /\w+/) &&
                !&member(substr($questr,$i+$l,1),$Qios::Glb::trmchr)) {
                next L3;
            } elsif (length($k1) > $n) { 
                $n = length($k1);
                $e = $k1;
            }
        }
    }
    return $e;
}



sub match_word {
    my $i = shift;

     
    my $w = '';
    while (!&member($ql[$i],$Qios::Glb::trmchr) && ($i < $#ql)) {
        $w .= $ql[$i];
        $i++;
    }
    return $w;
}



sub tokenize1 {
    $questr = &trim(shift);
    @ql = split(//, $questr);
    my ($ni,$sc) = (0,undef);


    my $i = 0;
    L: while ($i <= $#ql) {
        while ($ql[$i] =~ /\s/) { 
            $i++; 
        }
        if (&member($ql[$i],["\'","\""])) {

            $ni = $ql[$i++];
            my $tl = '';
            while (($ql[$i] ne $ni) && ($i < $#ql)) {
                print $ql[$i].":";
                $tl .= $ql[$i];
                $i++;
            }


            push @quelst, ($tl);
            push @quetok, ($token{const});
            $i++;
            next L;
        }  
        if ($ql[$i] =~ /^\d/) {
            my $tl = '';
            while ($ql[$i] =~ /\d/) {
                $tl .= $ql[$i];
                $i++;
            }
            push @quelst, ($tl);
            push @quetok, ($token{const});
            next L;
        }  
        if ($k = &match_opkey($i)) {
            $i += length($k);
            push @quelst, ($k);
            push @quetok, ($token{$k});
            next L;
        }  
        if ($k = &match_word($i)) {
            $i += length($k);
            push @quelst, ($k);
            push @quetok, ($token{name});
            next L;
        }  
    } 

    return $Ok;
}



sub printPStr {
    my $i;
    print $pp.":\n";
    for ($i=0; $i <= $#quelst; $i++) {
        print " >>" if ($i == $pp);
        print $quelst[$i];
        print "<< " if ($i == $pp);
    }
    print "\n";
}





sub parse {
    my ($pk,$s) = @_;
    
    Qios::Parse->clean();

    &tokenize1($s);
    return &con_A_expr();
}



sub con_A_expr {
    my $nd = undef;

    CE: {
        if ($quetok[$pp] == $op{select}) { 
            if ($pp++,!defined($nd = &con_select())) {
                setError ParseSelectErr;
                return undef;
            }
            last CE;
        }

        if ($quetok[$pp] == $op{project}) {
            if ($pp++,!defined($nd = &con_project())) {
                setError ParseProjectErr;
                return undef;
            } 
            last CE;
        }

        if ($quetok[$pp] == $op{join}) {
            if ($pp++,!defined($nd = &con_join())) { 
                setError ParseJoinErr;
                return undef;
	    }
            last CE;
        }

        if ($quetok[$pp] == $op{copy}) { 
  	    if ($pp++,!defined($nd = &con_copy())) {
                setError ParseCopyErr;
                return undef;
	    }
            last CE;
        }

        if ($quetok[$pp] == $op{name}) { 
  	    if (!defined($nd = &con_T_expr())) {
                setError ParseTabExprErr;
                return undef;
	    }
            last CE;
        }

        if ($quetok[$pp] == $op{rdef}) { 
  	    if ($pp++,!defined($nd = &con_span())) {
                setError ParseSpanErr;
                return undef;
	    }
            last CE;
        }

        setInternal ParseQueOpExpct;
        return undef; 
    } 

    return $nd;
}



sub con_select {
    if ($quetok[$pp] != $op{lpar}) {
        setEror ParseLParExpct;
        return undef;                    
    }
    my $nd1 = undef;
    if ($pp++,!defined($nd1 = &con_A_expr())) { 
        return undef;                    
    }

    push @rstck, ($nd1);
    if ($quetok[$pp] != $op{lcon}) {
        setError ParseLconExpct;
        return undef;                    
    }

    my $nd2 = undef;
    if (++$pp,!defined($nd2 = &con_fpara())) { 
        setError ParseFParaErr;
        return undef; 
    } 
    if ($quetok[$pp] != $op{rpar}) {
        setError ParseRParExpct;
        return undef;                    
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{select};
    $nd->{nodOuter} = $nd1;
    $nd->{nodParam} = $nd2;
    $pp++;
    return $nd;
}



sub con_project { 
    if ($quetok[$pp] != $op{lpar}) {
        setError ParseLParExpct;
        return undef;                    
    }
    my $nd1 = undef;
    if ($pp++,!defined($nd1 = &con_A_expr())) { 
        return undef; 
    }
    if ($quetok[$pp] != $op{lcon}) {  
        setError ParseLconExpct;
        return undef;                    
    }
    my $nd2 = undef;
    if ($pp++,!defined($nd2 = &con_tpara())) {  
        setError ParseTParaErr;
        return undef;                    
    }
    if ($quetok[$pp] != $op{rpar}) {
        setError ParseRParExpct;
        return undef;                    
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{project};
    $nd->{nodOuter} = $nd1;
    $nd->{nodParam} = $nd2;
    $pp++;
    return $nd;
}



sub con_join { 
    if ($quetok[$pp] != $op{lpar}) {
        setError ParseLParExpct;
        return undef;                    
    }

    my $nd1 = undef;
    if ($pp++,!defined($nd1 = &con_A_expr())) {
        return undef;                    
    }
    push @rstck, ($nd1);
    if ($quetok[$pp] != $op{lcon}) {
        setError ParseLconExpct;
        return undef;                    
    }

    my $nd2 = undef;
    if ($pp++,!defined($nd2 = &con_A_expr())) {
        return undef;                    
    }
    push @rstck, ($nd2);
    if ($quetok[$pp] != $op{lcon}) {
        setError ParseLconExpct;
        return undef;                    
    }

    my $nd3 = undef;
    if (++$pp,!defined($nd3 = &con_fpara())) { 
        setError ParseFParaErr;
        return undef; 
    } 

    if ($quetok[$pp] != $op{rpar}) {
        setError ParseRParExpct;
        return undef;                    
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{join};
    $nd->{nodOuter} = $nd1;
    $nd->{nodInner} = $nd2;
    $nd->{nodParam} = $nd3;
    $pp++;
    return $nd;
}



sub con_copy { 
    if ($quetok[$pp] != $op{lpar}) {
        setError ParseLParExpct;
        return undef;                    
    }
    my $nd1 = undef;
    if ($pp++,!defined($nd1 = &con_A_expr())) { 
        return undef; 
    }
    if ($quetok[$pp] != $op{lcon}) {  
        setError ParseLconExpct;
        return undef;                    
    }
    my $nd2 = undef;
    if ($pp++,!defined($nd2 = &con_cpara())) { 
        setError ParseCParaErr;
        return undef; 
    }
    if ($quetok[$pp] != $op{rpar}) {
        setError ParseRParExpct;
        return undef;                    
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{copy};
    $nd->{nodOuter} = $nd1;
    $nd->{nodParam} = $nd2;
    $pp++;
    return $nd;
}



sub con_cpara { 
    my $v = undef;
    my $ty = undef;
    my $nm = undef;
    if ($quetok[$pp] == $op{rdef}) {
        if ($quetok[++$pp] == $op{addr}) {  

            $ty = 2;
            if ($quetok[++$pp] != $op{name}) {  
                setError ParseNameExpct;
                return undef;
            }
            $nm = "\&".$quelst[$pp];
            $v = $symtab{$quelst[$pp]};   
        } else {

            if ($quetok[$pp] != $op{name}) {  
                setError ParseNameExpct;
                return undef;
            }
            $ty = 1;
            $nm = $quelst[$pp];   
        } 
    } else {
        if ($quetok[$pp] != $op{name}) {  
            setError ParseNameExpct;
            return undef;
        }
        $ty = 0;
        $v = $quelst[$pp];   
    } 

    my $np = Qios::Que->new();
    $np->{nodOp} = $op{cpara};
    $np->{nodCval} = $v if (defined $v);
    $np->{nodKind} = $ty;
    $np->{nodName} = $nm if (defined $nm);

    if (($ty == 1) || ($ty == 2)) {
    }
    $pp++;
    return $np;
}



sub con_T_expr { 
    my $vn = $quelst[$pp];

    if (!&Qios::Osm::obj_exists($vn)) {
        setError ParseObjNotExist;
        return undef;
    }

    if ($quetok[++$pp] != $op{dot}) { 
        setError ParseDotExpct;
        return undef; 
    } 

    if ($quetok[++$pp] != $op{ext} &&
        $quetok[$pp] != $op{ins}) {
        setError ParseTabExprExpct;
        return undef; 
    } 

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{$quelst[$pp]};
    $nd->{nodName} = $vn;
    $nd->{nodOuter} = &obj_fetch($vn)->{$Qios::Osm::idnt};

    if (defined $nd->{nodOuter}->{$Qios::Osm::site}) {
        $nd->{nodSite} = $nd->{nodOuter}->{$Qios::Osm::site};  
    } else {
        $nd->{nodSite} = $Qios::Glb::defaultSite;
    }
    $pp++;   
    return $nd;
}



sub con_span { 
    if ($quetok[$pp] != $op{name}) { 
        setError ParseNameExpct;
        return undef; 
    } 

    my $vn = $quelst[$pp];
    if (exists $symtab{$vn}) {
        $pp++;
        return $symtab{$vn};
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{span};
    $nd->{nodName} = $vn;

    $symtab{$vn} = $nd;
    $pp++;   
    return $nd;
}



sub con_fpara { 
    my @vlst = ();

    W1: while ($quetok[$pp] == $op{name}) {
        push @vlst, ($quelst[$pp]);
        last W1 if ($quetok[++$pp] != $op{lcon});
        $pp++;
    }

    my ($rs,$i);
    my %vrmap = ();
    for ($i=$#vlst; $i>=0; $i--) {
        $rs = pop @rstck;
        $vrmap{$vlst[$i]} = $rs;
        $symtab{$vlst[$i]} = $rs;
    }

    my $nd1 = undef; 
    my $pred = $False;
    if ($quetok[$pp] == $op{rdef}) { 

        if ($quetok[++$pp] == $op{name}) { 
            $v = $quelst[$pp++];
            $pred = $Ok;
        } else {
            setError ParseNameExpct;
            return undef; 
	}
    } elsif ($quetok[$pp] == $op{vdef}) {

        @pvars = ();
        if ($pp++,!defined($nd1 = &con_P_expr())) { 
            setError ParseParExprErr;
            return undef; 
        }
    } else {
        setError ParseRVdefExpct;
        return undef; 
    } 

    my $np = Qios::Que->new();
    $np->{nodOp} = $op{fpara};
    $np->{nodKind} = $pred;
    $np->{nodVRmap} = { %vrmap }; 
    $np->{nodPnam} = [ @vlst ]; 
    if ($pred) {    
        push @{$para{$v}}, ($np);
        $np->{nodName} = $v;
    } else {
        $np->{nodOuter} = $nd1;      
        $np->{nodPvars} = [ @pvars ];  
        $np->{nodBind} = $Ok;
    }
    return $np;
}



sub con_P_expr {  
    my $nd1 = undef;
    if (!defined($nd1 = &con_P_term())) { 
        return undef; 
    } 
    my $nd2 = undef;
    if (!defined($nd2 = &con_Pa_expr())) { 
        return undef; 
    } 

    if ($nd2 == $Null) { 
        return $nd1;
    } else {
        $nd2->{nodOuter} = $nd1; 
        return $nd2;
    }
}

sub con_Pa_expr {  
    if ($quetok[$pp] != $op{lor}) {
        return $Null;           
    }
    my $nd1 = undef;
    if ($pp++,!defined($nd1 = &con_P_term())) { 
        return undef; 
    } 

    my $nd2 = undef;
    if (!defined($nd2 = &con_Pa_expr())) { 
        return undef;
    } 

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{lor};
    if ($nd2 == $Null) { 
        $nd->{nodInner} = $nd1;
    } else {
        $nd2->{nodOuter} = $nd1; 
        $nd->{nodInner} = $nd2;
    }
    return $nd; 
}

sub con_P_term {  
    my $nd1 = undef;
    if (!defined($nd1 = &con_P_factor())) { 
        return undef; 
    } 
    my $nd2 = undef;
    if (!defined($nd2 = &con_Pa_term())) { 
        return undef;
    } 

    if ($nd2 == $Null) {
        return $nd1;
    } else {
        $nd2->{nodOuter} = $nd1;
        return $nd2;
    } 
}

sub con_Pa_term {  
    if ($quetok[$pp] != $op{land}) {
        return $Null;           
    }
    my $nd1 = undef; 
    if ($pp++,!defined($nd1 = &con_P_factor())) { 
        return undef; 
    } 
    my $nd2 = undef;
    if (!defined($nd2 = &con_Pa_term())) { 
        return undef;
    } 

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{land};
    if ($nd2 == $Null) { 
        $nd->{nodInner} = $nd1;
    } else {
        $nd2->{nodOuter} = $nd1; 
        $nd->{nodInner} = $nd2;
    }
    return $nd; 
}

sub con_P_factor {  
    my $nd = undef;


    if ($quetok[$pp] != $op{lpar}) { 
        if (defined($nd = &con_C_expr())) { 
            return $nd; 
	} else {
            return undef; 
        }
    } 

    if ($pp++,!defined($nd = &con_P_expr())) { 
        return undef; 
    } 
    if ($quetok[$pp] != $op{rpar}) { 
        setError ParseRParExpct;
        return undef; 
    }

    $pp++;
    return $nd;
}



sub con_tpara { 
    if ($quetok[$pp] != $op{lbrk}) { 
        setError ParseLBrkExpct;
        return undef; 
    } 

    my @tup = ();
    my $nm;
    W1: while ($quetok[++$pp] == $op{name}) {
        $nm = $quelst[$pp];

        if ($quetok[++$pp] == $op{cntx}) { 
            if ($quetok[++$pp] == $op{name}) { 
                $nm .= $Qios::Osm::attrGlue.$quelst[$pp++];
            } else {
                setError ParseNameExpct;
                return undef; 
            }
        } 

        push @tup, ($nm);
        last W1 if ($quetok[$pp] != $op{lcon});
    }

    if ($quetok[$pp] != $op{rbrk}) { 
        setError ParseRBrkExpct;
        return undef; 
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{tpara};
    $nd->{nodCval} = [ @tup ];         #\(?)
    $pp++;
    return $nd;
}



sub con_C_expr { 
    my $nd1 = undef;
    if (!defined($nd1 = &con_C_term())) { 
        return undef;
    } 
    my $nd2 = undef;
    if (!defined($nd2 = &con_binop())) { 
        setError ParseBOpExprErr;
        return undef; 
    } 
    my $nd3 = undef;
    if (!defined($nd3 = &con_C_term())) { 
        return undef;
    } 

    $nd2->{nodOuter} = $nd1;
    $nd2->{nodInner} = $nd3;
    return $nd2; 
}



sub con_C_term {
    my $nd = undef;

    if (!defined($nd = &con_constant())) {
        setError ParseConsErr;
        return undef;
    } elsif ($nd != $Null) {
        return $nd;
    }

    my $nd1 = undef;
    if (!defined($nd1 = &con_variable())) {
        setError ParseVarErr;
        return undef;
    } 

    my $nd2 = undef;
    while (defined($nd2 = &con_attribute()) && 
           ($nd2 != $Null)) {
        $nd2->{nodOuter} = $nd1;
        $nd1 = $nd2;
    } 

    if (!defined $nd2) {
        return $nd1;
    } else {
        return $nd1; 
    }
}



sub con_binop {

    if (!&member($quetok[$pp],\@binops)) {
        setError ParseBinOpExpct;
        return undef;
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $quetok[$pp];
    $pp++;
    return $nd;
}



sub con_uniop {

    if (!&member($quetok[$pp],\@uniops)) {
        setError ParseUniOpExpct;
        return undef;
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $quetok[$pp];
    $pp++;
    return $nd;
}



sub con_constant {
    if ($quetok[$pp] != $op{const} &&
        $quetok[$pp] != $op{lbrk}) { 
        return $Null;
    } 

    my $val;
    if ($quetok[$pp] == $op{lbrk}) {
        @tup = ();

        W1: while ($quetok[++$pp] == $op{const}) {
            $val = $quelst[$pp];
            $val =~ s/[\'\"]//g;
            push @tup, ($val);
            last W1 if ($quetok[++$pp] != $op{lcon});
        }
        if ($quetok[$pp] != $op{rbrk}) { 
            setError ParseRBrkExpct;
            return undef; 
        }

        $val = \@tup;
    } else {

        $val = $quelst[$pp];      
        $val =~ s/[\'\"]//g;
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{const};
    $nd->{nodCval} = $val;
    $pp++;
    return $nd;
}



sub con_attribute {
    if ($quetok[$pp] != $op{dot}) { 
        return $Null;
    } 
    my $atcl = '';
    if ($quetok[++$pp] == $op{name}) { 
        if ($quetok[$pp+1] == $op{cntx}) { 
            $atcl = &obj_fetch($quelst[$pp])->{$Qios::Osm::idnt};

            $pp += 2;
        } 
    }

    if ($quetok[$pp] != $op{name}) { 
        setError ParseNameExpct;
        return undef;
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{attv};
    if ($atcl eq '') { 
        $nd->{nodName} = $quelst[$pp];
    } else {
        $nd->{nodName} = $atcl."\#".$quelst[$pp];
    }
    $pp++;
    return $nd;
}



sub con_variable {
    if ($quetok[$pp] != $op{name}) {
        setError ParseNameExpct;
	return undef;
    }
    if (!exists $symtab{$quelst[$pp]}) {
        setError ParseVarNotExist;
        return undef;
    }

    my $nd = Qios::Que->new();
    $nd->{nodOp} = $op{var};
    $nd->{nodName} = $quelst[$pp];
    $nd->{nodOuter} = $symtab{$quelst[$pp]};

    push @pvars, ($nd);
    $pp++;
    return $nd;
}



sub test1 {
    my $pk = shift;








return $e;
}

1;

