
=head1 NAME

B<Que.pm> -- query tree manipulation and type-checking. 

=head1 DESCRIPTION

This package includes the data structures and procedures for the
representation and manipulation of query trees. Qios uses a single
representation of queries for parsing, type-checking, optimization and
query evaluation. The main skeleton of the query tree is fixed after
the parsing is completed. The skeleton, ie. operations; symbols and
links among them, is then used during the subsequent phases of the
query execution for attaching the information relevant for the
particular phase.

The main procedure implemented in this module is type-checking which
is performed on query trees. The query trees are constructed by module
I<Parse.pm> which also retrieves class objects linked to the access
methods. The module I<Que.pm> checks the type of query trees in a
bottom-up manner starting with access methods progressing toward the
root node of query tree. 

The type-checking procedure is used also for rule matching procedure
described with I<Rule.pm>. In this case, the successful computation of
type of query tree denotes formalization of the successful rule
matching. For this purpose, the type-checking procedure is defined
also for spans.

=head1 Query trees

A query tree is a graph (dag) which vertices are query nodes. The
vertices are linked with respect to the parse tree formed during
parsing of query. 

Query nodes represent the operations, operation parameters, spans
linking rules, and access methods. The query nodes store the data for
the different phases of query processing: parsing, type checking,
optimization, and query evaluation. The data elements of query 
nodes used in I<Que.pm> are as follows. 

The basic links among query nodes are generated during parsing. 
They are implemented by using the following data members of query
nodes. 

    nodOprt    -- type of operator 
    nodOuter   -- outer operand 
    nodInner   -- inner operand (in the case it exists) 
    nodParam   -- predicate of algebraic operation 
    nodName    -- name of node (meaning depends on node type) 

The parameters of the set operations I<select>, I<project> and I<join>
are abstracted away by using the separate type of query node. The
parameter node stores information about the query sub-tree
corresponding to the parameter expression; it includes references
to the root of query sub-tree, variables and the references to the
data sources of variables. 

The parameter nodes play vital role for the representation of rules
as well as for the rule matching procedure. The following are the data
members of parameter nodes used for the representation of the 
interface of parameter expression. 

    nodKind    -- kind of param (normal,rule) 
    nodBinded  -- parameter binded to param expression 
    nodVrmap   -- mapping var names to data sources 
    nodPnam    -- list of variable names (ordered by data sources)
    nodPvars   -- list of variable nodes (ordered by parse) 

Here are some of the remaining important data members of the query
nodes used for type-checking, query optimization and query
evaluation. 
 
    nodType    -- type associated to node 
    nodExpr    -- expr of query rooted in this node 
    nodCval    -- current value for iterator 
    nodSite    -- site of operation 
    nodEquc    -- reference to equiv class of query 

=head1 Type-checking

Type-checking is implemented by procedure computing the types of query
nodes bottom-up, that is, from access methods toward the root of
query tree. The class objects related to the access methods are
retrieved during the parsing process. The computation of the types of
the particular query nodes is simple since the tuple-structured types
are flat, that is, they do not include nested structured elements
except the arrays of scalars. The type of a query node is stored by
constructing a new class object.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Que;


use Qios::Glb;
use Qios::Osm;
use Qios::Eval;




BEGIN {

    $free = $nil;      # first free object in @obs /* out 20/5/00 */
    $fink = 'free';    # name of attr forming list of free objects

}




sub new() {
    my $cl = shift;
    $self = {};
    bless $self;
    return $self;
}



sub delete() {
    my $n = shift;


    undef %{$n};
}



sub deleteQuery() {
    my $n = shift;

    if (&member($n->{nodOp},
                [$op{ext},$op{ins},$op{var},$op{const}])) {
        $n->delete();
        return $Ok;
    }
 
    if (defined $n->{nodOuter}) { 
        $n->{nodOuter}->deleteQuery() 
            if (ref($n->{nodOuter}) eq 'Qios::Que');
    }
    if (defined $n->{nodInner}) {
        $n->{nodInner}->deleteQuery()
            if (ref($n->{nodInner}) eq 'Qios::Que');
    }
    if (defined $n->{nodParam}) {
        $n->{nodParam}->deleteQuery()
            if (ref($n->{nodParam}) eq 'Qios::Que');
    }

    $n->delete();
    return $Ok;
}



sub clean() {
    my $n = shift;

    delete $n->{nodExpr};
    delete $n->{nodType};
    delete $n->{nodSite};
    return $Ok;
}



sub clean_fpara() {
    my $n = shift;

    $n->clean();

    if (($n->{nodOp} == $op{var}) || ($n->{nodOp} == $op{const})) {
        return $Ok;
    }
 
    $n->{nodOuter}->clean_fpara() if (defined $n->{nodOuter});
    $n->{nodInner}->clean_fpara() if (defined $n->{nodInner});
    return $Ok;        
}



sub duplicate() {
    my $n = shift;
    my $n1 = Qios::Que->new();
    my ($k,$v);
    while (($k,$v) = each %{$n}) {
        if (ref($n->{$k}) eq "ARRAY") {
            $n1->{$k} = [ @{$v} ];
        } elsif (ref($n->{$k}) eq "HASH") {
            $n1->{$k} = { %{$v} };
	} else {
            $n1->{$k} = $v;
        }
    }
    return $n1;
}





sub atts_union() {
    my ($l1,$l2) = @_;
    
    my $k;
    foreach $k (@{$l2}) {
        push @{$l1}, ($k);
    }
    return $Ok;
}


 
sub atts_split() {
    my ($l1,$l2,$l3,$l4) = @_;
    
    my $k;
    foreach $k (@{$l1}) {
        if (!&member($k,$l2)) {
            push @{$l4}, ($k);
        } else {
            push @{$l3}, ($k);
        }
    }
    return $Ok;
}



sub atts_project {
    my ($n,$l) = @_;
    
    if ($n->{nodOp} == $op{span}) {
        return $Ok;
    }
 
    if (!defined &obj_fetch($n->{nodType})->obj_project($l)) {
        setError QueAttsProjErr;
        return undef;
    }

    if (&member($n->{nodOp},[$op{ins},$op{ext}])) {
        return $Ok;
    }
 
    my $l1 = [];
    if (!defined &obj_fetch($n->{nodType})->obj_atts($l1)) {
        setError QueAttsProjErr;
        return undef;
    }

    my $l2 = $l1;
    my $l3;
    if (&member($n->{nodOp},[$op{select},$op{join}])) {
        $l2 = []; $l3 = [];

        &atts_union($l1,$n->{nodParam}->{nodAtts});

        my $la = [];
        &obj_fetch($n->{nodOuter}->{nodType})->obj_atts($la);
        &atts_split($l1,$la,$l2,$l3);
        undef @{$la};
    } 

    if (!defined $n->{nodOuter}->atts_project($l2)) {
        setError QueCompAttsProjErr;
        return undef;
    }

    if (exists $n->{nodInner}) {
        if (!defined $n->{nodInner}->atts_project($l3)) {
            setError QueCompAttsProjErr;
            return undef;
        }
    }
    return $Ok;
}




BEGIN {
    @atts = ();
}


sub compile {
    my ($pk,$s,$cn) = @_;
    log_hey1 "Compiling a query.\n";

    my $n;
    if (!defined ($n = Qios::Parse->parse($s))) {
        &Qios::Parse::printPStr();
        return undef;
    }

    if (defined $cn) {
        $n->{nodDefty} = $cn;
    }
    
    log_say1 "Type-checking.\n";
    if (!defined $n->typ_query()) {
        &Qios::Parse::printPStr();
        return undef;
    }
    return $n;
    
}



sub typ_query {
    my $n = shift;
    my $rvl;

    return $Ok if defined($n->{nodExpr});

    C1: {
        if ($n->{nodOp} == $op{select}) {
            $rvl = $n->typ_select();
            last C1;
        } elsif ($n->{nodOp} == $op{project}) {
            $rvl = $n->typ_project();
            last C1;
        } elsif ($n->{nodOp} == $op{join}) {
            $rvl = $n->typ_join();
            last C1;
        } elsif ($n->{nodOp} == $op{copy}) {
            $rvl = $n->typ_copy();
            last C1;
        } elsif (&member($n->{nodOp},[$op{ins},$op{ext}])) {
            $rvl = $n->typ_AM();
            last C1;
        } elsif ($n->{nodOp} == $op{span}) {
            $rvl = $n->typ_span();
        }
    }
    return $rvl;
}



sub typ_select {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_query()) {
        setError QueSelectErr;
        return undef;
    }
    if (!defined $n->{nodParam}->typ_fpara()) {
        setError QueFParaErr, "of select()";
        return undef;
    }

    if (!defined &Qios::Cost::slct_select($n)) {
        setError QueSlctCompErr;
        return undef;
    }

    if (defined $n->{nodType}) {
        &obj_fetch($n->{nodType})->obj_delete();
    }

    if (!defined ($n->{nodType} = 
   	          &Qios::Osm::obj_dupl(&obj_fetch($n->{nodOuter}->{nodType}),
                                 $n->{nodDefty}))) {
        setError QueCantCreTyp;
        return undef;
    } 

    if (!defined $n->{nodOuter}->{nodSite}) {
        setError QueNoSiteInf;
        return undef;
    }
    $n->{nodSite} = $n->{nodOuter}->{nodSite};

    $n->{nodExpr} = "s(".$n->{nodOuter}->{nodExpr}.",".
		    $n->{nodParam}->{nodExpr}.")";
    return $Ok;
}



sub typ_project {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_query()) {
        setError QueProjectErr;
        return undef;
    }

    if (!defined $n->{nodParam}->{nodCval}) {
        setError QueNoCVal;
        return undef;
    }

    if (defined $n->{nodType}) {
        &obj_fetch($n->{nodType})->obj_delete();
    }

    if (!defined ($n->{nodType} = 
                  &Qios::Osm::obj_dupl(&obj_fetch($n->{nodOuter}->{nodType}),
                                 $n->{nodDefty}))) {
        setError QueCantCreTyp;
        return undef;
    }

    my $pl = $n->{nodParam}->{nodCval};
    my ($at,$nm); 
    @ls = ();
    while ($at = pop @{$pl}) {

        if (!defined ($nm = &obj_fetch($n->{nodType})->obj_atrfn($at))) {
            setError QueFNameErr, "$at";
            return undef;
        }
        push @ls, ($nm);
    }
    $n->{nodParam}->{nodCval} = \@ls;

    if (!defined $n->atts_project($n->{nodParam}->{nodCval})) {
        return undef;
    } 

    if (!defined $n->{nodOuter}->{nodSite}) {
        setError QueNoSiteInf;
        return undef;
    }
    $n->{nodSite} = $n->{nodOuter}->{nodSite};

    $n->{nodExpr} = "p(".$n->{nodOuter}->{nodExpr}.",[".
 	            join(',',@{$n->{nodParam}->{nodCval}})."])";

    return $Ok;
}



sub typ_join {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_query()) {
        setError QueJoinErr;
        return undef;
    }
    if (!defined $n->{nodInner}->typ_query()) {
        setError QueJoinErr;
        return undef;
    }
    if (!defined $n->{nodParam}->typ_fpara()) {
        setError QueFParaErr, "of join()";
        return undef;
    }

    my $t1 = &obj_fetch($n->{nodOuter}->{nodType});
    my $t2 = &obj_fetch($n->{nodInner}->{nodType});

    if (!defined &Qios::Cost::slct_join($n)) {
        setError QueSlctCompErr;
        return undef;
    }

    if (defined $n->{nodType}) {
        &obj_fetch($n->{nodType})->obj_delete();
    }

    if (!defined ($n->{nodType} = 
                  &Qios::Osm::obj_join($t1,$t2,$n->{nodDefty}))) {
        setError QueCantCreTyp;
        return undef;
    } 

    if ($n->{nodOuter}->{nodSite} ne $n->{nodInner}->{nodSite}) {
        setError QueNoSiteInf;
        return undef;
    }
    $n->{nodSite} = $n->{nodOuter}->{nodSite};

    $n->{nodExpr} = "j(".$n->{nodOuter}->{nodExpr}.",".
                    $n->{nodInner}->{nodExpr}.",".
	 	    $n->{nodParam}->{nodExpr}.")";
    return $Ok;
}



sub typ_copy {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_query()) {
        setError QueCopyErr;
        return undef;
    }

    if ($n->{nodParam}->{nodKind} == 1) {
        $n->{nodParam}->{nodCval} = 
	$Rule::crl->{rulPara}->{$n->{nodParam}->{nodName}};
    }

    if ($n->{nodParam}->{nodKind} == 2) {
        $n->{nodParam}->{nodCval} = 
	    $n->{nodParam}->{nodCval}->{nodSite};
    }

    if (!defined $n->{nodParam}->{nodCval}) {
        setError QueNoSiteInf;
        return undef;
    }
    $n->{nodSite} = $n->{nodParam}->{nodCval};

    $n->{nodExpr} = "c(".$n->{nodOuter}->{nodExpr}.",".
                    $n->{nodParam}->{nodCval}.")";
    $n->{nodType} = $n->{nodOuter}->{nodType}; 

    return $Ok;
}



sub typ_AM {
    my $n = shift;

    if (!defined $n->{nodOuter}) {
        setInternal QueNoAMClass;
        return undef;
    } 

    if (!defined $n->{nodSite}) { 
        setError QueNoSiteInf;
        $n->{nodSite} = $Qios::Glb::defaultSite;
    } 

    $n->{nodExpr} = $n->{nodOuter}."\.".&op_key($n->{nodOp});

    if (defined $n->{nodType}) {
        &obj_fetch($n->{nodType})->obj_delete();
    }

    if (!defined ($n->{nodType} = 
                  &Qios::Osm::obj_dupl(&obj_fetch($n->{nodOuter}),
                                 $n->{nodDefty}))) {
         setError QueCantCreTyp;
         return undef; 
    }

    if (!defined &obj_fetch($n->{nodOuter})->{$Qios::Osm::stat}) { 
        Qios::Cost->statistics($n->{nodOuter});
    } 
    return $Ok;
}



sub typ_span {
    my $n = shift;

    if (!defined $n->{nodSite}) {
        setError QueNoSiteInf;
        return undef;
    }

    if ((!defined $n->{nodType}) || (!defined $n->{nodExpr})) {
        setError QueNoTyp;
        return undef;
    }
    return $Ok;
}



sub typ_fpara {
    my $n = shift;

    if ($n->{nodOp} != $op{fpara}) {
        setInternal QueFParaErr;
        return undef;
    }

    if ($n->{nodKind} == $Ok) {
 
        $n->{nodType} = bool;
        $n->{nodExpr} = join(',',@{$n->{nodPnam}})."#".$n->{nodName};
 
        return $Ok;
    }

    @atts = ();
    if (!defined $n->{nodOuter}->typ_pexpr()) {
        return undef;
    }

    $n->{nodType} = $n->{nodOuter}->{nodType};
    $n->{nodAtts} = [ @atts ];

    $n->{nodExpr} = join(',',@{$n->{nodPnam}}).":".
                             $n->{nodOuter}->{nodExpr};
    return $Ok;
}



sub typ_pexpr {
    my $n = shift;

    C1: {
        if (&member($n->{nodOp},
                    [$op{var},$op{attv},$op{const}])) {
            return $n->typ_term();
        }
        if (&member($n->{nodOp},\@binops)) {
            return $n->typ_bin();
        }
        if (&member($n->{nodOp},\@uniops)) {
            return $n->typ_uni();
        }
    }

    setInternal QueNoSuchOp;
    return undef;
}



sub typ_bin {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_pexpr()) {
        return undef;
    }
    if (!defined $n->{nodInner}->typ_pexpr()) {
        return undef;
    }

    $n->{nodExpr} = $n->{nodOuter}->{nodExpr}.
                   &op_key($n->{nodOp}).
  	           $n->{nodInner}->{nodExpr};

    C1: {
        if (&member($n->{nodOp},[$op{land},$op{lor}])) {
            if ($n->{nodOuter}->{nodType} eq bool &&
                $n->{nodInner}->{nodType} eq bool) {
	        $n->{nodType} = bool;
                last C1;
            } else {
                setError QueBoolOprExpct;
                return undef;
            }
        }    

        if (&member($n->{nodOp},
                    [$op{plus},$op{minus},$op{times},$op{delete}])) {
            if ((($n->{nodOuter}->{nodType} eq 'number') ||
                 ($n->{nodOuter}->{nodType} eq 'string')) &&
                (($n->{nodInner}->{nodType} eq 'number') ||
                 ($n->{nodInner}->{nodType} eq 'string'))) {
	        $n->{nodType} = number;
                last C1;
            } else {
                setError QueNumOprExpct;
                return undef;
            }
        }

        if (&member($n->{nodOp},[$op{ieq},$op{deq},$op{mtch}])) {
            if ($Ok) {       
	        $n->{nodType} = bool;
                last C1;
            } else {
                setError QueTypNotEq;
                return undef;
            }
        }

        if (&member($n->{nodOp},
                    [$op{lt},$op{le},$op{gt},$op{ge}])) {
            if ((($n->{nodOuter}->{nodType} eq 'number') ||
                 ($n->{nodOuter}->{nodType} eq 'string')) &&
                (($n->{nodInner}->{nodType} eq 'number') ||
                 ($n->{nodInner}->{nodType} eq 'string'))) {
   	        $n->{nodType} = bool;
                last C1;
            } else {
                return undef;
            }
        }
    
        if ($n->{nodOp} == $op{in}) {
            if ($n->{nodOuter}->{nodType} eq 
                $n->{nodInner}->{nodType}->[0]) {        # (!!!)
                $n->{nodType} = bool;
            } else {
                setError QueTypNotCmp;
                return undef;
            }
        }

        if ($n->{nodOp} == $op{subset}) {
            if ($n->{nodOuter}->{nodType}->[0] eq 
                $n->{nodInner}->{nodType}->[0]) {
                $n->{nodType} = bool;
            } else {
                setError QueTypNotCmp;
                return undef;
            }
        }
    }
    return $Ok;
}



sub typ_uni {
    my $n = shift;

    if (!defined $n->{nodOuter}->typ_pexpr()) {
        return undef;
    }

    if ($n->{nodOp} == $op{uneg}) {
        $n->{nodType} = bool;
        $n->{nodExpr} = "!(".$n->{nodOuter}->{nodExpr}.")";
        return $Ok;
    }

    setInternal QueNoSuchOp;
    return undef;
}



sub typ_term {
    my $n = shift;

    if ($n->{nodOp} == $op{const}) {
        my $ta;
        if (defined ($ta = typ_const($n->{nodCval}))) {

            $n->{nodType} = $ta;

            $n->{nodExpr} = $n->{nodCval};
        } else {
            setError QueConstErr;
            return undef;
        }
        return $Ok;
    }

    if ($n->{nodOp} == $op{var}) {

        $n->{nodType} = $n->{nodOuter}->{nodType};

        $n->{nodExpr} = $n->{nodName};
        return $Ok;
    }

    if ($n->{nodOp} == $op{attv}) {

        if (!defined $n->{nodOuter}->typ_term()) {
            return undef; 
        } 

        my ($to,$an);
        if (!&Qios::Osm::obj_exists($n->{nodOuter}->{nodType})) {
            setError QueNoSuchObj, "\'$n->{nodOuter}->{nodType}\'";
            return undef;
        }
        $to = &obj_fetch($n->{nodOuter}->{nodType});

        if ($n->{nodFlnm}) {                       # already as full name?
            $an = $n->{nodName};
        } else {
            if (!defined ($an = $to->obj_atrfn($n->{nodName}))) {
                setError QueAttrNamErr, "of \'$n->{nodName}\'";
                return undef;
            }
            $n->{nodName} = $an;                   # name as in $to
            $n->{nodFlnm} = $Ok;
        }
        push @atts, ($an);

        my ($t,$ta);
        if (!defined ($t = $to->attr_get($an))) {
            setError QueNoSuchObj;
	    return undef;
        }


        $n->{nodType} = $t;

        $n->{nodExpr} = $n->{nodOuter}->{nodExpr}."\.".$n->{nodName};
        return $Ok;
    }

    setInternal QueNoSuchOp;
    return undef;    
}



sub typ_acomp {
    my $t = shift;

    if (&member($t,[number,string,bool])) {
        return $t;
    }
    
    if (ref($t) eq "ARRAY") {
        my $tm;
        if (defined ($tm = &typ_acomp($t->[0]))) {
            return [ $tm ];
        } else {
            setError QueMvdCompTypErr;
            return undef;
        }
    }

    if (defined &Qios::Osm::obj_exists($t)) {
        return $t;
    } else {
        setError QueCompTypErr;
        return undef;
    }
}



sub typ_const {
    my $t = shift;

    if (ref($t) eq "ARRAY") {
        return [ &typ_const($t->[0]) ];           # only first checked (!)
    } elsif ($t =~ /^\&\d+/) {      
        if (&Qios::Osm::obj_exists($t)) {
            return &obj_fetch($t)->{$Qios::Osm::supc}->[0];  # access to Osm (!)
        } else {
            setError QueNoSuchObj;
            return undef;
        }
    } elsif ($t =~ /^[+-]?\d+/) {      
        return "number";
    } elsif ($t =~ /\w+/) {
        return "string";
    }

    return undef;
}



sub typ_data {
    my $data = shift;

    if (ref($data) ne "ARRAY") {
        if (&Qios::Osm::obj_exists($data)) {

            if (&Qios::Osm::obj_isclass($data)) {
                return $data;
            } else {
                return &obj_fetch($data)->{$Qios::Osm::supc}->[0];
            }
        } else {

            return &typ_const($data);
        }

    } else {

        return [ &typ_data($data->[0]) ];
    }

    setInternal QueNoSuchTyp;
    return undef;
}





sub next {
    my $n = shift;

    C1: {
        if ($n->{nodOp} == $op{select}) {
            return &Qios::Eval::next_select($n);
        }
        if ($n->{nodOp} == $op{project}) {
            return &Qios::Eval::next_project($n);
        }
        if ($n->{nodOp} == $op{join}) {
            return &Qios::Eval::next_join($n);
        }
        if ($n->{nodOp} == $op{copy}) {
            return &Qios::Eval::next_copy($n);
        }
        if (&member($n->{nodOp},[$op{ins},$op{ext}])) {
            return &Qios::Eval::next_AM($n);
        }
    }

    setInternal QueNoSuchOp;
    return undef;
}



sub scan {
    my ($n,$mt,$at,$av) = @_;

    C1: {
        if ($n->{nodOp} == $op{select}) {
            return &Qios::Eval::scan_select($n,$mt,$at,$av);
        }
        if ($n->{nodOp} == $op{project}) {
            return &Qios::Eval::scan_project($n,$mt,$at,$av);
        }
        if ($n->{nodOp} == $op{join}) {
            return &Qios::Eval::scan_join($n,$mt,$at,$av);
        }
        if ($n->{nodOp} == $op{copy}) {
            return &Qios::Eval::scan_copy($n,$mt,$at,$av);
        }
        if (&member($n->{nodOp},[$op{ins},$op{ext}])) {
            return &Qios::Eval::scan_AM($n,$at,$av);
        }
    }

    setInternal QueNoSuchOp;
    return undef;
}



sub close {
    my $n = shift;

    C1: {
        if ($n->{nodOp} == $op{select}) {
            return &Qios::Eval::close_select($n);
        }
        if ($n->{nodOp} == $op{project}) {
            return &Qios::Eval::close_project($n);
        }
        if ($n->{nodOp} == $op{join}) {
            return &Qios::Eval::close_join($n);
        }
        if ($n->{nodOp} == $op{copy}) {
            return &Qios::Eval::close_copy($n);
        }
        if (&member($n->{nodOp},[$op{ins},$op{ext}])) {
            return &Qios::Eval::close_AM($n);
        }
    }

    setInternal QueNoSuchOp;
    return undef;
}






sub print_sub {
    my $n = shift;

    Qios::Glb->print_raw($n);
    my $oprt = $n->{nodOp};

    return if (&member($oprt,[$op{var},$op{ext},$op{ins}]));

    if (exists $n->{nodOuter} &&
        $n->{nodOuter} != "undef") {
        $n->{nodOuter}->print_sub();
    }
    if (exists $n->{nodInner} &&
        $n->{nodInner} != "undef") {
        $n->{nodInner}->print_sub();
    }
    if (exists $n->{nodParam} &&
        $n->{nodParam} != "undef") {
        $n->{nodParam}->print_sub();
    }
}


1;









