
=head1 NAME

I<Rsm.pm> -- record storage manager.

=head1 DESCRIPTION

Module implements data store for records representing individual
objects. Apart from the indexing structures realizing oids and basic
information form the object level, records are treated as arrays of
bytes the structure of which is known at the object level realized by
I<Osm.pm>.

The record storage manager is based on the Berkeley DB (1.x) storage
system providing access to various database formats; I<Rsm.pm> uses
hash data files. Records are stored as oid/value pairs where the
values are packed in the sequences of bytes.

=head1 ARCHITECTURE

Current version of the module provides a very simple record manager
relying heavily on DB_File package. The module can be described thru
the groups of routines serving as the interface to I<Rsm.pm>.

The first group of routines allows the manipulation of object
identifiers. Free oids obtained by releasing the object from the
database are stored in the list of free oids realized directly as
key/value pairs.

Object manager I<Osm.pm> handles relationships between instances and
class records (objects) in a similar manner as in early network and
hierarchical database storage systems. Each class is realized as a
list including all instances as well as the sub-class objects. The
implementation of the relationships among class and instance objects
follows the semantics of the object hierarchy (poset) comprising a
database is described in [].

The simple routines for storing and retrieving records (I<rec_read>,
I<rec_write>) are provided. Next, the set of routines 
(I<inx_create>, I<inx_defined>, I<inx_print>, I<inx_drop>) are
implemented to manipulate main-memory hash indexes based on a single
class attribute. Finally, the set of procedures (I<db_create>,
I<db_open>, I<db_close>, I<db_commit>, I<db_clean>) for the
manipulation of the database files are provided.

=head1 NOTES

The module I<Rsm.pm> is planned to be (should be:) rewritten in a
stand-alone file-based record storage system. The features planned:
the custom object address space tuned for storing larger text records,
efficient object packing which allows selective access to record
components, and grouping of instance objects by their classes.

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Rsm;



require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw( $db %db );

use Qios::Glb;
use Qios::Osm;
use DB_File;




BEGIN {

    $dbnull = '@@';
    $dbnew  = '$$';
    $dbrec  = undef;   # common db record for record manipulation 
}




sub oid_free {
    my $id = shift;


    if (defined $db->{oidfree}) {
        $db->{pspace}->{$id} = $dbnull.$db->{oidfree};
    } else {
        $db->{pspace}->{$id} = $dbnull;
    }
    $db->{oidfree} = $id;
    return $Ok;
}



sub oid_new {

    my $id;
    if (defined $db->{oidfree}) {

        $id = $db->{oidfree};
        $db->{oidfree} = &oid_chop($db->{pspace}->{$id});

    } else {
 
        if ($db->{oidnxt} >= $db->{dbsize}) { 
            setError RsmDbSpaceFull;
            return undef;
        }

        $id = '&'.$db->{oidnxt}++;
        $db->{$Qios::Osm::chng} = $Ok;
    }

    return $id;
}



sub oid_alias {
    my ($oid,$id) = @_;

    if (!exists $db->{pspace}->{$id}) {
        $db->{pspace}->{$id} = $oid;
    } else {
        setError RsmRecIdExist;
        return undef;
    }
    return $id;
}



sub oid_set {
    my ($en,$id) = @_;

    if (!exists $db->{vspace}->{$id}) {
        $db->{vspace}->{$id} = $en;
    } else {
        setError RsmRecIdExist;
        return undef;
    }
    return $Ok;
}



sub oid_del {
    my $id = shift; 

    delete $db->{pspace}->{$id};

    return $Ok;
}



sub oid_is_sysgen {
    my $id = shift; 
    return ($id =~ /^\&\w+$/);
}



sub oid_is_dbnew {
    my $id = shift; 
    return ($$id =~ /^$dbnew(.*)$/);
}



sub oid_is_void {
    my $id = shift;
    return ($$id =~ /^$dbnull(.*)/);
}



sub oid_is_alias {
    my $st = shift;
    return ($$st =~ /^\&\w+$/);
}



sub oid_chop {
    my $id = shift;
    
    $id =~ s/^$dbnull(.*)/$1/;
    $id = undef if ($id eq '');
    return $id;
}



sub oid_exists {
    my $id = shift; 


    if (&oid_is_sysgen($id)) {
        if (exists $db->{pspace}->{$id}) {
            my $rec = $db->{pspace}->{$id};
            return !&oid_is_void(\$rec);
        } else {
            return $False;
        }
    } else {
        if (exists $db->{pspace}->{$id}) {
            my $rec = $db->{pspace}->{$id};
            if (&oid_is_alias(\$rec)) {
                return (exists $db->{pspace}->{$rec});
            } else { 
                return $False;
	    } 
        } else {
            return $False;
        }
    } 
}





sub subo_add {
    my ($ob1,$ob2) = @_;

    my $id1 = $ob1->{$Qios::Osm::idnt};
    my $id2 = $ob2->{$Qios::Osm::idnt};
    my ($idp,$obp);

    if ($ob1->{$Qios::Osm::kind}) {
        $idp = $ob2->{"<$id2"};
	$obp = &Qios::Osm::obj_fetch($idp);
    } else {
        $idp = $id2;
	$obp = $ob2;
    }

    my $idn = $obp->{">$id2"};
    my $obn = &Qios::Osm::obj_fetch($idn);
    $obp->{">$id2"} = $id1;
    $obn->{"<$id2"} = $id1;
    $ob1->{">$id2"} = $idn;
    $ob1->{"<$id2"} = $idp;

    $obp->{$Qios::Osm::chng} = $Ok;
    $obn->{$Qios::Osm::chng} = $Ok;
    return $Ok;
}



sub subo_del {
    my ($ob1,$ob2) = @_;

    my $id1 = $ob1->{$Qios::Osm::idnt};
    my $id2 = $ob2->{$Qios::Osm::idnt};
    my $idn = $ob1->{">$id2"};
    my $idp = $ob1->{"<$id2"};

    my $obn = &Qios::Osm::obj_fetch($idn);
    my $obp = &Qios::Osm::obj_fetch($idp);
    $obp->{">$id2"} = $idn;
    $obn->{"<$id2"} = $idp;

    $obp->{$Qios::Osm::chng} = $Ok;
    $obn->{$Qios::Osm::chng} = $Ok;
    return $Ok;
}



sub subo_assoc {
    my ($ob,$csp) = @_;

    undef @{$ob->{$Qios::Osm::supc}}; 
    $ob->{$Qios::Osm::supc} = [ @$csp ];   # creates a copy?(!)

    my $cl;
    foreach $cl (@$csp) {
        &subo_add($ob,&Qios::Osm::obj_fetch($cl));
    }
    return $Ok;
}



sub subo_disoc {
    my $ob = shift;

    my $cl;
    foreach $cl (@{$ob->{$Qios::Osm::supc}}) {
        &subo_del($ob,&Qios::Osm::obj_fetch($cl));
    }
    return $Ok;
}











sub rec_read {
    my ($id,$l) = @_;
    
    if (!&oid_exists($id)) {
        setError RsmRecNotDef, $id;
        return undef;
    }
    $Rsm::dbrec = $db->{pspace}->{$id};

    if (&oid_is_alias(\$Rsm::dbrec)) {
        if (!exists $db->{pspace}->{$Rsm::dbrec}) {
            setInternal RsmAliasRdErr, "\'$dbrec\'";
            return undef;
        } 
        $Rsm::dbrec = $db->{pspace}->{$Rsm::dbrec};
    }
    return \$Rsm::dbrec;
}



sub rec_write {
    my ($rc,$id) = @_;


    $db->{pspace}->{$id} = $$rc;
}





sub inx_create {
    my ($pk,$cl,$at,$ty) = @_;
    my $opk = &Qios::Glb::op_name($ty);
    log_enter1 "$cl->{$Qios::Osm::idnt}:$at:$opk";
    log_say1 "Creating index for $at ($opk) on class $cl->{$Qios::Osm::idnt}.\n";

    if (defined $cl->{$Qios::Osm::indx}->{$at}) {
        return $Ok;
    }

    $cl->{$Qios::Osm::indx}->{$at} = {};
    my $inx = $cl->{$Qios::Osm::indx}->{$at};

    my $sc;
    if ($ty == $op{phixex}) { 
        $sc = Qios::Osm->scan($cl->{$Qios::Osm::idnt},$op{phsqex}); 
    } elsif ($ty == $op{phixin}) {
        $sc = Qios::Osm->scan($cl->{$Qios::Osm::idnt},$op{phsqin}); 
    } else {
        setInternal RsmNoSuchPhsOp;
        return undef;
    }

    my $ob;
    my $at1 = $cl->obj_atrfn($at);
    while (defined ($ob = Qios::Osm->next($sc))) {
        push @{$inx->{$ob->{$at1}}}, ($ob->{$Qios::Osm::idnt});
    }

    Qios::Osm->close($sc);
    log_exit1 $cl->{$Qios::Osm::idnt};
    return $Ok;
}



sub inx_defined {
    my ($cl,$at) = @_;

    return exists $cl->{$Qios::Osm::indx}->{$at};
}



sub inx_print {
    my ($pk,$cl,$at) = @_;
    log_hey1 "Printing index $cl.$at.\n";

    if (!exists $cl->{$Qios::Osm::indx}->{$at}) {
        print "Index for $at on $cl->{$Qios::Osm::idnt} does\'t exist.\n";
    }

    my ($k,$k1);
    my @ks = sort keys %{$cl->{$Qios::Osm::indx}->{$at}};
    foreach $k (@ks) {
        print "$k: ";
        foreach $k1 (@{$cl->{$Qios::Osm::indx}->{$at}->{$k}}) {
            print "$k1 ";
        }
        print "\n";
    }
    return $Ok;
}



sub inx_drop {
    my ($pk,$cl,$at) = @_;
    log_enter1 "$cl->{$Qios::Osm::idnt}:$at";
    log_say1 "Dropping index for $at on class $cl->{$Qios::Osm::idnt}.\n";

    my $inx = $cl->{$Qios::Osm::indx}->{$at};
    my ($k,$v);
    while (($k,$v) = each %{$inx}) {
        undef @{$v};
    }

    undef %{$inx};
    delete $cl->{$Qios::Osm::indx}->{$at};

    return $Ok;
}



sub inx_drop_all {
    my ($pk,$cl) = @_;

    my ($at,$inx);
    foreach $at (keys %{$cl->{$Qios::Osm::indx}}) {
        $inx = $cl->{$Qios::Osm::indx}->{$at};
        log_say1 "Dropping index for $at on class $cl->{$Qios::Osm::idnt}.\n";

        my ($k,$v);
        while (($k,$v) = each %{$inx}) {
            undef @{$v};
        }
    
        undef %{$inx};
        delete $cl->{$Qios::Osm::indx}->{$at};
    }

    $cl->{$Qios::Osm::indx} = undef;
    return $Ok;
}






sub db_create {
    my $pck = shift;
    log_enter0;

    if (defined $Qios::Glb::db{$Qios::Glb::db->{dbname}}) { 
        setError RsmDbNameExist;
        return undef;
    }
    log_say1 "Creating new database ".$db->{dbfile}.".\n";

    $db->{$Qios::Osm::idnt} = '&0';
    $db->{$Qios::Osm::idnt1} = 'space';
    $db->{$Qios::Osm::kind} = $Qios::Osm::iscl;
    
    $db->{bufloc} = undef;
    $db->{buflst} = undef;
    $db->{bufcnt} = 0;

    $db->{oidnxt} = 1;
    $db->{oidfree} = undef;

    $db->{dbdesc} = new DB_File::HASHINFO;            #  BTREEINFO;
    tie %{$db->{pspace}}, 'DB_File', $db->{dbfile}, 
        O_CREAT|O_RDWR, 0777, $db->{dbdesc};


    $db->{pspace}->{'&0'} = '';        # from scratch...
    my $rc = $db->obj_pack();
    &rec_write($rc,'&0');
    $db->{pspace}->{'space'} = '&0';

    Qios::Osm->obj_new([],$Qios::Osm::iscl,object);   
    Qios::Osm->obj_new([],$Qios::Osm::iscl,iterator);
    Qios::Osm->obj_new([],$Qios::Osm::iscl,temp);

    my $dbn = $db->{dbname};
    $Qios::Glb::dbname = $dbn;
    $Qios::Glb::db{$dbn} = $Qios::Glb::db;

    Qios::Glb->root_log();
    return $Ok;
}



sub db_open {
    my $pck = shift;
    log_enter0;

    if (defined $Qios::Glb::db{$Qios::Glb::db->{dbname}}) { 
        setError RsmDbNameExist;
        return undef;
    }
    log_say1 "Opening database ".$Qios::Glb::db->{dbfile}.".\n"; 

    $Qios::Glb::db->{dbdesc} = new DB_File::HASHINFO;   # BTREEINFO;
    tie %{$Qios::Glb::db->{pspace}}, 'DB_File', $Qios::Glb::db->{dbfile}, 
        O_CREAT|O_RDWR, 0777, $Qios::Glb::db->{dbdesc};

    my $rc = &rec_read('&0');
    $Qios::Glb::db->obj_unpack($rc);

    $Qios::Glb::db->{bufloc} = undef;
    $Qios::Glb::db->{buflst} = undef;
    $Qios::Glb::db->{bufcnt} = 0;


    $Qios::Glb::dbname = $Qios::Glb::db->{dbname};
    $Qios::Glb::db{$Qios::Glb::db->{dbname}} = $Qios::Glb::db;

    Qios::Glb->root_log();
    return $db;
}



sub db_close {
    my $pck = shift;
    log_enter0;
    log_say1 "Closing database ".$Qios::Glb::db->{dbfile}.".\n";

    Qios::Optor->mesh_clean();

    Qios::Rsm->db_commit(); 
    Qios::Rsm->db_clean();

    my $rc = $Qios::Glb::db->obj_pack();
    &rec_write($rc,'&0');

    delete $Qios::Glb::db{$Qios::Glb::db->{dbname}};
    
    untie %{$Qios::Glb::db->{pspace}};
    undef %{$Qios::Glb::db->{pspace}};
    undef %{$Qios::Glb::db->{vspace}};
    $Qios::Glb::dbname = undef;
    undef %{$Qios::Glb::db};
    $Qios::Glb::db = undef;

    return $Ok;
}



sub db_commit {
    my $pck = shift;
    log_enter0;
    log_say1 "Committing changes in $db->{dbfile}.\n";

    my $rc;
    my $ob = $db->{bufloc};
    while (defined $ob) {

        if ($ob->{$Qios::Osm::chng}) {
            $rc = $ob->obj_pack();
            &rec_write($rc,$ob->{$Qios::Osm::idnt});
            $ob->{$Qios::Osm::chng} = $False;
        } 

        $ob = $ob->{$Qios::Osm::bufnxt};
    }

    return $Ok;
}



sub db_clean {
    my $pck = shift;
    log_enter0;
    log_say1 "Cleaning $db->{dbname} environment.\n";
 
    my $obn = undef;
    my $ob = $db->{bufloc};
    while (defined $ob) {

        $obn = $ob->{$Qios::Osm::bufnxt};
        $ob->obj_dispose();
        $ob = $obn;
    }

    return $Ok;
}



sub db_list_pspace {
    my ($pk,$cls) = @_; 
    log_enter0;
    log_hey1 "Verifying pspace.\n";

    my @obs0 = (); 
    my @obs1 = (); 
    my @obs2 = (); 
    my @obs4 = (); 

    my ($k,$v,$ob,$isc,$id,$s1,$s2);
    W2: while (($k,$v) = each %{$db->{pspace}}) {

        if (&oid_is_sysgen($k)) {
            if (&oid_is_void(\$v)) {
                push @obs1, ("$k ");
                next W2;
            }
            push @obs0, ("$k ");
            next W2;
        }

        if (&oid_is_alias(\$v)) {
            push @obs2, ("$k ");
            next W2;
        }

        push @obs3, ("$k ");
    }    

    print "Sysgen ids\n\n";
    print @obs0;
    print "\n\nVoid ids\n\n";
    print @obs1;
    print "\n\nAliases\n\n";
    print @obs2;
    print "\n\nRest\n\n";    
    print @obs3;
    print "\n";    
    return $Ok;
}



sub db_check_pspace {
    my ($pk,$cls) = @_; 
    log_enter0;
    log_hey1 "Verifying pspace.\n";

    my @obs0 = (); 
    my @obs1 = (); 
    my @obs2 = (); 
    my @obs3 = (); 

    my ($k,$v,$ob,$isc,$id,$s1,$s2);
    W2: while (($k,$v) = each %{$db->{pspace}}) {

        if (&oid_is_sysgen($k)) {
            if (&oid_is_void(\$v)) {
                push @obs1, ("$k ");
                next W2;
            }
            push @obs0, ("$k ");
            next W2;
        }

        if (&oid_is_alias(\$v)) {
            push @obs2, ("$k ");
            next W2;
        }

        push @obs3, ($k);
    }    

    print "Sysgen ids\n\n";
    print @obs0;
    print "\n\nVoid ids\n\n";
    print @obs1;
    print "\n\nAliases\n\n";
    print @obs2;

    print "\n\nRest\n\n";    
    foreach $id (@obs3) {
	delete ${$db->{pspace}}{$id};
        print "deleted $id\n";
    }
    print "\n";    
    return $Ok;
}



sub db_print_pspace {
    my ($pk,$cls) = @_; 
    log_enter0;
    log_hey1 "Verifying pspace.\n";

    my @obs0 = ();   # classes
    my @obs1 = ();   # normal objects
    my @obs2 = ();   # aliases
    my @obs5 = ();   # rest!

    my ($k,$v,$ob,$isc,$id,$s1,$s2);
    W2: while (($k,$v) = each %{$db->{pspace}}) {

        if (!&oid_exists($k)) {
            push @obs5, ($k);
            next W2;
        }



        if (&oid_is_alias(\$v)) {
            push @obs2, ($k);
            next W2;
        }

        $ob = &Qios::Osm::obj_fetch($k);
        if ($ob->{$Qios::Osm::kind}) {
            push @obs0, ($k);
        } else {
            push @obs1, ($k);
        }
    }    

    print "\nAliases\n\n";
    foreach $id (sort @obs2) {
        $ob = &Qios::Osm::obj_fetch($id);
        if (&oid_is_sysgen($id)) {
            $s1 = $id."($ob->{$Qios::Osm::idnt1})",
        } else {
            $s1 = $id."($ob->{$Qios::Osm::idnt})",
        }
        if (defined $ob->{$Qios::Osm::supc}->[0]) { 
            $s2 = &Qios::Osm::obj_fetch($ob->{$Qios::Osm::supc}->[0])->{$Qios::Osm::idnt1};
        }
        printf "%-20s %-10s \n",$s1,$s2;
    }


    print "\nClasses\n\n";
    foreach $id (sort @obs0) {
        $ob = &Qios::Osm::obj_fetch($id);
        if (&oid_is_sysgen($id)) {
            $s1 = $id."($ob->{$Qios::Osm::idnt1})",
        } else {
            $s1 = $id."($ob->{$Qios::Osm::idnt})",
        }
        if (defined $ob->{$Qios::Osm::supc}->[0]) { 
            $s2 = &Qios::Osm::obj_fetch($ob->{$Qios::Osm::supc}->[0])->{$Qios::Osm::idnt1};
        }
        printf "%-20s %-10s \n",$s1,$s2;
    }


    print "\nGround objects\n\n";
    foreach $id (sort @obs1) {
        $ob = &Qios::Osm::obj_fetch($id);
        if (defined $ob->{$Qios::Osm::supc}->[0]) { 
            $s1 = &Qios::Osm::obj_fetch($ob->{$Qios::Osm::supc}->[0])->{$Qios::Osm::idnt1};
        }
        printf "%-20s %-10s \n", 
               $id,
               $s1;
    }

    print "\nOthers\n\n";
    foreach $id (sort @obs5) {
        printf "%-20s %-10s \n", 
               $id, 
               $db->{pspace}->{$id};

    }

    print "\nRoot info\n\n";
    print "oidfree: $db->{oidfree}\n";
    print "oidnxt: $db->{oidnxt}\n";

    return $Ok;
}



sub db_print_pspace_raw {
    my ($pk,$cls) = @_; 
    log_hey1 "Verifying dbfile $db->{dbname}.\n";

    my ($k,$v);
    while (($k,$v) = each %{$db->{pspace}}) {
        print "$k\t\t$v\n";
    }    
    return $Ok;
}





BEGIN { 

}



1;




