
=head1 NAME

B<Rule.pm> -- logical optimization rules.

=head1 DESCRIPTION 

The I<query transformation rules> are used for the transformation of
query trees into logically equivalent query trees having different
syntactical structure and therefore, potentially, a faster method for
the evaluation of queries. Module I<Rule.pm> deals with the rules for
logical level of the query optimization.

The logical optimization rules are in Qios specified in a language
that follows strictly the syntax and the semantics of the Qios query
expressions. The new constructs are added to the language to be able
to express transformation rules.

=head1 Rules

A rule is composed of the input pattern and the output pattern. The
input and output patterns are connected by means of common variables
and common data sources. The links among the variables define the
meaning of the rule. The input and output patterns, ie. query
expressions, are "terminated" using special operators called I<spans>.
A span is a virtual operation which can match any operation defined on
sets of objects. The rule (i/o) patterns can then be seen, from the
perspective of parse trees, as the upper parts of query trees with the
abstract leaves.

The examples of the logical transformation rules are presented by the
currently used rules in Qios which are given at the beginning of the
module I<Rule.pm>.

The query transformation rules are from the external form (augumented
query expressions) transformed into the representation which is based
on the query tree representation of query expressions. The input and
output patterns of the rule are represented by the input and output
query trees. These are connected by the links relating variables, data
sources and parameters of the input and output patterns of rule.

=head1 Matching

Given a query tree and a rule, the matching procedure can be viewed as
an attempt to cover the root of query tree with the input rule pattern
ie. query tree. It is realized by a procedure that descends in paralel
on both query trees and matches the corresponding query nodes. The
first part of matching completes successfully if all nodes from the
input rule pattern match the input query tree.

The second phase of matching is type-checking of the transformed input
query expression. The construction of transformed query, ie. the
target rule pattern on top of the matched input query tree, is
realized together with the successful first phase of matching since
the nodes of the input and output patterns (query trees) are linked in
spans as well as thru some operations. The output pattern of the rule
can serve as the "test" construction on which the type-checking
procedure can be executed. The rule matches completely if the
type-checking routine completes successfully. 

The final transformation phase is now easy since the output pattern
can be simply duplicated and linked to the input query tree at spans. 

=head1 COMMENTS

This module is part of the I<qios> programming environment.

=cut

package Qios::Rule;


use Qios::Glb;
use Qios::Osm;
use Qios::Que;
use Qios::Parse;




BEGIN {
    @rule = ();       

    $rules = 
"--
-- rules for relational operatios
--
match0 = select( join( #q1, #q2, x,y#p1 ), z#p2 )
subst0 = join( select( #q1, z#p2 ), #q2, x,y#p1 )
match1 = select( join( #q1, #q2, x,y#p1 ), z#p2 )
subst1 = join( #q1, select( #q2, z#p2 ), x,y#p1 )
match2 = join( select( #q1, z#p2 ), #q2, x,y#p1 )
subst2 = select( join( #q1, #q2, x,y#p1 ), z#p2 )
--
match3 = join( join( #q1, #q2, x,y#p1 ), #q3, z,u#p2 )
subst3 = join( join( #q1, #q3, z,u#p2 ), #q2, x,y#p1 )
-- 
match4 = join( #q1, #q2, x,y#p )
subst4 = join( #q2, #q1, y,x#p )
--
end.
--
-- rules for copy
--
match5 = copy( select( #q, x#p ), #s )
subst5 = select( copy( #q, #s ), x#p )
match6 = select( copy( #q, #s ), x#p )
subst6 = copy( select( #q, x#p ), #s )
--
match7 = copy( join( #q1, #q2, x,y#p ), #s )
subst7 = join( copy( #q1, #s ), copy( #q2, #s ), x,y#p )
match8 = join( copy( #q1, #s ), copy( #q2, #s ), x,y#p )
subst8 = copy( join( #q1, #q2, x,y#p ), #s )
--
match9 = join( copy( #q1, #s1 ), #q2, x,y#p )
subst9 = copy( join( #q1, copy( #q2, #\&q1 ), x,y#p ), #s1 )
match10 = join( #q1, copy( #q2, #s1 ), x,y#p )
subst10 = copy( join( copy( #q1, #\&q2 ), #q2, x,y#p ), #s1 )
--
end.
--
-- right-deep --
match2 = select( join( #s1, #s2, x,y#p1 ), z#p2 )
subst2 = join( #s1, select( #s2, z#p2 ), x,y#p1 )";

}




sub free() {
    my $self = shift;
    undef %{$self};
}



sub new() {
    my $cl = shift;
    $self = {};
    bless $self;
    return $self;
}



sub clean() {
    my ($r,$dt) = @_;

    my ($k,$pi,$po);
    foreach $k (keys %{$r->{rulPara}}) {
        if (ref($r->{rulPara}->{$k}) eq "ARRAY" ) {

            $pi = $r->{rulPara}->{$k}->[0];
            $po = $r->{rulPara}->{$k}->[1];

            if ($#{$r->{rulPara}->{$k}} == 2) {
                pop @{$r->{rulPara}->{$k}};
            }

            $po->{nodKind} = 1;  
            if (exists $po->{nodOuter}) {
                delete $po->{nodOuter};
            }
            if (exists $po->{nodPvars}) {
                delete $po->{nodPvars};
            }

            if (exists $pi->{savePnam} && 
                exists $pi->{saveVRmap}) {
                $po->{nodPnam} = $pi->{savePnam};
                $po->{nodVRmap} = $pi->{saveVRmap};
            }

            if (exists $pi->{saveVRmap}) {
                delete($pi->{saveVRmap});
            }
            if (exists $pi->{savePnam}) {
                delete $pi->{savePnam};
            }
            if (exists $pi->{rulVVmap}) {
                undef %{$pi->{rulVVmap}};
                delete $pi->{rulVVmap};
            }    
        } else {

            $r->{rulPara}->{$k} = undef;
        }
    }

    &clean_outpat($r->{rulOutpat},$dt);
    return $Ok;    
}



sub clean_outpat() {
    my ($n,$dt) = @_;

    if ($n->{nodOp} == $op{span}) {
        $n->clean();
        return $Ok; 
    }

    if ($dt && &Qios::Osm::obj_exists($n->{nodType})) {
        &obj_fetch($n->{nodType})->obj_delete();
    }
    $n->clean();

    &clean_outpat($n->{nodOuter},$dt) if (defined $n->{nodOuter});
    &clean_outpat($n->{nodInner},$dt) if (defined $n->{nodInner});
    &clean_outpat($n->{nodParam},$dt) if (defined $n->{nodParam});
    return $Ok;        
}



BEGIN {
    $crl = undef;
}



sub apply() {
    my $r = shift;

    my $n = &match_apply($r->{rulOutpat});


    return $n    
}



sub match_apply() {
    my $np = shift;

    my ($nc,$o1,$o2,$p);
    CA: {
        if ($np->{nodOp} == $op{select}) {

            if (exists $Optor::qexp{$np->{nodExpr}}) {
                if (&Qios::Osm::obj_exists($np->{nodType})) {
                    &obj_fetch($np->{nodType})->obj_delete();
                }
                $nc = $Optor::qexp{$np->{nodExpr}};

            } else {
                $o1 = &match_apply($np->{nodOuter});
 	        $p  = &match_apply($np->{nodParam});

                $nc = Qios::Que->new();
                &ncopy($np,$nc);
                $nc->{nodOuter} = $o1;
                $nc->{nodParam} = $p;
	    }

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc;
        } 

        if ($np->{nodOp} == $op{project}) {

            if (exists $Optor::qexp{$np->{nodExpr}}) {
                if (&Qios::Osm::obj_exists($np->{nodType})) {
                    &obj_fetch($np->{nodType})->obj_delete();
                }
                $nc = $Optor::qexp{$np->{nodExpr}};

            } else {
                $o1 = &match_apply($np->{nodOuter});
 	        $p  = &match_apply($np->{nodParam});

                $nc = Qios::Que->new();
                &ncopy($np,$nc);
                $nc->{nodOuter} = $o1;
                $nc->{nodParam} = $p;
	    }

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc;
        } 

        if ($np->{nodOp} == $op{join}) {

            if (exists $Optor::qexp{$np->{nodExpr}}) {
                if (&Qios::Osm::obj_exists($np->{nodType})) {
                    &obj_fetch($np->{nodType})->obj_delete();
                }
                $nc = $Optor::qexp{$np->{nodExpr}};

            } else {
                $o1 = &match_apply($np->{nodOuter});
                $o2 = &match_apply($np->{nodInner});
 	        $p  = &match_apply($np->{nodParam});

                $nc = Qios::Que->new();
                &ncopy($np,$nc);
                $nc->{nodOuter} = $o1;
                $nc->{nodInner} = $o2;
                $nc->{nodParam} = $p;
	    }

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc; 
        } 

        if ($np->{nodOp} == $op{copy}) {

            if (exists $Optor::qexp{$np->{nodExpr}}) {
                if (&Qios::Osm::obj_exists($np->{nodType})) {
                    &obj_fetch($np->{nodType})->obj_delete();
                }
                $nc = $Optor::qexp{$np->{nodExpr}};

            } else {
                $o1 = &match_apply($np->{nodOuter});
                $o2 = &match_apply($np->{nodParam});

                $nc = Qios::Que->new();
                &ncopy($np,$nc);
                $nc->{nodOuter} = $o1;
                $nc->{nodParam} = $o2;
	    }

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc;
        } 

        if ($np->{nodOp} == $op{fpara}) {

            $nc = Qios::Que->new();
            &ncopy($np,$nc);

            my $pi = $crl->{rulPara}->{$np->{nodName}}->[0];  # ref to inpat
            my $v1;

            foreach $v1 (keys %{$np->{nodVRmap}}) { 
                $nc->{nodVRmap}->{$v1} = $np->{nodVRmap}->{$v1}->{rulNewi}; 
            }
            $nc->{nodBind} = $False;

            $np->{nodPnam} = $pi->{savePnam};
            $np->{nodVRmap} = $pi->{saveVRmap};

            delete($nc->{nodName}) if (exists $nc->{nodName});   

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc;
        } 

        if ($np->{nodOp} == $op{cpara}) {

            $nc = Qios::Que->new();
            &ncopy($np,$nc);

            $nc->{nodKind} = 0; 

            delete($nc->{nodName}) if (exists $nc->{nodName});   

            $np->clean();
            $np->{rulNewi} = $nc;
            return $nc;
        } 

        if ($np->{nodOp} == $op{span}) {

            $np->{rulNewi} = $np->{match};

            $np->clean();

            return $np->{match};
        } 
        return $False;
    }
}



sub ncopy() {
    my ($n1,$n2) = @_;
    my ($k,$v);
    while (($k,$v) = each %{$n1}) {
        $n2->{$k} = $v;
    }
}



sub match() {
    my $r = shift;
    my $nd = shift;

    $crl = $r;

    my $satts = [];
    &obj_fetch($nd->{nodType})->obj_atts($satts);
    
    if (!&match_query($r->{rulInpat},$nd)) {
        return $False;
    }

    foreach $p (keys %{$r->{rulPara}}) {

        if (ref($r->{rulPara}->{$p}) eq "ARRAY" ) {

            my $po = $r->{rulPara}->{$p}->[1];
            my $pm = $r->{rulPara}->{$p}->[2];

            $po->{nodOuter} = $pm->{nodOuter};
            $po->{nodPvars} = $pm->{nodPvars}; 


            my $n = undef;
            foreach $n (@{$po->{nodPvars}}) {
                 $n->{nodOuter} = $po->{nodVRmap}->{$n->{nodName}};
	    }

            $po->{nodKind} = 0;    # ordinary param
        }
    }

    if (!($r->{rulOutpat}->typ_query())) {
        &cleanErrors();
        return $False;
    }

    $r->{rulOutpat}->atts_project($satts);

    return $Ok;
}



sub match_fpara() {
    my ($pi,$pm) = @_;

    push(@{$crl->{rulPara}->{$pi->{nodName}}},($pm));

    my $fnd = $Ok;
    my ($v1,$v2);
    my $po = $crl->{rulPara}->{$pi->{nodName}}->[1];   # param in outpat

    F1: foreach $v1 (keys %{$pi->{nodVRmap}}) {
        foreach $v2 (keys %{$pm->{nodVRmap}}) {
  	    if ($pi->{nodVRmap}->{$v1}->{match} == $pm->{nodVRmap}->{$v2}) {
                $pi->{rulVVmap}->{$v1} = $v2;
                next F1;
	    }
        }
        $fnd = $False;
        last F1;
    }
    return $False if !$fnd;

    $pi->{saveVRmap} = $po->{nodVRmap};
    $po->{nodVRmap} = {};
 

    foreach $v1 (keys %{$pi->{saveVRmap}}) {    
         $v2 = $pi->{rulVVmap}->{$v1};
         $po->{nodVRmap}->{$v2} = $pi->{saveVRmap}->{$v1}; 
    }

    $pi->{savePnam} = $po->{nodPnam};  # save it
    $po->{nodPnam} = (); 
    foreach $v1 (@{$pi->{savePnam}}) {
         push(@{$po->{nodPnam}},($pi->{rulVVmap}->{$v1}));
    }

    return $Ok;
}



sub match_cpara() {
    my ($p1,$p2) = @_;

    if ($p1->{nodKind} == 1) {
        $crl->{rulPara}->{$p1->{nodName}} = $p2->{nodCval};
    }

    return $Ok;
}



sub match_query() {
    my $r = shift;  # rule root
    my $n = shift;  # query root

    if ($r->{nodOp} != $n->{nodOp} &&
        $r->{nodOp} != $op{span}) {
        return $False;
    }

    $r->{match} = $n;

    my $m = undef;
    CA: {
        if ($r->{nodOp} == $op{select}) {
            $m = &match_query($r->{nodOuter},$n->{nodOuter});
 	    $m &&= &match_fpara($r->{nodParam},$n->{nodParam}) if $m;
            return $m;
        } 

        if ($r->{nodOp} == $op{project}) {
            $m = &match_query($r->{nodOuter},$n->{nodOuter});
            return $m;
        } 

        if ($r->{nodOp} == $op{join}) {
            $m = &match_query($r->{nodOuter},$n->{nodOuter});
            $m &&= &match_query($r->{nodInner},$n->{nodInner}) if $m;
	    $m &&= &match_fpara($r->{nodParam},$n->{nodParam}) if $m;
            return $m;
        } 

        if ($r->{nodOp} == $op{copy}) {
            $m = &match_query($r->{nodOuter},$n->{nodOuter});
            $m &&= &match_cpara($r->{nodParam},$n->{nodParam}) if $m;
            return $m;
        } 

        if ($r->{nodOp} == $op{span}) {
            $r->{nodSite} = $n->{nodSite};
            $r->{nodType} = $n->{nodType};
            $r->{nodExpr} = $n->{nodExpr};
            return $Ok;
        } 
        return $False;
    }
}



sub load() { 
    my $p = shift;  # package
    log_say1 "Loading rules from \'$fn\'.\n";


    my @line = ();
    my ($l1,$l2);
    my @rs = split /[\r\f\n]+/, $rules;

    W: while () {
        $l1 = shift @rs;

        last W if ($l1 eq 'end.');
        next W if ($l1 =~ /^\-\-/);

        @line = split(/=/,$l1);
        &Qios::Parse::tokenize($line[1]);
        my $r = Qios::Rule->new();
        $r->{rulInpat} = &Qios::Parse::con_A_expr();

        my $l2 = shift @rs;
        @line = split(/=/,$l2);
        &Qios::Parse::tokenize($line[1]);
        $r->{rulOutpat} = &Qios::Parse::con_A_expr();

        $r->{rulPara} = { %Qios::Parse::para };

        push(@rule,($r));
        Qios::Parse->clean();
    }  
    return $Ok;
}



1;


